Package: mda
Version: 0.5-5
Date: 2024-11-05
Authors@R: c(person(given = "Trevor",
                      family = "Hastie",
                      role = c("aut","cre"),
		      comment="Original co-author of the S package `mda`",
                      email = "hastie@stanford.edu"),
		      person(given =  "Robert",
                      family = "Tibshirani",
                      role = "aut",
		      comment="Original co-author of the S package `mda`"),
		      person(given =  "Balasubramanian",
                      family = "Narasimhan",
                      role = "ctb",
		      comment="Contributed to the upgrading of code"),
	      person(given =  "Friedrich",
                      family = "Leisch",
                      role = "ctb",
		      comment="Original R port from the S package"),
              person(given = "Kurt",
                      family = "Hornik",
                      role = "ctb",
		      comment="Original R port from the S package"),
              person(given = "Brian",  
                      family = "Ripley",
                      role = "ctb",
		      comment="Original R port from the S package"))
Description: Mixture and flexible discriminant analysis, multivariate
        adaptive regression splines (MARS), BRUTO, and vector-response smoothing splines.
	Hastie, Tibshirani and Friedman (2009) "Elements of Statistical Learning (second edition, chap 12)" Springer, New York. 
Title: Mixture and Flexible Discriminant Analysis
Depends: R (>= 3.5.0), stats, class
Suggests: earth, testthat
License: GPL-2
Packaged: 2024-11-07 00:50:42 UTC; hastie
Repository: CRAN
Date/Publication: 2024-11-07 19:50:05 UTC
NeedsCompilation: yes
Author: Trevor Hastie [aut, cre] (Original co-author of the S package `mda`),
  Robert Tibshirani [aut] (Original co-author of the S package `mda`),
  Balasubramanian Narasimhan [ctb] (Contributed to the upgrading of code),
  Friedrich Leisch [ctb] (Original R port from the S package),
  Kurt Hornik [ctb] (Original R port from the S package),
  Brian Ripley [ctb] (Original R port from the S package)
Maintainer: Trevor Hastie <hastie@stanford.edu>
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 13:42:32 UTC; unix
Archs: mda.so.dSYM

Package: mda
Version: 0.5-5
Date: 2024-11-05
Authors@R: c(person(given = "Trevor",
                      family = "Hastie",
                      role = c("aut","cre"),
		      comment="Original co-author of the S package `mda`",
                      email = "hastie@stanford.edu"),
		      person(given =  "Robert",
                      family = "Tibshirani",
                      role = "aut",
		      comment="Original co-author of the S package `mda`"),
		      person(given =  "Balasubramanian",
                      family = "Narasimhan",
                      role = "ctb",
		      comment="Contributed to the upgrading of code"),
	      person(given =  "Friedrich",
                      family = "Leisch",
                      role = "ctb",
		      comment="Original R port from the S package"),
              person(given = "Kurt",
                      family = "Hornik",
                      role = "ctb",
		      comment="Original R port from the S package"),
              person(given = "Brian",  
                      family = "Ripley",
                      role = "ctb",
		      comment="Original R port from the S package"))
Description: Mixture and flexible discriminant analysis, multivariate
        adaptive regression splines (MARS), BRUTO, and vector-response smoothing splines.
	Hastie, Tibshirani and Friedman (2009) "Elements of Statistical Learning (second edition, chap 12)" Springer, New York. 
Title: Mixture and Flexible Discriminant Analysis
Depends: R (>= 3.5.0), stats, class
Suggests: earth, testthat
License: GPL-2
Packaged: 2024-11-07 00:50:42 UTC; hastie
Repository: CRAN
Date/Publication: 2024-11-07 19:50:05 UTC
NeedsCompilation: yes
Author: Trevor Hastie [aut, cre] (Original co-author of the S package `mda`),
  Robert Tibshirani [aut] (Original co-author of the S package `mda`),
  Balasubramanian Narasimhan [ctb] (Contributed to the upgrading of code),
  Friedrich Leisch [ctb] (Original R port from the S package),
  Kurt Hornik [ctb] (Original R port from the S package),
  Brian Ripley [ctb] (Original R port from the S package)
Maintainer: Trevor Hastie <hastie@stanford.edu>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 17:17:32 UTC; unix
Archs: mda.so.dSYM

useDynLib("mda")
importFrom(class,lvqinit, olvq1, lvq3, lvqtest)
importFrom(utils,head,tail,packageDescription)
importFrom("graphics", "par", "plot", "points")
importFrom("stats", "contr.helmert", "delete.response", "fitted",
            "kmeans", "model.extract", "model.frame", "model.matrix",
            "predict", "terms", "update", "var")

export("bruto", "coef.fda",
       "confusion",
       "fda",
       "gen.ridge", "laplacian","mspline",
       "mars", "mda", "mda.start",
       "polyreg",
       "softmax")

S3method("coef", "fda")
S3method("confusion", "default")
S3method("confusion", "fda")
S3method("confusion", "list")
S3method("model.matrix", "mars")
S3method("plot", "fda")
S3method("predict", "bruto")
S3method("predict", "fda")
S3method("predict", "gen.ridge")
S3method("predict", "mars")
S3method("predict", "mda")
S3method("predict", "polyreg")
S3method("print", "fda")
S3method("print", "mda")

This is a first shot at porting the Hastie & Tibshirani mda package to
R.  mda and fda with MARS seem to work fine, but bruto is still broken
and segfaults.  Hopefully, this can be fixed soon.

Kurt Hornik
Friedrich Leisch

Actually, mda was not fine: it was doing the pruning incorrectly.
bruto now works.

Brian Ripley 99/08/17

subroutine marss(nx,n,p,nclass,y,x,w,tagx,maxorder,mmax,penalty,thresh,forwstep,interms,prune,bx,fullin,lenb, bestgcv, bestin, flag, cut,dir, res,alpha,beta,
scrat,iscrat,trace)
#
# multivariate mars
#
# input 
#
# integer nx- number of rows of x, bx, and tagx
# integer n- number of observations
# integer p- number of variables
# integer nclass - number of classes (resposne variables)
# double precision y(n,nclass)- matrix of response values
# double precision x(nx,p) -matrix of predictors
# double precision w(n)- weights (ge 0)
# integer tagx(nx,p) - tag array for sorting x
# integer maxorder- maxumum order of interaction (mi in Friedman's notation)
#
# integer mmax- maximum number of basis functions (nk in  Friedman's notation)
#     if it is desired to add say 3 terms to an existing model, one can set mmax=interms+3
#     and set interms appropriately (see below)
#
# double precision penalty- cost per knot selection (df in Friedman's notation)
#
# double precision thresh- forward stepwise threshold - forw step stops   if (rss/rssnull).le.thresh)
#    to prevent numerical problems with overfit solutions
#
# logical forwstep- if true, forward stepwise is carried out. if false, procedure
#   starts with basis matrix bx and carried out backward stepwise. This saves
#   a great deal of time in choosing the penalty parameter 
#
# integer interms- number of terms in existing model. for a fresh start interms=1.
#
# logical prune- true if pruning (backward stepwise deletion) is desired
#
#
#
# 
# 
# double precision bx(nx,mmax)- basis matrix- used on input only if interms>1 or forwstep=false
#
# integer fullin(mmax)- basis column indicator. columns marked with a 1 indicate active
#    columns-  used on input only if interms>1 or forwstep=false
#    
# integer lenb- number of active columns of bx -
#      used on input only if interms>1 or forwstep=false
#
# double precision res(n,nclass) - residuals - used as input only if interms>1, in which case they
# should be the residuals from the current fitted model
#
# double precision scrat(1)- scratch array of length at least 
# 1+n+ 2*n*mmax +4*mmax*mmax +3*mmax + 3*mmax*nclass +3*nclass
#                                               +28n+51 (for linpack)
#
# integer iscrat(1)- scratch array of length at least 4*mmax
#
#logical trace - should there be a printing of progress
#
#
# output
#
# double precision bx(nx,mmax)- basis matrix. only first lenb columns are active. first column is 1
# integer fullin(mmax)- basis column indicator. columns marked with a 1 indicate
#    the linearly indep set 
# integer lenb- number of active columns of bx
# double precision bestgcv- gcv for best model
# integer bestin(mmax) basis column indicator for best model. 1 indicates term is in model
# integer flag(mmax,p)- ijth element is 1 if predictor j is in model term i 
#    (zero otherwise). note for a zero column of bx (fullin=0), the row of flag
#    will be all 1, for a technical reason, it should be ignored of course.
#
# double precision cut(mmax,p) - ijth element cutpoint for variable j in model term i
# double precision dir(mmax,p) - ijth element is 1 if factor involving predictor j in term i is
#  of the form (x-t)+ ; it is -1 if it has the form (t-x)+
#
# double precision res(n,nclass) - residuals from final fit
# double precision alpha(nclass)- intercepts from final fit
# double precision beta(mmax,nclass)- slopes from final fit



implicit double precision (a-h,o-z)

integer nx, n,p,nclass,tagx(nx,p),maxorder,mmax,bestin(mmax),flag(mmax,p),
fullin(mmax)
double precision y(n,nclass),x(nx,p),w(n),bx(nx,mmax),bestgcv,cut(mmax,p),dir(mmax,p),res(nx,nclass),
alpha(nclass),beta(mmax,nclass)

double precision scrat(1)
integer iscrat(1)
logical forwstep, prune, trace, tracec
common tracec
tracec=trace



# iscrat must be of len at least 3*mmax
# scrat must be of length at least 
#  1+n+ 2*n*mmax +4*mmax*mmax +3*mmax + 3*mmax*nclass +3*nclass

#1. bxorth
#2. bxorthm
#3. cov
#4. covsy
#5. ybar
#6. scr1
#7. scr5
#8. scr6
#9.  temp

#10. bxsc
#11. r
#12. betasc,
#13
#14
#15  work

len1=n*mmax
len2=mmax
len3=mmax*mmax
len4=mmax*nclass
len5=nclass
len6=mmax
len7=mmax
len8=nclass
len9=n


len10=n*mmax
len11=mmax*mmax
len12=mmax*nclass
len13=mmax*mmax
len14=mmax*mmax

n1=1
n2=n1+len1
n3=n2+len2
n4=n3+len3
n5=n4+len4
n6=n5+len5
n7=n6+len6
n8=n7+len7
n9=n8+len8
n10=n9+len9
n11=n10+len10
n12=n11+len11
n13=n12+len12
n14=n13+len13
n15=n14+len14




call marsnew1(nx,n,p,nclass,y,x,w,tagx,maxorder,mmax,bx,bestgcv, bestin,fullin,lenb,flag, cut,dir,res,
alpha,beta,penalty,thresh,forwstep,interms,prune,
scrat,scrat(n2),scrat(n3),scrat(n4),scrat(n5),
scrat(n6),scrat(n7),scrat(n8),scrat(n9),
scrat(n10),scrat(n11),scrat(n12),scrat(n13),scrat(n14),scrat(n15),
iscrat,iscrat(1+mmax),iscrat(1+2*mmax),iscrat(1+3*mmax))

return
end

subroutine marsnew1(nx,n,p,nclass,y,x,w,tagx,maxorder,mmax,bx,bestgcv, bestin,fullin,lenb,flag, cut,dir,res,alpha,beta,penalty,thresh,forwstep,interms,prune,
bxorth,bxorthm,cov,covsy,ybar,
scr1,scr5,scr6, temp, 
  bxsc, r, betasc,varsc,var,work,
termlen,in, tempin, qpivot)
# input n,p,nclass,x,tagx,y,w,mmax,maxorder

# nclass is # of response variables

# output- bestin, bestgcv, flag, cut,dir, bx,res

implicit double precision (a-h,o-z)

integer n,nterms2,p,mmax,flag(mmax,p),tagx(nx,p),termlen(mmax), nclass,fullin(mmax)

double precision cov(mmax,mmax),covsy(mmax,nclass),
critmax,x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),y(n,nclass),ybar(nclass),
scr1(mmax),scr5(mmax),scr6(nclass)

double precision temp(n),w(n), cut(mmax,p),dir(mmax,p),alpha(nclass),beta(mmax,nclass),  bxsc(n,mmax), r(mmax,mmax), dofit,
  res(nx,nclass),betasc(mmax,nclass), varsc(mmax,mmax), var(mmax,mmax), stopfac,work(1)
  
integer   tempin(mmax), bestin(mmax),qrank, qpivot(mmax)

logical forwstep,go, prune, newform, cvar, trace
common trace

#
# forw stepwise stops if gcv/gcvnull > stopfac, to avoid adding lots of
# noisy terms
#
double precision rtemp(4)
integer itemp(4)

tolbx=.01
stopfac=10.0
#stopfac=10e9
prevcrit=10e9
#write(6,*) "in marsnew mmax=",mmax, " forwstep=",forwstep, "interms=",interms

#call intpr("M1",2,n,1)
#call dblepr("penalty",7,penalty,1)
#call intpr("maxo",4,maxorder,1)

if(interms.eq.1) {dofit=0}
 else{dofit=0;do j=2,lenb {dofit=dofit+fullin(j)}
     nterms=interms
}


if(forwstep){
fullin(1)=1
do i=2,mmax{
 fullin(i)=0
}

do i=1,n {
  w(i)=1
}

do i=1, mmax{
  termlen(i)=0
  do j=1, p{
   flag(i,j)=0
   cut(i,j)=0
}}



nterms=1;  nterms2=2;
do i=1,n {
   bx(i,1)=1
  bxorth(i,1)=1.0/dsqrt(dfloat(n))
}

bxorthm(1)=1/dsqrt(dfloat(n))

do i=1,n {
 do j=1, mmax{
   bx(i,j)=0.0
}}

do i=1,n {
   bx(i,1)=1
}

do k=1, nclass{
 ybar(k)=0.0
 do i=1,n {
   ybar(k)=ybar(k)+y(i,k)/n
}}
if(interms.eq.1){
rssnull=0.0
 do k=1, nclass{ 
 do i=1,n {
   rssnull=rssnull+(y(i,k)-ybar(k))**2
 }}
}
else{
 rssnull=0.0
 do k=1, nclass{ 
 do i=1,n {
   rssnull=rssnull+res(i,k)**2
 }}
 
}
 


rss=rssnull

cmm= (1+dofit) + penalty*(.5*dofit)
gcvnull=(rssnull/n)/(1.0-cmm/n)**2

#write(6,*) "initial rss=", rssnull, " initial gcv=", gcvnull
if(trace)call dblepr("initial rss=",11,rssnull,1)
if(trace)call dblepr("initial gcv=",11,gcvnull,1)

lenb=1
ii=interms-1
go=.true.
while( (ii.lt.(mmax-1)).and.((rss/rssnull).gt.thresh).and.go){
  ii=ii+2

#  write(6,*) "bef addtrm",ii, nterms

   do i1=1, nterms{
     do i2=1, nterms{
     cov(i1,i2)=0
  }}

  do j=1, nterms{
    cov(j,j)=0.0
    do i=1,n {
      cov(j,j)=cov(j,j)+(bxorth(i,j)-bxorthm(j))*(bxorth(i,j)-bxorthm(j))
    }
  }
  


  
 do k=1,nclass{
   
   do j=1, nterms{
     covsy(j,k)=0.0
     do i=1,n {
      covsy(j,k)=covsy(j,k)+(y(i,k)-ybar(k))*bxorth(i,j)
      }
    }
  }




  do ik=1,mmax { tempin(ik)=fullin(ik)}

  call addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,rss,prevcrit,
     cov,covsy,y,ybar,x,tagx,w,termlen,mmax,tolbx,
  nterms,flag,maxorder,scr1,scr5,scr6,imax,jmax,kmax,critmax, newform,bxsc, r, betasc, temp)
 
# check whether to accept term
doftemp=dofit
 doftemp=doftemp+1
if((imax>1).and.(newform)) {doftemp=doftemp+1}




temprss=rss-critmax
cmm= (1+doftemp) + penalty*(.5*doftemp)
gcv=(temprss/n)/(1.0-cmm/n)**2

go=.false.
if(((critmax/rss).gt.thresh).and.((gcv/gcvnull).lt.stopfac)){
   go=.true.
   dofit=doftemp

    rss=rss-critmax
    kk=tagx(imax,jmax)
#write(6,256) jmax, imax ,kmax, critmax,x(kk,jmax), rss,gcv, dofit
256 format(" ","adding term"," jmax=",i3, "  imax=",i3 ,"  kmax=",i3, "  critmax= ",f8.2,
"  cutp=", f9.5," rss=",f8.2, " gcv=",f8.2, " dofit=",f9.3)
#    write(6,*) 
itemp(1)=jmax;itemp(2)=imax;itemp(3)=kmax;
rtemp(1)=critmax;rtemp(2)=x(kk,jmax);rtemp(3)=rss;rtemp(4)=gcv

if(trace) call intpr("adding term ",12,ii,1)
if(trace)call intpr("var, sp index, parent",21,itemp,3)
if(trace)call dblepr("critmax cut rss gcv",19,rtemp,4)

prevcrit=critmax
  do j=1,p {
    flag(ii,j)=flag(kmax,j)
   flag(ii+1,j)=flag(kmax,j)
   cut(ii,j)=cut(kmax,j)
   cut(ii+1,j)=cut(kmax,j)
   dir(ii,j)=dir(kmax,j)
   dir(ii+1,j)=dir(kmax,j)
   

  }
  termlen(ii)=termlen(kmax)+1
  termlen(ii+1)=termlen(kmax)+1


 


 do i=1,n { temp(i)=x(tagx(i,jmax),jmax)}
 temp1=temp(imax)

  fullin(ii)=1;

  if((imax.gt.1).and.(newform)) {fullin(ii+1)=1 }

        flag(ii,jmax)=1;
       flag(ii+1,jmax)=1;
       cut(ii,jmax)=temp1
       cut(ii+1,jmax)=temp1
        dir(ii,jmax)=1
        dir(ii+1,jmax)=-1
#
#this is to prevent trying to split a 0 column later

 if(fullin(ii+1).eq.0)
  {
#do j=1,p {flag(ii+1,j)=1}
#   
       termlen(ii+1)=maxorder+1
  }


  
  do i=1,n { if( (x(i,jmax)-temp1).gt.0) bx(i,ii)=bx(i,kmax)*(x(i,jmax)-temp1)
              if((temp1-x(i,jmax)).ge.0) bx(i,ii+1)=bx(i,kmax)*(temp1-x(i,jmax))
           }
 


 
    if(nterms.eq.1){
      temp1=0.0;  do i=1,n { temp1=temp1+bx(i,2)/n; } 
            do i=1,n {bxorth(i,2)=bx(i,2)-temp1;}
    }
    else{
##     write(6,*) "bef call to orthreg"


 call orthreg(n,n,nterms,bxorth,fullin, bx(1,ii),bxorth(1,nterms2))
          }

 if(fullin(ii+1).eq.1){


 call orthreg(n,n,nterms+1,bxorth,fullin, bx(1,ii+1),bxorth(1,nterms2+1))
#
  }
  else {do i=1,n {bxorth(i,nterms2+1)=0}}

  bxorthm(nterms2)=0.0 ; 
  bxorthm(nterms2+1)=0.0

   do i=1,n {
     bxorthm(nterms2)=bxorthm(nterms2)+bxorth(i,nterms2)/n
     bxorthm(nterms2+1)=bxorthm(nterms2+1)+bxorth(i,nterms2+1)/n
   }

 

   temp1=0.0;temp2=0.0;do i=1,n {temp1=temp1+bxorth(i,nterms2)**2
                                 temp2=temp2+bxorth(i,nterms2+1)**2 }
##write(6,*) "norm of new columns",temp1,temp2
   if (temp1.gt.0.0) {do i=1,n { bxorth(i,nterms2)  =bxorth(i,nterms2)/dsqrt(temp1) }}
   if (temp2.gt.0.0) {do i=1,n { bxorth(i,nterms2+1)=bxorth(i,nterms2+1)/dsqrt(temp2) }}
   lenb=lenb+2
  nterms=nterms+2; nterms2=nterms2+2;

}}

#write(6,*) "stopping forw stepwise"
#if((rss/rssnull).le.thresh) write(6,*) "rss ratio=",rss/rssnull
#if((critmax/rss).le.thresh) write(6,*) "crit ratio=",critmax/rss
#if((gcv/gcvnull).gt.stopfac) write(6,*) "gcv ratio=",gcv/gcvnull

rtemp(1)=rss/rssnull;rtemp(2)=critmax/rss;rtemp(3)=gcv/gcvnull
if(trace)call dblepr("stopping forw step; rss crit and gcv ratios",43,rtemp,3)
if(trace){
 if((rss/rssnull).le.thresh) call dblepr("rss ratio=",10,rss/rssnull,1)
 if((critmax/rss).le.thresh) call dblepr ("crit ratio=",11,critmax/rss,1)
call dblepr("critmax",7,critmax,1)
call dblepr("rss",3,rss,1)
 if((gcv/gcvnull).gt.stopfac) call dblepr("gcv ratio=",10,gcv/gcvnull,1)
}

# write(6,*) "after forward step  ","go=",go," rss=",rss, "nterms=", nterms


}


#call intpr("M2",2,fullin,10)
#call intpr("nterms",6,nterms,1)
dofit= -1
do i=1,nterms{
 bestin(i)=fullin(i)
 dofit=dofit+fullin(i)
}
#call intpr("M21",3,n,1)


if(trace)call intpr("aft forw step",13,nterms,1)

##added by Trevor 5/24/2015
cvar=.false.
call  qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,bestin,y,qpivot,qrank,beta,res,rss,cvar,var,varsc,scr1, work) 


#call dblepr("rss",3,rss,1)
#call dblepr("beta",4,beta,4)
#call intpr("M3",2,n,1)

nt=dofit+1
if(qrank< nt) {
#write(6,*) "singular matrix"
 do i=qrank+1,nt{ 
  bestin(qpivot(i))=0
  fullin(qpivot(i))=0
#write(6,*) "removing col ",qpivot(i)
  dofit=dofit-1;
} }


##write(6,*) "bef mlin", nterms,dofit,(bestin(j),j=1,nterms),y(1,1),n,mmax,nclass,w(1)

cvar=.true.
#call intpr("M23",3,fullin,11)
#call intpr("bestin",6,bestin,11)
#
#call intpr("lenb",4,lenb,1)
#call intpr("mmax",4,mmax,1)
#call dblepr("res",3,res,10)
#call dblepr("bx",2,bx(1,1),5)
#call dblepr("bx",2,bx(1,2),5)
#call dblepr("bx",2,bx(n,11),1)
#call dblepr("y",1,y,5)
#call intpr("nx",2,nx,1)
#call intpr("n",1,n,1)




#write(6,*) "rank=",qrank, "rss=",rss


 





rssfull=rss

# dofit is the number of nonconstant, indep basis fns in the model



 cmm= (1+dofit) + penalty*(.5*dofit)
bestgcv=(rss/n)/(1.0-cmm/n)**2

#write(6,*) "full model,gcv=",bestgcv, " rss=",rssfull, "dofit=",dofit
rtemp(1)=bestgcv; rtemp(2)=rssfull;rtemp(3)=dofit
if(trace)call dblepr("full model: gcv rss dofit",25,rtemp,3)
if(trace)call intpr("terms",5,fullin,lenb)

# after forward stepwise, nterms and lenb = # of columns of bx
# dofit is number of nonzero columns 



if(prune){
 
do i=1,mmax{ tempin(i)=bestin(i)}
#write(6,*) "initial tempin", (tempin(i),i=1,nterms)

#call intpr("M5",2,n,1)
while(dofit>0 ) {
    
     
    jo=1;rsstemp=10e99;minterm=0
     do ii=2, lenb {

       if(tempin(ii).eq.1) {
        jo=jo+1
          temp7=0.0
          do kc=1,nclass{
           temp7=temp7+beta(jo,kc)**2/var(jo,jo)
          }

        if(temp7 < rsstemp) {minterm=ii; rsstemp=temp7}
      


      }}
         rss=rss+rsstemp
        dofit=dofit-1; 
        cmm= (1.0+dofit) + penalty*(.5*dofit)
         gcv=(rss/n)/(1.0-cmm/n)**2
        
         tempin(minterm)=0
      

#         write(6,100) minterm,gcv, rss,dofit,(tempin(ik),ik=1,lenb)
100 format(" ","pruning, minterm= ",i4, " gcv=",f9.3,2x, " rss=",f9.3,2x," dof=",f9.3,
" model= ",60(i1,1x))
   
 
    if(gcv< bestgcv) {bestgcv=gcv;do i=1,mmax {bestin(i)=tempin(i);}}


  if(dofit > 0) {  
cvar=.true.
call  qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,tempin,y,qpivot,qrank,beta,res,rss,cvar,var,varsc,scr1,work)


          }
}



call  qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,bestin,y,qpivot,qrank,beta,res,rss,cvar,var,varsc,
scr1, work)
#write(6,101) bestgcv,rss,(bestin(ik),ik=1,lenb)
101 format(" ","best model gcv=",f9.3," rss=",f9.3,2x,"model= ",60(i1,1x))
if(trace)call intpr("best model",10,bestin,lenb)
if(trace)call dblepr(" gcv=",4,bestgcv,1)


}





return
end
subroutine addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,
rss,prevcrit,cov,covsy,y,ybar,x,tagx,w,termlen,mmax,tolbx,
nterms,flag,maxorder,scr1,scr5,scr6,imax,jmax,kmax,critmax, newform,bxsc,r,betasc, scrat)


implicit double precision (a-h,o-z)

integer n,nterms,nterms2,p,mmax,flag(mmax,p),v,tagx(nx,p),termlen(mmax), nclass, tempin(mmax), minspan, iendspan

double precision cov(mmax,mmax),covsy(mmax,nclass),
critmax,x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),y(n,nclass),ybar(nclass),
scr1(mmax),scr5(mmax),scr6(nclass), bxsc(n,mmax), r(mmax,mmax),
 betasc(mmax,nclass), scrat(n),w(n)

double precision temp1, temp2, scr2,sumb, sumbx, su, st, tem
logical newform, tnewform, trace
common trace
critmax=0;jmax=0;imax=0;kmax=0;


# finds best term to add to current mars model







##write(6,*) "in addtrm n,p, nclass, nterms,mmax, y(1,1), x(1,1), tagx(1,1),flag(1,1),bx(1,1#),bxorth(1,1)", n,p, nclass, nterms,mmax, y(1,1), x(1,1), tagx(1,1),flag(1,1),bx(1,1),bxor#th(1,1)





do m=1,nterms {

 nm=0
 do jjj=1,n {
  if(bx(jjj,m).gt.0) {nm=nm+1}
 }

# tem=-(1d0/(n*nm))*dlog(1d0 - 5d-2)  n should be p thanks to  Gints Jekabsons/Stephen Milborrow
 tem=-(1d0/(p*nm))*dlog(1d0 - 5d-2)

 minspan= -1d0*(dlog(tem)/dlog(2d0))/2.5

# tem=(5d-2)/n   n should be p thanks to  Gints Jekabsons/Stephen Milborrow
 tem=(5d-2)/p
 iendspan=3d0-dlog(tem)/dlog(2d0)

# # write(6,*) "minspan,iendspan",minspan,iendspan

if(termlen(m)< maxorder){

 do v=1,p {

  if(flag(m,v).eq.0){
# check if model term type is already in model, to avoid a linear dependence


  tnewform=.true.
  mm=1
  while((mm.le.nterms).and.tnewform) {
     mm=mm+1
    if(tempin(mm).eq.1){
     tnewform=.false.
     if(flag(mm,v).ne.1){tnewform=.true.;go to 9911}
     do j=1,p {if(j.ne.v) { if(flag(mm,j).ne.flag(m,j)) {tnewform=.true.;go to 9911 }}}
    }
9911 continue
}

# if new form of term, fit term bm*x
# should it be bxorth below?

if(tnewform) {

nterms2=nterms+1

     do i=1,n {
        scrat(i)=x(i,v)*bx(i,m)
      }
#
#

if(nterms>1){

 call orthreg(n,n,nterms,bxorth,tempin, scrat,bxorth(1,nterms2))
   }
else{
   tem=0;do i=1,n {tem=tem+scrat(i)/n;}
    do i=1,n{ bxorth(i,2)=scrat(i)-tem}
   }


    bxorthm(nterms2)=0.0 ; 
   do i=1,n {
     bxorthm(nterms2)=bxorthm(nterms2)+bxorth(i,nterms2)/n 
   }
   temp1=0.0;do i=1,n {temp1=temp1+bxorth(i,nterms2)**2}                               
   if (temp1.gt.tolbx) {do i=1,n { bxorth(i,nterms2)=bxorth(i,nterms2)/dsqrt(temp1) }}
     else {do i=1,n {bxorth(i,nterms2)=0};tnewform=.false.;}

    do i1=1, nterms2{
         cov(i1,nterms2)=0.0
         cov(nterms2, i1)=0.0
    }

 
      cov(nterms2,nterms2)=1


  do kc=1,nclass{
     covsy(nterms2,kc)=0.0
     do i=1,n {
      covsy(nterms2,kc)=covsy(nterms2,kc)+(y(i,kc)-ybar(kc))*bxorth(i,nterms2)

      }
   }


    critnew=0.0
    do kc=1,nclass {
      temp1=0
      do i=1,n { temp1=temp1+y(i,kc)*bxorth(i,nterms2)}
      critnew=critnew+temp1**2
     }
## write(6,*) "geoff",nterms,v,critnew
    if(critnew.gt.critmax) {
       jmax=v
       critmax=critnew
       imax=1
       kmax=m
      }

}


if(tnewform) {nterms2=nterms+1; nterms21=nterms+2} else{nterms2=nterms; nterms21=nterms+1;critnew=0.0}

#if((nterms.eq.7).and.(v.eq.2)) {
# call dblepr("bxorth",6,bxorth(1,nterms2),10)
# call dblepr("temp1",5,temp1,1)
# call dblepr("critmax",7,critmax,1)
 
#}





# try other knot locations





     do kc=1, nclass{
      covsy(nterms21,kc)=0
     }
     do ii=1,nterms21 {
        cov(ii,nterms21)=0
        cov(nterms21,ii)=0
     }
   do kc=1,nclass {
     scr6(kc)=0
   }
    do ii=1,nterms21{
        scr1(ii)=0;
    }
scr2=0;su=0;st=0;sumbx2=0;sumb=0.0;sumbx=0.0;
   for (k=n-1;k>0;k=k-1) {
 
   do i=1,nterms2 {
       kk=tagx(k,v)
       kk1=tagx(k+1,v)
       scr1(i)=scr1(i)+(bxorth(kk1,i)-bxorthm(i))*bx(kk1,m)
       cov(i,nterms21)=cov(i,nterms21)+ (x(kk1,v)-x(kk,v))*scr1(i)
       cov(nterms21,i)=cov(i,nterms21)   
      }
     scr2=scr2+(bx(kk1,m)**2)*x(kk1,v)

     sumbx2=sumbx2+bx(kk1,m)**2
    sumb=sumb+bx(kk1,m)
    sumbx=sumbx+bx(kk1,m)*x(kk1,v)
     su=st
    st=sumbx-sumb*x(kk,v)
     cov(nterms21,nterms21)= cov(nterms21,nterms21)+ 
      (x(kk1,v)-x(kk,v))*(2*scr2-sumbx2*(x(kk,v)+x(kk1,v)))+ 
          ( (su*su)-(st*st) )/n


   crittemp=critnew
     do kc=1, nclass{

     scr6(kc)=scr6(kc)+(y(kk1,kc)-ybar(kc))*bx(kk1,m)
     covsy(nterms21,kc)=covsy(nterms21,kc )+(x(kk1,v)-x(kk,v))*scr6(kc)

    temp1=covsy(nterms21,kc)
    temp2=cov(nterms21,nterms21)

     do jk=1,nterms2 {
        temp1=temp1-covsy(jk,kc)*cov(jk,nterms21)
         temp2=temp2-cov(jk,nterms21)*cov(jk,nterms21)
       }
#
# this has to be fixed!!
#



 if(cov(nterms21,nterms21)>0){
        if((temp2/cov(nterms21,nterms21)) > tolbx) {critadd=(temp1*temp1)/temp2} else {critadd=0.0}}
 else{critadd=0}

 crittemp=crittemp+critadd

#if((nterms.eq.7).and.(v.eq.2).and.(critadd.gt.0)) {

#call intpr("k",1,k,1)
 #call dblepr("temp2",5,temp2,1)
#call dblepr("cov",3,cov(nterms21,nterms21),1)
# call dblepr("critadd",7,critadd,1)
# call dblepr("crittemp",8,crittemp,1)

#}



 
 if(crittemp.gt.(1.01*rss)) {crittemp=0.0}
 if(crittemp.gt.(2*prevcrit)) {crittemp=0.0}

 }
         
   


if(k.gt.1){
  k0=tagx(k-1,v)
}
#
# decide whether to accept term
#

#if((k.eq.70).and.(v.eq.10)){
##write(6,*) "HI",k,v,critmax,minspan,mod(k,minspan),iendspan,n-iendspan,x(kk,v),x(kk0,v)
#}
#call intpr("k",1,k,1)
#call intpr("v",1,v,1)
#call dblepr("crittemp",8,crittemp,1)
#call intpr("minspan",7,minspan,1)
#call intpr("iendspan",8,iendspan,1)
     if((crittemp.gt.critmax).and.(mod(k,minspan).eq.0).and.
(k.ge.iendspan).and.(k.le.(n-iendspan)).and.(bx(kk1,m).gt.0).and.
(.not.((k.gt.1).and.(x(kk,v).eq.x(k0,v))) )) {
       jmax=v
       critmax=crittemp
       imax=k
       kmax=m
       newform=tnewform

       
      }
   


   } }
9999 continue
 }}


}



return
end

subroutine sknotl(x,n,knot,k)
implicit double precision(a-h,o-z) 
integer     n,k,ndk,j
double precision x(n),knot(n+6),a1,a2,a3,a4


     # Allocate knots acording to the cutoffs given below


	# Cutoff constants

	 a1 = log(50e0)/log(2e0)  ; a2 = log(100e0)/log(2e0)
	 a3 = log(140e0)/log(2e0) ; a4 = log(200e0)/log(2e0)

	# Cutoff Criteria

        if(n<50)                    { ndk = n }
        else if (n>=50  & n<200)    { ndk = 2.**(a1+(a2-a1)*(n-50.)/150.)  }
        else if (n>=200 & n<800)    { ndk = 2.**(a2+(a3-a2)*(n-200.)/600.)  }
        else if (n>=800 & n<3200)   { ndk = 2.**(a3+(a4-a3)*(n-800.)/2400.)  }
        else if (n>=3200)           { ndk = 200. + (n-3200)**.2 }
		 
		 k = ndk + 6

     
     # Allocate Knots  ( note no account is taken of any weighting vector )

	do j=1,3   { knot(j) = x(1) }
	do j=1,ndk { knot(j+3) = x( 1 + (j-1)*(n-1)/(ndk-1) ) }
	do j=1,3   { knot(ndk+3+j) = x(n) }

return
end

subroutine pack(n,p,match,x,xbar)
integer n,p,match(n)
double precision x(n),xbar(n)
	do i=1,p
		xbar(i)=0e0
	do i=1,n
		xbar(match(i))=xbar(match(i))+x(i)
return
end

subroutine suff(n,p,match,x,y,w,xbar,ybar,wbar,work)
integer n,p,match(n)
double precision x(n),xbar(n),y(n),ybar(n),w(n),wbar(n),work(n)
call pack(n,p,match,w,wbar)
do i=1,n
	xbar(match(i))=x(i)
do i=1,n
	work(i)=y(i)*w(i)
call pack(n,p,match,work,ybar)
	do i=1,p{
		if(wbar(i)>0e0) 
			ybar(i)=ybar(i)/wbar(i) 
		else ybar(i)=0e0

	}
return
end
subroutine unpack(n,p,match,xbar,x)
integer n,p, match(n)
double precision x(n),xbar(p+1)
if(p<n)xbar(p+1)=0e0
do i = 1,n
	x(i)=xbar(match(i))
return
end

subroutine qrreg(nx,n,px,p,nclass,x,xsc,in,y,qpivot,qrank,beta,res,rss,cvar,var,varsc,scr1,work)

implicit double precision (a-h,o-z)

integer nx,n,p,px, qpivot(p),qrank,nclass,in(p)
double precision x(nx,p), xsc(n,p), y(n,nclass),res(nx,nclass),beta(px,nclass),work(1),scr1(p),var(px,p),varsc(px,p)
logical cvar

ii=0
do j=1,p {
 if(in(j).eq.1){
  ii=ii+1
  do i=1,n {
  xsc(i,ii)=x(i,j)
   }
}}
nt=ii
ijob=101
info=1
temp3=1d-2
do i=1,p {qpivot(i)=i}
call dqrdca(xsc,n,n,nt,scr1,qpivot,work,qrank,temp3)
# computes both fits and beta
rss=0.0
 do k=1,nclass{
	call dqrsl(xsc,n,n,qrank,scr1,y(1,k),work(1),work(1),beta(1,k),
		work(1),res(1,k),ijob,info)
        do i=1,n { res(i,k)=y(i,k)-res(i,k); rss=rss+res(i,k)*res(i,k)}
 }

if(cvar) {call calcvar(nx,n,px,xsc,qrank,qpivot,var,varsc,work)}
return
end

subroutine sspl(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,
		dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs,sg,abd,p1ip,ier)
implicit double precision(a-h,o-z) 

# A Cubic B-spline Smoothing routine.
#
#          The algorithm minimises:
#
#      (1/n) * sum w(i)* (y(i)-s(i))**2 + lambda* int ( s"(x) )**2 dx
#
#	for each of p response variables in y
#Input

#  x(n)		vector containing the ordinates of the observations
#  y(ldy,p)	matrix (n x p) of responses (ldy can be greater than n)
#  w(n)		vector containing the weights given to each data point
#  n    	number of data points
#  ldy		leading dimension of y
#  p		number of columns in y	
#  knot(nk+4)	vector of knot points defining the cubic b-spline basis.
#  nk		number of b-spline coefficients to be estimated
#			nk <= n+2
#  method 	method for selecting amount of smoothing, lambda
#		1 = fixed lambda 
#		2 = fixed df
#		3 = gcv
#		4 = cv
#  tol		used in Golden Search routine
#  wp(p)	weights, length p,  used to combine cv or gcv in 3 or 4 above
#  ssy(p)	offsets for weighted sum of squares for y; can be all zero,
#			else should be the variability lost due to collapsing
#			onto unique values
#  dfoff	offset df used in gcv calculations (0 is good default)
#  dfmax	maximum value for df allowed when gcv or cv are used
#  		routine simply returns the value at dfmax if it was exceeded
#  cost		cost per df (1 is good default)
#Input/Output
#  lambda	penalised likelihood smoothing parameter
#  df		trace(S)
#Output
#  cv		omnibus cv criterion
#  gcv		omnibus gcv criterion (including penalty and offset)
#  coef(nk,p)	vector of spline coefficients
#  s(ldy,p)	matrix of smoothed y-values
#  lev(n)	vector of leverages
# Working arrays/matrix
#  xwy(nk,p)	X'Wy
#  hs(nk,4)	the diagonals of the X'WX matrix
#  sg(nk,4)   	the diagonals of the Gram matrix
#  abd(4,nk)	[ X'WX+lambda*SIGMA] in diagonal form
#  p1ip(4,nk)	inner products between columns of L inverse
#  ier          error indicator
#                  ier = 0 ___  everything fine
#                  ier = 1 ___  spar too small or too big
#                               problem in cholesky decomposition
double precision x(n),y(ldy,p),w(n),knot(nk+4),tol,wp(p),ssy(p),
	 	dfoff,dfmax,cost,lambda,df,cv,gcv,coef(nk,p),s(ldy,p),lev(n),
		xwy(nk,p),hs(nk,4),sg(nk,4),abd(4,nk),p1ip(4,nk)
integer         n,p,ldy,nk,method,ier

#  Compute SIGMA, X'WX, X'WY, trace, ratio, s0, s1.

# SIGMA-> sg[]
# X'WX -> hs[]
# X'WY -> xwy[]
call sgram(sg(1,1),sg(1,2),sg(1,3),sg(1,4),knot,nk)
call stxwx2(x,y,w,n,ldy,p,knot,nk,xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,4))

# Compute estimate

if(method==1) {# Value of lambda supplied

	call sslvr2(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,4),
		sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
	}

else {# Use Forsythe, Malcom and Moler routine to minimise criterion
	call fmm(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs,sg,abd,p1ip,ier)
	if(method>2&df>dfmax){
		df=dfmax
		call fmm(x,y,w,n,ldy,p,knot,nk,2,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs,sg,abd,p1ip,ier)
		}
	}
return
end

subroutine fmm(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs,sg,abd,p1ip,ier)
double precision xs(n),ys(ldy,nvar),ws(n),knot(nk+4),tol,wp(nvar),ssy(nvar),
	 	dfoff,cost,lambda,df,cv,gcv,coef(nk,nvar),s(ldy,nvar),lev(n),
		xwy(nk,nvar),hs(nk,4),sg(nk,4),abd(4,nk),p1ip(4,nk)
integer         n,ldy,nvar,nk,method,ier

# Local variables

double precision  t1,t2,ratio,
                  a,b,c,d,e,eps,xm,p,q,r,tol1,tol2,u,v,w,
                  fu,fv,fw,fx,x,targdf,
		  ax,bx

integer i
ax=1d-10  	#used to be lspar
bx=1.5		#used to be uspar
t1=0. ; t2=0.
targdf=df
do i=3,nk-3 { t1 = t1 + hs(i,1) }
do i=3,nk-3 { t2 = t2 + sg(i,1) }
ratio = t1/t2
#
#  an approximation  x  to the point where f  attains a minimum  on
#  the interval  (ax,bx)  is determined.
#
#
#  input..
#
#  ax	 left endpoint of initial interval
#  bx	 right endpoint of initial interval
#  f	 function subprogram which evaluates  f(x)  for any  x
#	 in the interval  (ax,bx)
#  tol	 desired length of the interval of uncertainty of the final
#	 result ( .ge. 0.0)
#
#
#  output..
#
#  fmin  abcissa approximating the point where	f  attains a minimum
#
#
#      the method used is a combination of  golden  section  search  and
#  successive parabolic interpolation.	convergence is never much slower
#  than  that  for  a  fibonacci search.  if  f  has a continuous second
#  derivative which is positive at the minimum (which is not  at  ax  or
#  bx),  then  convergence  is	superlinear, and usually of the order of
#  about  1.324....
#      the function  f	is never evaluated at two points closer together
#  than  eps*dabs(fmin) + (tol/3), where eps is	approximately the square
#  root  of  the  relative  machine  precision.   if   f   is a unimodal
#  function and the computed values of	 f   are  always  unimodal  when
#  separated by at least  eps*dabs(x) + (tol/3), then  fmin  approximates
#  the abcissa of the global minimum of  f  on the interval  ax,bx  with
#  an error less than  3*eps*dabs(fmin) + tol.  if   f	is not unimodal,
#  then fmin may approximate a local, but perhaps non-global, minimum to
#  the same accuracy.
#      this function subprogram is a slightly modified	version  of  the
#  algol  60 procedure	localmin  given in richard brent, algorithms for
#  minimization without derivatives, prentice - hall, inc. (1973).
#
#
#      double precision  a,b,c,d,e,eps,xm,p,q,r,tol1,tol2,u,v,w
#      double precision  fu,fv,fw,fx,x
#
#  c is the squared inverse of the golden ratio
#
      c = 0.5*(3. - dsqrt(5d0))
#
#  eps is approximately the square root of the relative machine
#  precision.
#
      eps = 1d0
   10 eps = eps/2d0
      tol1 = 1d0 + eps
      if (tol1 .gt. 1d0) go to 10
      eps = dsqrt(eps)
#
#  initialization
#
      a = ax
      b = bx
      v = a + c*(b - a)
      w = v
      x = v
      e = 0.0
	lambda = ratio*16.**(-2. + x*(6.))
	call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,4),
		sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
switch(method){
	case 2:
		fx=3d0+(targdf-df)**2
	case 3:
		fx=gcv
	case 4:
		fx=cv
	}
      fv = fx
      fw = fx
#
#  main loop starts here
#
   20 xm = 0.5*(a + b)
      tol1 = eps*dabs(x) + tol/3d0
      tol2 = 2d0*tol1
#
#  check stopping criterion
#
      if (dabs(x - xm) .le. (tol2 - 0.5*(b - a))) go to 90
#
# is golden-section necessary
#
      if (dabs(e) .le. tol1) go to 40
#
#  fit parabola
#
      r = (x - w)*(fx - fv)
      q = (x - v)*(fx - fw)
      p = (x - v)*q - (x - w)*r
      q = 2.00*(q - r)
      if (q .gt. 0.0) p = -p
      q =  dabs(q)
      r = e
      e = d
#
#  is parabola acceptable
#
   30 if (dabs(p) .ge. dabs(0.5*q*r)) go to 40
      if (p .le. q*(a - x)) go to 40
      if (p .ge. q*(b - x)) go to 40
#
#  a parabolic interpolation step
#
      d = p/q
      u = x + d
#
#  f must not be evaluated too close to ax or bx
#
      if ((u - a) .lt. tol2) d = dsign(tol1, xm - x)
      if ((b - u) .lt. tol2) d = dsign(tol1, xm - x)
      go to 50
#
#  a golden-section step
#
   40 if (x .ge. xm) e = a - x
      if (x .lt. xm) e = b - x
      d = c*e
#
#  f must not be evaluated too close to x
#
   50 if (dabs(d) .ge. tol1) u = x + d
      if (dabs(d) .lt. tol1) u = x + dsign(tol1, d)

		lambda = ratio*16.**(-2. + u*(6.))
	call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
		xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,4),
		sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
switch(method){
	case 2:
		fu=3d0+(targdf-df)**2
	case 3:
		fu=gcv
	case 4:
		fu=cv
	}
#
#  update  a, b, v, w, and x
#
      if (fu .gt. fx) go to 60
      if (u .ge. x) a = x
      if (u .lt. x) b = x
      v = w
      fv = fw
      w = x
      fw = fx
      x = u
      fx = fu
      go to 20
   60 if (u .lt. x) a = u
      if (u .ge. x) b = u
      if (fu .le. fw) go to 70
      if (w .eq. x) go to 70
      if (fu .le. fv) go to 80
      if (v .eq. x) go to 80
      if (v .eq. w) go to 80
      go to 20
   70 v = w
      fv = fw
      w = u
      fw = fu
      go to 20
   80 v = u
      fv = fu
      go to 20
#
#  end of main loop
#
   90 continue 
if(method==2){call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,1,tol,wp,ssy,
	dfoff,cost,lambda,df,cv,gcv,coef,s,lev,
	xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,4),
	sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)}
	
return 
end
subroutine stxwx2(x,z,w,k,ldy,pz,xknot,n,y,hs0,hs1,hs2,hs3)
implicit double precision(a-h,o-z) 
double precision  z(ldy,pz),w(k),x(k),xknot(n+4),
		 y(n,pz),hs0(n),hs1(n),hs2(n),hs3(n),
		 eps,vnikx(4,1),work(16) # local
integer  k,n,pz,ldy,
j,i,pp,ileft,mflag  # local

# Initialise the output vectors

do i=1,n { 
	hs0(i)=0d0  
	hs1(i)=0d0
	hs2(i)=0d0
	hs3(i)=0d0
	do j=1,pz { y(i,j)=0d0 }
	}

# Compute X'WX -> hs0,hs1,hs2,hs3  and X'WZ -> y

eps = .1d-9
do i=1,k {
	call interv(xknot(1),(n+1),x(i),ileft,mflag)
	if(mflag== 1) {
		if(x(i)<=(xknot(ileft)+eps)){ileft=ileft-1}
		else{return}
	}
	call bsplvd (xknot,4,x(i),ileft,work,vnikx,1)

	j= ileft-4+1
	do pp=1,pz {y(j,pp) = y(j,pp)+w(i)*z(i,pp)*vnikx(1,1)}
	hs0(j)=hs0(j)+w(i)*vnikx(1,1)**2
	hs1(j)=hs1(j)+w(i)*vnikx(1,1)*vnikx(2,1)
	hs2(j)=hs2(j)+w(i)*vnikx(1,1)*vnikx(3,1)
	hs3(j)=hs3(j)+w(i)*vnikx(1,1)*vnikx(4,1)

	j= ileft-4+2
	do pp=1,pz {y(j,pp) =  y(j,pp)+w(i)*z(i,pp)*vnikx(2,1)}
	hs0(j)=hs0(j)+w(i)*vnikx(2,1)**2
	hs1(j)=hs1(j)+w(i)*vnikx(2,1)*vnikx(3,1)
	hs2(j)=hs2(j)+w(i)*vnikx(2,1)*vnikx(4,1)

	j= ileft-4+3
	do pp=1,pz {y(j,pp) =  y(j,pp)+w(i)*z(i,pp)*vnikx(3,1)}
	hs0(j)=hs0(j)+w(i)*vnikx(3,1)**2
	hs1(j)=hs1(j)+w(i)*vnikx(3,1)*vnikx(4,1)

	j= ileft-4+4
	do pp=1,pz {y(j,pp) =  y(j,pp)+w(i)*z(i,pp)*vnikx(4,1)}
	hs0(j)=hs0(j)+w(i)*vnikx(4,1)**2  
	}

return
end
subroutine sslvr2(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,
		dfoff,cost,lambda,df,cv,gcv,coef,sz,lev,
		xwy,hs0,hs1,hs2,hs3,sg0,sg1,sg2,sg3,
		abd,p1ip,info)

implicit double precision(a-h,o-z) 
double precision x(n),y(ldy,p),w(n),knot(nk+4),tol,wp(p),ssy(p),
	 	dfoff,cost,lambda,df,cv,gcv,coef(nk,p),sz(ldy,p),lev(n),
		xwy(nk,p),
		hs0(nk),hs1(nk),hs2(nk),hs3(nk),
		sg0(nk),sg1(nk),sg2(nk),sg3(nk),
		abd(4,nk),p1ip(4,nk)
integer         n,p,ldy,nk,method,info
#local storage
double precision b0,b1,b2,b3,eps,vnikx(4,1),work(16),
	 xv,bvalue,rss,tssy
integer  ld4,i,icoef,ileft,ilo,j,mflag
logical fittoo
fittoo= (method!=2)

         ilo = 1 ; eps = .1d-10 ; ld4=4

# Purpose : Solves the smoothing problem and computes the
#           criterion functions (CV and GCV).
# The coeficients of estimated smooth

if(fittoo){
	do i=1,nk { do j=1,p {coef(i,j) = xwy(i,j) } }
	}
do i=1,nk { abd(4,i)   = hs0(i)+lambda*sg0(i) }
do i=1,(nk-1) { abd(3,i+1) = hs1(i)+lambda*sg1(i) }
do i=1,(nk-2) { abd(2,i+2) = hs2(i)+lambda*sg2(i) }
do i=1,(nk-3) { abd(1,i+3) = hs3(i)+lambda*sg3(i) }

call dpbfa(abd,ld4,nk,3,info)
if(info.ne.0) { return } 
if(fittoo){
	do j=1,p{ call dpbsl(abd,ld4,nk,3,coef(1,j)) }

# Value of smooths at the data points

	icoef = 1
	do i=1,n { 
		xv = x(i)
		do j=1,p{ sz(i,j) = bvalue(knot,coef(1,j),nk,4,xv,0) }
		}
	}
# Compute the criterion functions

# Get Leverages First

call sinrp2(abd,ld4,nk,p1ip)

do i=1,n { 
	xv = x(i)
	call interv(knot(1),(nk+1),xv,ileft,mflag)
	if(mflag==-1) { ileft = 4   ; xv = knot(4)+eps }
	if(mflag==1)  { ileft = nk  ; xv = knot(nk+1)-eps }
	j=ileft-3
	call bsplvd(knot,4,xv,ileft,work,vnikx,1)
	b0=vnikx(1,1);b1=vnikx(2,1);b2=vnikx(3,1);b3=vnikx(4,1)
	lev(i) = (p1ip(4,j)*b0**2   + 2.*p1ip(3,j)*b0*b1 +
				   2.*p1ip(2,j)*b0*b2 + 2.*p1ip(1,j)*b0*b3 +
		  p1ip(4,j+1)*b1**2 + 2.*p1ip(3,j+1)*b1*b2 +
				   2.*p1ip(2,j+1)*b1*b3 +
		  p1ip(4,j+2)*b2**2 + 2.*p1ip(3,j+2)*b2*b3 +
		  p1ip(4,j+3)*b3**2 )*w(i)    
	}
# Evaluate Criteria
rss = 0d0 ; df = 0d0 ; sumw=0d0;gcv=0d0;cv=0d0;
do i=1,n { 
	df  = df  + lev(i)
	}
if(fittoo){
	do i=1,n { 
		sumw  = sumw  + w(i)
		do j=1,p{
			rss = rss + w(i)*wp(j)*(y(i,j)-sz(i,j))**2
			cv = cv +w(i)*wp(j)*((y(i,j)-sz(i,j))/(1-lev(i)))**2
			} 
		}
tssy=0d0
do j=1,p{tssy=tssy+wp(j)*ssy(j)}

gcv=((rss+tssy)/sumw)/((1d0-((dfoff+df-1)*cost+1)/sumw)**2)

#note: the weights should sum to n (the number of original observations)

cv=(cv+tssy)/sumw
	}
#lev includes the weights
#Note that this version of cv omits ALL observations at
#tied x values, since the data are already collapsed here
return 
end
subroutine sinrp2(abd,ld4,nk,p1ip)
implicit double precision(a-h,o-z) 
double precision abd(ld4,nk),p1ip(ld4,nk),
	wjm3(3),wjm2(2),wjm1(1),c0,c1,c2,c3

integer	ld4,nk,i,j

	# Purpose :  Computes Inner Products between columns of L(-1)
	#	     where L = abd is a Banded Matrix with 3 subdiagonals

	#		A refinement of Elden's trick is used.
	#	Coded by Finbarr O'Sullivan
wjm3(1)=0d0; wjm3(2)=0d0; wjm3(1)=0d0
wjm2(1)=0d0; wjm2(2)=0d0
wjm1(1)=0d0

do i=1,nk { 
	j=nk-i+1
	c0 = 1d0/abd(4,j)
	if(j<=nk-3) {
		c1 = abd(1,j+3)*c0
		c2 = abd(2,j+2)*c0
		c3 = abd(3,j+1)*c0 
		}
	else if(j==nk-2) {
		c1 = 0d0
		c2 = abd(2,j+2)*c0
		c3 = abd(3,j+1)*c0 
		}
	else if(j==nk-1) {
		c1 = 0d0
		c2 = 0d0
		c3 = abd(3,j+1)*c0 
		}
	else if(j==nk) {
		c1 = 0d0
		c2 = 0d0
		c3 = 0d0
	}
	p1ip(1,j) = 0d0- (c1*wjm3(1)+c2*wjm3(2)+c3*wjm3(3))
	p1ip(2,j) = 0d0- (c1*wjm3(2)+c2*wjm2(1)+c3*wjm2(2))
	p1ip(3,j) = 0d0- (c1*wjm3(3)+c2*wjm2(2)+c3*wjm1(1))

	p1ip(4,j) = c0**2 + 
		c1**2*wjm3(1)+2.*c1*c2*wjm3(2)+2.*c1*c3*wjm3(3) +
		c2**2*wjm2(1)+2.*c2*c3*wjm2(2) +
		c3**2*wjm1(1)

	wjm3(1)=wjm2(1) ; wjm3(2)=wjm2(2) ; wjm3(3)=p1ip(2,j)
	wjm2(1)=wjm1(1) ; wjm2(2)=p1ip(3,j); wjm1(1)=p1ip(4,j)
	}
return
end
subroutine suff2(n,p,ny,match,y,w,ybar,wbar,ssy,work)
integer match(n),n,ny,p,i,j
double precision y(n,ny),ybar(p+1,ny),w(n),wbar(p+1),ssy(ny),work(n)
double precision tsum
#ssy is the within response variability that is lost by collapsing
call pack(n,p,match,w,wbar)
do j=1,ny{
	do i=1,n
		work(i)=y(i,j)*w(i)
	call pack(n,p,match,work,ybar(1,j))
	do i=1,p{
		if(wbar(i)>0d0) 
			ybar(i,j)=ybar(i,j)/wbar(i) 
		else ybar(i,j)=0d0
		}
	call unpack(n,p,match,ybar(1,j),work)
	tsum=0d0
	do i=1,n
		tsum=tsum+ w(i)*(y(i,j)-work(i))**2
	ssy(j)=tsum
	}
return
end
subroutine sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,center,
		dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
		xrange,work,ier)
double precision x(n),y(n,p),w(n),knot(nk+4),tol,wp(p),
	 	dfoff,dfmax,cost,lambda,df,cv,gcv,coef(1),s(n,p),lev(nef),
		xrange(2),work(1)
#workspace must be (2*p+2)*nefp1 + (p+16)*nk + n +p
integer         n,p,nk,method,ier,nef, nefp1, n2,match(1)
logical center
double precision  xmiss,sigtol,xdiff,temp
if(nef==0){# match has not been initialized
	xmiss=1d20
	sigtol=1d-5
	call namat(x,match,n,nef,work,work(n+1),xmiss,sigtol)
	xrange(1)=work(1) #work is actually the sorted unique xs
	xrange(2)=work(nef)
	}
else{
	do i=1,n {work(match(i))=x(i)}
	}
xdiff=xrange(2)-xrange(1)
do i=1,nef {work(i)=(work(i)-xrange(1))/xdiff}
if(nk==0){
	call sknotl(work,nef,knot,nk)
	nk=nk-4
	}
if(dfmax > dble(nk))dfmax=dble(nk)
if(cost>0){
	temp=dble(n-dble(center))/cost - dfoff
	if(dfmax>temp)dfmax=temp
	}
nefp1=nef+1
n2=nefp1*(2*p+2)+1
call sspl1(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,center,
	dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
	work(1),work(nefp1+1), 		#xin,yin
	work(nefp1*(p+1)+1),work(nefp1*(p+2)+1), 	#win, sout
	work(n2),work(n2+p*nk),work(n2+(p+4)*nk), 	#xwy, hs,sg
	work(n2+(p+8)*nk),work(n2+(p+12)*nk),work(n2+(p+16)*nk),
	work(n2+(p+16)*nk+p),ier)

return
end

#Memory management subroutine
subroutine sspl1(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,center,
		dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
		xin,yin,win,sout,
		xwy,hs,sg,abd,p1ip,ssy,work,ier)
double precision x(n),y(n,p),w(n),knot(nk+4),tol,wp(p),
	 	dfoff,dfmax,cost,lambda,df,cv,gcv,coef(nk,p),s(n,p),lev(nef),
		xin(nefp1),yin(nefp1,p),win(nefp1),sout(nefp1,p),
		xwy(nk,p),hs(nk,4),sg(nk,4),abd(4,nk),p1ip(4,nk),
		ssy(p),work(n)
integer         n,p,nefp1,nk,method,ier,match(n),nef
logical center
double precision  sbar, wmean,tdfoff
call suff2(n,nef,p,match,y,w,yin,win,ssy,work)
if(center){
	if(cost>0){tdfoff=dfoff-1/cost}
}
call sspl(xin,yin,win,nef,nefp1,p,knot,nk,method,tol,wp,ssy,
		tdfoff,dfmax,cost,lambda,df,cv,gcv,coef,sout,lev,
		xwy,hs,sg,abd,p1ip,ier)
	
#now unpack the results
do j=1,p{
	call unpack(n,nef,match,sout(1,j),s(1,j))
	if(center){
		sbar=wmean(nef,sout(1,j),win)
		do i=1,n{ s(i,j)=s(i,j)-sbar}
	}
}
if(center)df=df-1
return
end
subroutine namat(x,match,n,nef,work,iwork,xmiss,tol)
#returns match (order) and work(1:nef) is the sorted unique x
implicit double precision(a-h,o-z)
integer match(1),n,nef,iwork(1),index
double precision x(1),xmiss,work(1),tol,xend,xstart
do i=1,n {
	work(i)=x(i)
	iwork(i)=i
	}

call sortdi(work,n,iwork,1,n)
xstart=x(iwork(1))
index=n
xend=x(iwork(n))
while(xend >= xmiss & index > 1){
	index=index-1
	xend=x(iwork(index))
	}

tol=tol*(xend-xstart)
index=1
work(1)=xstart
for(i=1;i<=n;i=i+1){
	while((x(iwork(i))-xstart)<tol){
		match(iwork(i))=index
		i=i+1
		if(i>n)goto 10
		}
	xstart= x(iwork(i))
	index=index+1
	match(iwork(i))=index
	work(index)=xstart
	}
10 if(xstart >= xmiss) 
	{nef=index-1}
	 else {nef=index}
return 
end

subroutine simfit(x,y,w,n,p,dfoff,cost,wp,gcv,coef,s,type,center,work)
#
# computes constant and linear fits, and selects the best using gcv
#
implicit double precision (a-h,o-z)

integer n,p,type
double precision x(n),y(n,p),w(n),cost,dfoff,wp(p),gcv,coef(2,p),
	s(n,p),work(p)
logical center
double precision sx,sy,sumw, dcent,sxx,syy,sxy
dcent=1-dble(center)
#center is F for no centering, else T
#Note: if type enters 1 or 2, no selection is made
sumw=0d0;gcvc=0d0;
do i=1,n {
  sumw=sumw+w(i)
}
if(type!=1){#either 0 or 2 in which case the linear is needed as well
	sx=0.0 ;  sxx=0.0; gcvl=0d0;

	do i=1,n {
  	sx=sx+w(i)*x(i)
	}
	xbar=sx/sumw
	do i=1,n {
	  sxx=sxx+w(i)*(x(i)-xbar)*x(i)
	}
}
do j=1,p{
	sy=0d0;syy=0d0;
	do i=1,n{
		sy=sy+w(i)*y(i,j)
	}
	work(j)=sy/sumw
  	do i=1,n{
		syy=syy+w(i)*(y(i,j)-work(j))*y(i,j)
	}
	gcvc=gcvc+wp(j)*syy
	if(type!=1){ #once again, do for linear as well
		sxy=0.0;
  		do i=1,n{
  			sxy=sxy+w(i)*(x(i)-xbar)*y(i,j)
		}
		coef(2,j)=sxy/sxx
		gcvl=gcvl+wp(j)*(syy -sxy*coef(2,j))
	}
}
if(type==0){
	gcvc =gcvc/ (sumw* (1 - (dfoff*cost + dcent)/sumw)**2 )
	gcvl=gcvl/(sumw* (1 - (dcent + (dfoff +1)* cost)/sumw)**2)
	if(gcvc<=gcvl){
		type=1
		gcv=gcvc
	}
	else{
		type=2
		gcv=gcvl
	}
}
else {

	if(type==1) {gcv=gcvc/(sumw* (1 - (dfoff*cost + dcent)/sumw)**2 )}
	else {gcv=gcvl/(sumw* (1 - (dcent + (dfoff + 1)*cost)/sumw)**2)}
}
	
if(type==1){
	do j=1,p{
		coef(1,j)=work(j)*dcent
		coef(2,j)=0d0
		do i=1,n {s(i,j)=coef(1,j)}
		}
	}
else{
	do j=1,p{
		coef(1,j)=work(j)*dcent-xbar*coef(2,j)
		do i=1,n {s(i,j)=coef(1,j)+coef(2,j)*x(i)}
		}
	}	
return
end

 
subroutine sspl2(x,y,w,n,p,knot,nk,wp,match,nef,
		dfoff,dfmax,cost,lambda,df,gcv,coef,s,type,center,
		xrange,work,tol,ier)
double precision x(n),y(n,p),w(n),knot(nk+4),wp(p),
	 	dfoff,dfmax,cost,lambda,df,gcv,coef(1),s(n,p),
		xrange(2),work(1),tol
#this routine selects from the linear and constant model as well
#see documentation for sspl
#workspace must be (2*p+2)*nefp1 + (p+16)*nk + 2*n
#if type>0 then no selection is performed; the fit is simply computed.
integer         n,p,nk,nef,type,match(n),ier,method
double precision coef1,coef2,cv
logical center
#center is F for no centering, else T
if(type==3){
	method=1
call sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,center,
		dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,work(1),
		xrange,work(n+1),ier)
	return
}
if(type>0){
       call simfit(x,y,w,n,p,dfoff,cost,wp,gcv,coef,s,
		type,center,work)
	df=dble(type)- dble(center)
return
}
#selection is being performed
method=3
call sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,center,
		dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,work(1),
		xrange,work(n+1),ier)
gcv1=gcv
call simfit(x,y,w,n,p,dfoff,cost,wp,gcv,work,work(2*p+1),type,center,
		work((n+2)*p+1))
if(gcv<=gcv1){
#the coef swapping is so as not to destroy the spline coefs if needed
	df=dble(type)- dble(center)
	do j=1,p{
		coef1=work(1+(j-1)*2)
		coef2=work(2+(j-1)*2)
		if(type==1){
			do i=1,n {s(i,j)=coef1}
			}
		else{
			do i=1,n {s(i,j) =coef1+coef2*x(i)}
			}
		coef(1+(j-1)*2)=coef1
		coef(2+(j-1)*2)=coef2
		}
	}
else{
	type=3
	gcv=gcv1
	}
return
end
subroutine psspl2(x,n,p,knot,nk,xrange,coef,coefl,s,order,type)
implicit double precision(a-h,o-z) 
#make predictions from a fitted smoothing spline 
double precision x(n),knot(nk+4),xrange(2),coef(nk,p),coefl(2,1),s(n,p)
integer n,p,nk,order, type
double precision ytemp
switch(type){
	case 1:{
		do j=1,p{
			if(order>=1){ytemp=0d0} else {ytemp=coefl(1,j)}
			do i=1,n {s(i,j)=ytemp}
		}
	}
	case 2:{
		if(order>=1){
			do j=1,p{
				if(order==1){ytemp=coefl(2,j)}
					else {ytemp=0d0}
				do i =1,n {s(i,j)=ytemp}
			}
		}
		else{ 
			do j=1,p{
				do i=1,n {s(i,j)=coefl(1,j)+coefl(2,j)*x(i)}
			}
		}
	}
	case 3: {
		call psspl(x,n,p,knot,nk,xrange,coef,s,order)
	}
}
return
end

		

subroutine psspl(x,n,p,knot,nk,xrange,coef,s,order)
implicit double precision(a-h,o-z) 
#make predictions from a fitted smoothing spline, linear or constant
double precision x(n),knot(nk+4),xrange(2),coef(nk,p),s(n,p)
integer n,p,nk,order
double precision xcs,xmin,xdif, endv(2),ends(2),xends(2),stemp
double precision bvalue
integer where
if(order>2|order<0)return
xdif=xrange(2)-xrange(1)
xmin=xrange(1)
xends(1)=0d0
xends(2)=1d0
do j=1,p{
	if(order==0){
		endv(1)=bvalue(knot,coef(1,j),nk,4,0d0,0)
		endv(2)=bvalue(knot,coef(1,j),nk,4,1d0,0)
		}
	if(order<=1){
		ends(1)=bvalue(knot,coef(1,j),nk,4,0d0,1)
		ends(2)=bvalue(knot,coef(1,j),nk,4,1d0,1)
		}
	do i=1,n{
		xcs=(x(i)-xmin)/xdif
		where=0
		if(xcs<0d0){where=1}
		if(xcs>1d0){where=2}
		if(where>0){#beyond extreme knots
			switch(order){
				case 0:
					stemp=endv(where)+
						(xcs-xends(where))*ends(where)
				case 1:
					stemp=ends(where)
				case 2:
					stemp=0d0
			}
		}
		else {stemp=bvalue(knot,coef(1,j),nk,4,xcs,order)}
		if(order>0){s(i,j)=stemp/(xdif**dble(order))}
		else s(i,j)=stemp
	}
}
return
end

subroutine calcvar(nx,n,px,qr,qrank,qpivot,cov,tmpcov,work)

implicit double precision (a-h,o-z)
integer n,px,qrank,qpivot(px)
double precision qr(nx,px),cov(px,px), tmpcov(px,px),work(1)
    
double precision dsum
integer i,j,km
#  compute the unscaled coviance matrix for the linear coefficients
do i=1,qrank{
	do j=1,qrank{
		tmpcov(i,j)=0d0
		cov(i,j)=qr(i,j)
		}
	tmpcov(i,i)=1e0
	}
info=0

#
# notice I have put px in as the row dim of cov
#
call dbksl(cov,px,qrank,tmpcov,px,info)

do i=1,qrank{
	do j=i,qrank{
		dsum=0e0
		km=max(i,j)
		for(k=km;k<=qrank;k=k+1){
			dsum=dsum+tmpcov(i,k)*tmpcov(j,k)
			}
		tmpcov(i,j)=dsum
		tmpcov(j,i)=dsum
		}
	}
# no
do i=1,qrank{
	do j=1,qrank{		
                 cov(i,j)=tmpcov(i,j)
		}
	}
return
end

subroutine orthreg(nx,n,p,x,in, y,res)
#
# does lin reg of y on x. assumes that x is orthogonal with cols > 1 having mean 0
#
# 
# "in" is a vector of column indicators (0 means term is to be deleted)

implicit double precision (a-h,o-z)


integer n,nx,p, in(p)
double precision x(nx,p),y(n),res(n)
do i=1,n {
 res(i)=y(i)
}
do j=1,p {
 if(in(j).eq.1){
 temp1=0
 temp2=0
 do i=1,n {
     temp1=temp1+res(i)*x(i,j)
     temp2=temp2+x(i,j)*x(i,j)
 }
 beta=temp1/temp2
 do i=1,n {
   res(i)=res(i)-beta*x(i,j)
 }
}}
return
end

 

subroutine bruto(x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,
		dfmax,cost,lambda,df,coef,type,xrange,
		gcvsel,gcvbak,dfit,maxit,nit,
		eta,resid,thresh,work,trace)
implicit double precision(a-h,o-z) 
double precision x(n,q),y(n,p),w(n),knot(nkmax+4,q),wp(p),
	 	dfmax(q),cost,lambda(q),df(q),coef(nkmax*p,q),xrange(2,q),
		gcvsel(q,maxit(1)),gcvbak(q,maxit(2)),dfit(q,maxit(1)),
		eta(n,p),resid(n,p),thresh,work(1)
integer n,q,p,nkmax,nk(q),match(n,q),nef(q),type(q),maxit(2),
		nit(2)
logical trace, select

#Compute residuals
do j=1,p{
	do i=1,n{resid(i,j)=y(i,j)-eta(i,j)}
}

#bruto backfitting with selection
select=.true.
call bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,
		dfmax,cost,lambda,df,coef,type,xrange,
		gcvsel,dfit,maxit(1),nit(1),eta,resid,thresh*10d0,work,trace)
#regular backfitting
select=.false.
call bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,
		dfmax,cost,lambda,df,coef,type,xrange,
		gcvbak,dfit,maxit(2),nit(2),eta,resid,thresh,work,trace)
do j=1,p{
	do i=1,n{eta(i,j)=y(i,j)-resid(i,j)}
}
return
end
subroutine bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,
		dfmax,cost,lambda,df,coef,type,xrange,
		gcv,dfit,maxit,nit,s,resid,thresh,work,trace)
implicit double precision(a-h,o-z) 
double precision x(n,q),y(n,p),w(n),knot(nkmax+4,q),wp(p),
	 	dfmax(q),cost,lambda(q),df(q),coef(nkmax*p,q),
		xrange(2,q),gcv(q,maxit),dfit(q,maxit),
		s(n,p),resid(n,p),thresh,work(1)
integer n,q,p,nkmax,nk(q),match(n,q),nef(q),type(q),maxit,nit,ier,ntype
double precision   dfoff, gcv0, ndf,gcv1,gcvrat,ndfoff,wmean,sbar,rss,tol
logical center, select, trace
center=.true.
tol=1d-3  # this is the convergence criterion for gcv
#remove the mean from the residuals, and compute rss
rss=0d0
do j=1,p{
	sbar=wmean(n,resid(1,j),w)
	do i=1,n{
		resid(i,j)=resid(i,j) - sbar
		work(i)=resid(i,j)**2
		}
rss=rss+wp(j)*wmean(n,work,w)
}
dfoff=0
do k=1,q {dfoff=dfoff+df(k)}
gcv1=rss/((1-(1+dfoff*cost)/n)**2)
gcvrat=1d0
nit=0
while(nit<maxit&gcvrat >thresh ){
	gcv0=gcv1
	nit=nit+1
	do k=1,q{
		gcv(k,nit)=gcv1
		if(!select&type(k)==1)next
# form partial residuals if necessary
		if(type(k)>1){ #get the fitted values
			call psspl2(x(1,k),n,p,knot(1,k),nk(k),xrange(1,k),
				coef(1,k),coef(1,k),s,0,type(k))
			do j=1,p{
				sbar=wmean(n,s(1,j),w)
				do i=1,n{resid(i,j)=resid(i,j) + s(i,j)-sbar}
			}
		}

		ndfoff=dfoff-df(k)
		if(select) {ntype=0}
		else {ntype=type(k)}
		call sspl2(x(1,k),resid,w,n,p,knot(1,k),nk(k),wp,match(1,k),
		nef(k),ndfoff,dfmax(k),cost,lambda(k),ndf,
			gcv1,coef(1,k),s,ntype,center,
			xrange(1,k),work,tol,ier)
		gcv(k,nit)=gcv1
		if(select){
			dfit(k,nit)=ndf
			df(k)=ndf
			dfoff=ndfoff+ndf
			type(k)=ntype
		}
		if(type(k)>1){
			do j=1,p{
				do i=1,n{
					resid(i,j)=resid(i,j) - s(i,j)
					}
			}
		}
	}
gcvrat=dabs(gcv1-gcv0)/gcv0
if(trace){
	call intpr("outer iteration",15,nit,1)
	call dblepr("gcv  ",5,gcv1,1)
	call dblepr("ratio",5,gcvrat,1)
}
}
return
end
subroutine pbruto(x,n,q,ybar,p,knot,nkmax,nk,coef,type,xrange,
		eta,work)
implicit double precision(a-h,o-z) 
double precision x(n,q),ybar(p),knot(nkmax+4,q),coef(nkmax*p,q),
		xrange(2,q),eta(n,p),work(n,p)
integer n,q,p,nkmax,nk(q),type(q)
#initialization
do j=1,p{
	do i =1,n {eta(i,j)=ybar(j)}
}
do k=1,q{
	if(type(k)==1)next
	call psspl2(x(1,k),n,p,knot(1,k),nk(k),xrange(1,k),
			coef(1,k),coef(1,k),work,0,type(k))
	do j=1,p{
		do i=1,n{eta(i,j)=eta(i,j) + work(i,j)}
	}
}
return
end

polybasis <-
function (x, degree = 1, monomial = FALSE) 
{
    if (degree >= 3) 
        warning("This is not a smart polynomial routine. You may get numerical problems with a degree of 3 or more")
    x <- as.matrix(x)
    dn <- dimnames(x)
    dd <- dim(x)
    np <- dd[2]
    if (np == 1) 
        monomial <- TRUE
    if (degree > 1) {
        if (monomial) {
            px <- x
            cc <- sapply(split(paste(diag(np)), rep(seq(np), 
                rep(np, np))), paste, collapse = "")
            tx <- x
            for (i in 2:degree) {
                px <- px * tx
                x <- cbind(x, px)
                cc <- c(cc, sapply(split(paste(diag(np) * i), 
                  rep(seq(np), rep(np, np))), paste, collapse = ""))
            }
        }
        else {
            matarray <- array(x, c(dd, degree))
            for (i in 2:degree) matarray[, , i] <- x^i
            matarray <- aperm(matarray, c(1, 3, 2))
            x <- matarray[, , np]
            ad0 <- seq(degree)
            ad <- ad0
            ncol.mat0 <- degree
            ncol.x <- degree
            d0 <- paste(ad0)
            cc <- d0
            for (ii in seq(np - 1, 1)) {
                index0 <- rep(seq(ncol.mat0), ncol.x)
                index <- rep(seq(ncol.x), rep(ncol.mat0, ncol.x))
                newad <- ad0[index0] + ad[index]
                retain <- newad <= degree
                mat0 <- matarray[, , ii]
                if (any(retain)) 
                  newmat <- mat0[, index0[retain], drop = FALSE] * 
                    x[, index[retain], drop = FALSE]
                else newmat <- NULL
                ddn <- paste(d0[index0[retain]], cc[index[retain]], 
                  sep = "")
                zeros <- paste(rep(0, nchar(cc[1])), collapse = "")
                cc <- paste(0, cc, sep = "")
                d00 <- paste(d0, zeros, sep = "")
                x <- cbind(mat0, x, newmat)
                cc <- c(d00, cc, ddn)
                ad <- c(ad0, ad, newad[retain])
                ncol.x <- length(ad)
            }
        }
        if (!is.null(dn)) 
            dn[[2]] <- cc
        else dn <- list(NULL, cc)
        dimnames(x) <- dn
    }
    cbind(Intercept = 1, x)
}

      subroutine tpack(n,p,match,x,xbar)
      integer n,p,match(n)
      double precision x(n),xbar(n)
      do 23000 i=1,p
      xbar(i)=0e0
23000 continue
      do 23002 i=1,n
      xbar(match(i))=xbar(match(i))+x(i)
23002 continue
      return
      end
      subroutine suff(n,p,match,x,y,w,xbar,ybar,wbar,work)
      integer n,p,match(n)
      double precision x(n),xbar(n),y(n),ybar(n),w(n),wbar(n),work(n)
      call tpack(n,p,match,w,wbar)
      do 23004 i=1,n
      xbar(match(i))=x(i)
23004 continue
      do 23006 i=1,n
      work(i)=y(i)*w(i)
23006 continue
      call tpack(n,p,match,work,ybar)
      do 23008 i=1,p
      if(.not.(wbar(i).gt.0e0))goto 23010
      ybar(i)=ybar(i)/wbar(i)
      goto 23011
23010 continue
      ybar(i)=0e0
23011 continue
23008 continue
      return
      end
      subroutine untpack(n,p,match,xbar,x)
      integer n,p, match(n)
      double precision x(n),xbar(p+1)
      if(.not.(p.lt.n))goto 23012
      xbar(p+1)=0e0
23012 continue
      do 23014 i = 1,n
      x(i)=xbar(match(i))
23014 continue
      return
      end

polyreg <-
function (x, y, w, degree = 1, monomial = FALSE, ...) 
{
    x <- polybasis(x, degree, monomial)
    y <- as.matrix(y)                   # just making sure ...
    if (iswt <- !missing(w)) {
        if (any(w <= 0)) 
            stop("only positive weights")
        w <- sqrt(w)
        y <- y * w
        x <- x * w
    }
    qrx <- qr(x)
    coef <- as.matrix(qr.coef(qrx, y))
    fitted <- qr.fitted(qrx, y)
    if ((df <- qrx$rank) < ncol(x)) 
        coef[qrx$pivot, ] <- coef
    if (iswt) 
        fitted <- fitted/w
    structure(list(fitted.values = fitted, coefficients = coef, 
        degree = degree, monomial = monomial, df = df), class = "polyreg")
}

c$$$ Naras fix
c$$$      subroutine calcvar(nx,n,px,qr,qrank,qpivot,cov,tmpcov,work)
      subroutine calcvar(nx,px,qr,qrank,cov,tmpcov)
      implicit double precision (a-h,o-z)
c$$$ Naras fix
c$$$      integer n,px,qrank,qpivot(px)
c$$$      double precision qr(nx,px),cov(px,px), tmpcov(px,px),work(1)
      integer px,qrank
      double precision qr(nx,px),cov(px,px), tmpcov(px,px)
      double precision dsum
      integer i,j,km
      do 23000 i=1,qrank
      do 23002 j=1,qrank
      tmpcov(i,j)=0d0
      cov(i,j)=qr(i,j)
23002 continue
      tmpcov(i,i)=1e0
23000 continue
      info=0
c R version has different args
c      call dbksl(cov,px,qrank,tmpcov,px,info)
      do 20 j = 1, qrank
         call dtrsl(cov, px, qrank, tmpcov(1,j), 01, info)
 20   continue
      do 23004 i=1,qrank
      do 23006 j=i,qrank
      dsum=0e0
      km=max(i,j)
      k=km
23008 if(.not.(k.le.qrank))goto 23010
      dsum=dsum+tmpcov(i,k)*tmpcov(j,k)
      k=k+1
      goto 23008
23010 continue
      tmpcov(i,j)=dsum
      tmpcov(j,i)=dsum
23006 continue
23004 continue
      do 23011 i=1,qrank
      do 23013 j=1,qrank
      cov(i,j)=tmpcov(i,j)
23013 continue
23011 continue
      return
      end

gen.ridge <-
  function (x, y, weights, lambda = 1, omega, df, ...) 
{
  if (missing(df) && lambda <= .Machine$double.eps) 
    return(polyreg(x, y))
  d <- dim(x)
  mm <- apply(x, 2, mean)
  x <- scale(x, mm, FALSE)
  simple <- if (missing(omega)) 
    TRUE
  else FALSE
  if (!simple) {
    if (!all(match(c("values", "vectors"), names(omega), 
                   FALSE))) 
      stop("You must supply an  eigen-decomposed version of omega")
    vals <- pmax(sqrt(.Machine$double.eps), sqrt(omega$values))
    basis <- scale(omega$vectors, FALSE, vals)
    x <- x %*% basis
  }
  svd.x <- svd(x)
  dd <- svd.x$d
  if (!missing(df))lambda=df.gold(dd,df)
  df=sum(dd^2/(dd^2 + lambda))
  y <- ( t(t(y) %*% svd.x$u) * dd)/(dd^2 + lambda)
  coef <- svd.x$v %*% y
  fitted <- x %*% coef
  if (!simple) 
    coef <- basis %*% coef
  structure(list(fitted.values = fitted, coefficients = coef, 
                 df = df, lambda = lambda, xmeans = mm), class = "gen.ridge")
}

predict.mars <-
function (object, newdata, ...) 
{
    if (missing(newdata)) {
        z <- fitted(object)
        if (is.null(z)) 
            stop("need to supply newdata")
        else return(z)
    }
    model.matrix.mars(object, newdata) %*% object$coefficients
}

mars <-
function (x, y, w = rep(1, nrow(x)), wp, degree = 1, nk = max(21,
    2 * ncol(x) + 1), penalty = 2, thresh = 0.001, prune = TRUE,
    trace.mars = FALSE, forward.step = TRUE, prevfit = NULL, ...)
{
    this.call <- match.call()
    if ((nk%%2) != 1)
        nk <- nk - 1
    x <- as.matrix(x)
    np <- dim(x)
    n <- np[1]
    p <- np[2]
    y <- as.matrix(y)
    nclass <- ncol(y)
    if (is.null(np)) {
        np <- c(length(x), 1)
        x <- as.matrix(x)
    }
    if (forward.step) {
        interms <- 1
        lenb <- nk
        bx <- matrix(rep(0, nrow(x) * nk), nrow = n)
        res <- matrix(rep(0, nrow(x) * ncol(y)), nrow = n)
        fullin <- rep(0, nk)
        cuts <- NULL
        factor <- NULL
    }
    else {
        bx <- model.matrix.mars(prevfit, x, full = TRUE)
        interms <- ncol(bx)
        lenb <- prevfit$lenb
        o <- prevfit$all.terms
        fullin <- rep(0, ncol(bx))
        fullin[o] <- 1
        res <- prevfit$res
        factor <- prevfit$factor
        cuts <- prevfit$cuts
        if (missing(penalty))
            penalty <- prevfit$penalty
        degree <- prevfit$degree
        nk <- lenb
        thresh <- prevfit$thresh
    }
    if (missing(penalty) & (degree > 1))
        penalty <- 3
    if (!missing(wp)) {
        if (any(wp <= 0))
            stop("wp should all be positive")
        wp <- sqrt(wp/sum(wp))
        y <- y * outer(rep(1, n), wp)
    }
    else wp <- NULL
    tagx <- x
    storage.mode(tagx) <- "integer"
    for (j in 1:p) {
        tagx[, j] <- order(x[, j])
    }
    bestin <- rep(0, nk)
    flag <- matrix(rep(0, nk * p), nrow = nk, ncol = p)
    if (is.null(cuts))
        cuts <- matrix(rep(0, nk * p), nrow = nk, ncol = p)
    if (is.null(factor)) {
        dir <- matrix(rep(0, nk * p), nrow = nk, ncol = p)
    }
    else {
        dir <- factor
    }
    alpha <- rep(0, nclass)
    beta <- matrix(rep(0, nk * nclass), nrow = nk)
    bestgcv <- 0
    storage.mode(y) <- "double"
    storage.mode(x) <- "double"
    storage.mode(bx) <- "double"
    storage.mode(flag) <- "integer"
    storage.mode(cuts) <- "double"
    storage.mode(dir) <- "double"
    storage.mode(res) <- "double"
    storage.mode(beta) <- "double"
    lenscrat <- 1 + n + 2 * n * nk + 4 * nk * nk + 3 * nk + 3 *
        nk * nclass + 3 * nclass + 28 * n + 51
    junk <- .Fortran("marss", as.integer(n), as.integer(n), as.integer(p),
        as.integer(nclass), as.matrix(y), as.matrix(x), as.double(w),
        as.matrix(tagx), as.integer(degree), as.integer(nk),
#        as.double(penalty), as.double(thresh), as.logical(forward.step),
        as.double(penalty), as.double(thresh), as.integer(forward.step),
#        as.integer(interms), as.logical(prune), bx = as.matrix(bx),
        as.integer(interms), as.integer(prune), bx = as.matrix(bx),
        fullin = as.integer(fullin), lenb = as.integer(lenb),
        bestgcv = as.double(bestgcv), bestin = as.integer(bestin),
        flag = as.matrix(flag), cuts = as.matrix(cuts), dir = as.matrix(dir),
## Naras fixes
##         res = as.matrix(res), alpha = as.double(alpha), beta = as.matrix(beta),
        res = as.matrix(res), beta = as.matrix(beta),
        double(lenscrat), integer(4 * nk), as.integer(trace.mars), PACKAGE = "mda")
    lenb <- junk$lenb
    all.terms <- seq(lenb)[junk$fullin[1:lenb] == 1]
    selected.terms <- seq(lenb)[junk$bestin[1:lenb] == 1]
    coefficients <- junk$beta[seq(selected.terms), , drop = FALSE]
    residuals <- junk$res
    fitted.values <- y - residuals
    if (!is.null(wp)) {
        TT <- outer(rep(1, n), wp)
        residuals <- residuals/TT
        fitted.values <- fitted.values/TT
        coefficients <- coefficients/outer(rep(1, length(selected.terms)),
            wp)
    }
    dir <- junk$dir[seq(lenb), , drop = FALSE]
    dimnames(dir) <- list(NULL, dimnames(x)[[2]])
    cutss <- junk$cuts[seq(lenb), , drop = FALSE]
    x <- junk$bx[, selected.terms, drop = FALSE]
    structure(list(call = this.call, all.terms = all.terms, selected.terms = selected.terms,
        penalty = penalty, degree = degree, nk = nk, thresh = thresh,
        gcv = junk$bestgcv, factor = dir, cuts = cutss, residuals = residuals,
        fitted.values = fitted.values, lenb = junk$lenb, coefficients = coefficients,
        x = x), class = "mars")
}

mda.fit <-
function (x, g, weights, theta, assign.theta, Rj, sub.df=NULL, tot.df=NULL,
    dimension = R - 1, eps = 100 * .Machine$double.eps, method = polyreg,
    ...)
{
    this.call <- match.call()
    n <- nrow(x)
    cnames <- names(weights)
    J <- length(cnames)
    R <- sum(Rj)
    wtots <- lapply(weights, function(x) apply(x, 2, sum))
    sub.prior <- lapply(wtots, function(x) x/sum(x))
    dp <- unlist(wtots)
    subclass.names <- names(dp)
    dp <- dp/sum(dp)
    if (missing(theta))
        theta <- contr.helmert(R)
    theta <- contr.fda(dp, theta)
    if (!(is.null(sub.df) & is.null(tot.df))) {
        Q <- diag(dp) + transformPenalty(prior = dp, cl = rep(seq(J),
            Rj), df = sub.df, tot.df = tot.df)
        theta <- fix.theta(theta, Q)
    }
    Theta <- matrix(0, n, R - 1)
    obs.weights <- double(n)
    for (j in seq(J)) {
        Theta[g == j, ] <- weights[[j]] %*% theta[assign.theta[[j]],
            , drop = FALSE]
        obs.weights[g == j] <- weights[[j]] %*% rep(1, Rj[j])
    }
    fit <- method(x, Theta, obs.weights, ...)
    Theta <- Theta * obs.weights
    ssm <- t(Theta) %*% fitted(fit)/n
    ed <- svd(ssm, nu = 0)
    thetan <- ed$v
    lambda <- ed$d
    lambda[lambda > 1 - eps] <- 1 - eps
    discr.eigen <- lambda/(1 - lambda)
    pe <- (100 * cumsum(discr.eigen))/sum(discr.eigen)
    dimension <- min(dimension, sum(lambda > eps))
    if (dimension == 0) {
        warning("degenerate problem; no discrimination")
        return(structure(list(dimension = 0, fit = fit, call = this.call),
            class = "fda"))
    }
    thetan <- thetan[, seq(dimension), drop = FALSE]
    pe <- pe[seq(dimension)]
    alpha <- sqrt(lambda[seq(dimension)])
    sqima <- sqrt(1 - lambda[seq(dimension)])
    vnames <- paste("v", seq(dimension), sep = "")
    means <- scale(theta %*% thetan, FALSE, sqima/alpha)
    dimnames(means) <- list(subclass.names, vnames)
    names(lambda) <- c(vnames, rep("", length(lambda) - dimension))
    names(pe) <- vnames
    list(percent.explained = pe, values = lambda, means = means,
        theta.mod = thetan, dimension = dimension, sub.prior = sub.prior,
        fit = fit, call = this.call)
}

predict.polyreg <-
function (object, newdata, ...) 
{
    if (missing(newdata)) {
        z <- fitted(object)
        if (is.null(z)) 
            stop("need to supply newdata")
        else return(z)
    }
    degree <- object$degree
    monomial <- object$monomial
    polybasis(newdata, degree, monomial) %*% object$coef
}

lvq.start <-
function (x, g, subclasses) 
{
    cnames <- levels(fg <- factor(g))
    J <- length(cnames)
    g <- as.numeric(g)
    weights <- as.list(cnames)
    names(weights) <- cnames
    subclasses <- rep(subclasses, length = length(cnames))
    size <- sum(subclasses)
    cb <- lvqinit(x, g, size = size)
    TT <- olvq1(x, g, codebk = cb)
    TT <- lvq3(x, g, codebk = TT)
    cl <- as.numeric(TT$cl)
    R <- length(cl)
    cx <- TT$x
    p <- ncol(cx)
    for (j in seq(J)) {
        which <- cl == j
        number <- sum(which)
        if (number == 0) {
            cx <- rbind(cx, apply(x[g == j, ], 2, mean))
            cl <- c(cl, j)
            wmj <- matrix(1, sum(g == j), 1)
            number <- 1
        }
        else if (number == 1) 
            wmj <- matrix(1, sum(g == j), 1)
        else {
            jcx <- cx[which, ]
            jcl <- seq(number)
            jcluster <- lvqtest(list(x = jcx, cl = jcl), x[g == 
                j, ])
            needed <- unique(jcluster)
            rcl <- rep(0, number)
            rcl[needed] <- j
            cl[which] <- rcl
            wmj <- diag(number)[jcluster, needed, drop = FALSE]
            number <- length(needed)
        }
        dimnames(wmj) <- list(NULL, paste("s", seq(number), sep = ""))
        weights[[j]] <- wmj
    }
    TT <- cl > 0
    list(x = cx[TT, , drop = FALSE], cl = factor(cl[TT], labels = cnames), 
        weights = weights)
}

      subroutine sspl(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,dfoff,
     1     dfmax,cost,lambda,df,cv,gcv,coef,s,lev,xwy,hs,sg,abd,p1ip,
     2     ier)
      implicit double precision(a-h,o-z)
C     A Cubic B-spline Smoothing routine.
C     
C     The algorithm minimises:
C     
C     (1/n) * sum w(i)* (y(i)-s(i))**2 + lambda* int ( s"(x) )**2 dx
C     
C     for each of p response variables in y
C
C     Input:
C      
C     x(n)        vector containing the ordinates of the observations
C     y(ldy,p)    matrix (n x p) of responses (ldy can be greater than n)
C     w(n)        vector containing the weights given to each data point
C     n           number of data points
C     ldy         leading dimension of y
C     p           number of columns in y
C     knot(nk+4)  vector of knot points defining the cubic b-spline basis.
C     nk          number of b-spline coefficients to be estimated
C     nk <= n+2
C     method      method for selecting amount of smoothing, lambda
C          1 = fixed lambda
C          2 = fixed df
C          3 = gcv
C          4 = cv
C     tol         used in Golden Search routine
C     wp(p)       weights, length p,  used to combine cv or gcv in 3 or 4 above
C     ssy(p)      offsets for weighted sum of squares for y; can be all zero,
C                   else should be the variability lost due to collapsing
C                   onto unique values
C     dfoff       offset df used in gcv calculations (0 is good default)
C     dfmax       maximum value for df allowed when gcv or cv are used
C                   routine simply returns the value at dfmax if it was exceeded
C     cost        cost per df (1 is good default)
C
C     Input/Output:
C
C     lambda      penalised likelihood smoothing parameter
C     df          trace(S)
C
C     Output:
C
C     cv          omnibus cv criterion
C     gcv         omnibus gcv criterion (including penalty and offset)
C     coef(nk,p)  vector of spline coefficients
C     s(ldy,p)    matrix of smoothed y-values
C     lev(n)      vector of leverages
C
C     Working arrays/matrix:
C
C     xwy(nk,p)   X'Wy
C     hs(nk,4)    the diagonals of the X'WX matrix
C     sg(nk,4)    the diagonals of the Gram matrix
C     abd(4,nk)   [ X'WX+lambda*SIGMA] in diagonal form
C     p1ip(4,nk)  inner products between columns of L inverse
C     ier          error indicator
C        ier = 0 ___  everything fine
C        ier = 1 ___  spar too small or too big
C                     problem in cholesky decomposition
C
      integer n,p,ldy,nk,method,ier
      double precision x(n),y(ldy,p),w(n),knot(nk+4),tol,wp(p),ssy(p),
     *     dfoff,dfmax,cost,lambda,df,cv,gcv,coef(nk,p),s(ldy,p),
     *     lev(n),xwy(nk,p),hs(nk,4),sg(nk,4),abd(4,nk),p1ip(4,nk)

C  Compute SIGMA, X'WX, X'WY, trace, ratio, s0, s1.
C SIGMA-> sg[]
C X'WX -> hs[]
C X'WY -> xwy[]
      call sgram(sg(1,1),sg(1,2),sg(1,3),sg(1,4),knot,nk)
      call stxwx2(x,y,w,n,ldy,p,knot,nk,xwy,hs(1,1),hs(1,2),
     *     hs(1,3),hs(1,4))
C Compute estimate
      if(method.eq.1)then
C     Value of lambda supplied
c$$$Naras fix
c$$$         call sslvr2(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,dfoff,cost,
         call sslvr2(x,y,w,n,ldy,p,knot,nk,method,wp,ssy,dfoff,cost,
     1        lambda,df,cv,gcv,coef,s,lev,xwy,
     2        hs(1,1),hs(1,2),hs(1,3),hs(1,4),
     3        sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
      else
C     Use Forsythe, Malcom and Moler routine to minimise criterion
         call fmm(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,dfoff,cost,
     *        lambda,df,cv,gcv,coef,s,lev,xwy,hs,sg,abd,p1ip,ier)

         if(method.gt.2.and.df.gt.dfmax)then
            df=dfmax
            call fmm(x,y,w,n,ldy,p,knot,nk,2,tol,wp,ssy,dfoff,cost,
     *           lambda,df,cv,gcv,coef,s,lev,xwy,hs,sg,abd,p1ip,ier)
         endif
      endif
      return
      end

      subroutine fmm(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,
     1     dfoff,cost,lambda,df,cv,gcv,coef,s,lev,xwy,hs,sg,abd,p1ip,
     2     ier)
      integer n,ldy,nvar,nk,method,ier
      double precision xs(n),ys(ldy,nvar),ws(n),knot(nk+4),tol,wp(nvar),
     *     ssy(nvar),dfoff,cost,lambda,df,cv,gcv,coef(nk,nvar),
     *     s(ldy,nvar),lev(n),xwy(nk,nvar),hs(nk,4),sg(nk,4),abd(4,nk),
     *     p1ip(4,nk)
      double precision t1,t2,ratio, a,b,c,d,e,eps,xm,p,q,r,tol1,tol2,u,
     *     v,w, fu,fv,fw,fx,x,targdf,ax,bx
      integer i
C used to be lspar
      ax=1d-10
C used to be uspar
      bx=1.5
      t1=0.
      t2=0.
      targdf=df
      do 23004 i=3,nk-3
         t1 = t1 + hs(i,1)
23004 continue
      do 23006 i=3,nk-3
         t2 = t2 + sg(i,1)
23006 continue
      ratio = t1/t2
C
C  an approximation  x  to the point where f  attains a minimum  on
C  the interval  (ax,bx)  is determined.
C
C
C  input..
C
C  ax         left endpoint of initial interval
C  bx         right endpoint of initial interval
C  f         function subprogram which evaluates  f(x)  for any  x
C         in the interval  (ax,bx)
C  tol         desired length of the interval of uncertainty of the final
C         result ( .ge. 0.0)
C
C
C  output..
C
C  fmin  abcissa approximating the point where        f  attains a minimum
C
C
C      the method used is a combination of  golden  section  search  and
C  successive parabolic interpolation.        convergence is never much slower
C  than  that  for  a  fibonacci search.  if  f  has a continuous second
C  derivative which is positive at the minimum (which is not  at  ax  or
C  bx),  then  convergence  is        superlinear, and usually of the order of
C  about  1.324....
C      the function  f        is never evaluated at two points closer together
C  than  eps*dabs(fmin) + (tol/3), where eps is        approximately the square
C  root  of  the  relative  machine  precision.   if   f   is a unimodal
C  function and the computed values of         f   are  always  unimodal  when
C  separated by at least  eps*dabs(x) + (tol/3), then  fmin  approximates
C  the abcissa of the global minimum of  f  on the interval  ax,bx  with
C  an error less than  3*eps*dabs(fmin) + tol.  if   f        is not unimodal,
C  then fmin may approximate a local, but perhaps non-global, minimum to
C  the same accuracy.
C      this function subprogram is a slightly modified        version  of  the
C  algol  60 procedure        localmin  given in richard brent, algorithms for
C  minimization without derivatives, prentice - hall, inc. (1973).
C
C
C      double precision  a,b,c,d,e,eps,xm,p,q,r,tol1,tol2,u,v,w
C      double precision  fu,fv,fw,fx,x
C
C  c is the squared inverse of the golden ratio
C
      c = 0.5*(3. - dsqrt(5d0))
C
C  eps is approximately the square root of the relative machine
C  precision.
C
      eps = 1d0
 10   eps = eps/2d0
      tol1 = 1d0 + eps
      if(tol1 .gt. 1d0)then
      go to 10
      endif
      eps = dsqrt(eps)
C
C  initialization
C
      a = ax
      b = bx
      v = a + c*(b - a)
      w = v
      x = v
      e = 0.0
c$$$      Naras fix: This also seems to be copied from Numerical Recipes
c$$$      So same fix, i.e. setting d=0 seems reasonable afaik. Same for fx,fu!
      d = 0.0
      fx = 0.0
      fu = 0.0
      lambda = ratio*16.**(-2. + x*(6.))
c$$$Naras fix
c$$$      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,dfoff,
      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,wp,ssy,dfoff,
     1     cost,lambda,df,cv,gcv,coef,s,lev,xwy,
     2     hs(1,1),hs(1,2),hs(1,3),hs(1,4),
     3     sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
      I23010 = (method)
      if(.not.(I23010.eq.( 2)))goto 23011
      fx=3d0+(targdf-df)**2
      goto 23010
23011 continue
      if(.not.(I23010.eq.( 3)))goto 23012
      fx=gcv
      goto 23010
23012 continue
      if(.not.(I23010.eq.( 4)))goto 23013
      fx=cv
23013 continue
23010 continue
      fv = fx
      fw = fx
C
C  main loop starts here
C
20    xm = 0.5*(a + b)
      tol1 = eps*dabs(x) + tol/3d0
      tol2 = 2d0*tol1
C
C  check stopping criterion
C
      if(dabs(x - xm) .le. (tol2 - 0.5*(b - a)))then
      go to 90
C
C is golden-section necessary
C
      endif
      if(dabs(e) .le. tol1)then
         go to 40
C
C     fit parabola
C
      endif
      r = (x - w)*(fx - fv)
      q = (x - v)*(fx - fw)
      p = (x - v)*q - (x - w)*r
      q = 2.00*(q - r)
      if(q .gt. 0.0)then
         p = -p
      endif
      q = dabs(q)
      r = e
      e = d
C
C is parabola acceptable
C
      if(dabs(p) .ge. dabs(0.5*q*r))then
         go to 40
      endif
      if(p .le. q*(a - x))then
         go to 40
      endif
      if(p .ge. q*(b - x))then
         go to 40
      endif
C
C  a parabolic interpolation step
C
      d = p/q
      u = x + d
C
C  f must not be evaluated too close to ax or bx
C
      if((u - a) .lt. tol2)then
      d = dsign(tol1, xm - x)
      endif
      if((b - u) .lt. tol2)then
      d = dsign(tol1, xm - x)
      endif
      go to 50
C
C  a golden-section step
C
40    if(x .ge. xm)then
      e = a - x
      endif
      if(x .lt. xm)then
      e = b - x
      endif
      d = c*e
C
C  f must not be evaluated too close to x
C
50    if(dabs(d) .ge. tol1)then
      u = x + d
      endif
      if(dabs(d) .lt. tol1)then
      u = x + dsign(tol1, d)
      endif
      lambda = ratio*16.**(-2. + u*(6.))
c$$$Naras fix
c$$$      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,tol,wp,ssy,dfoff,
      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,method,wp,ssy,dfoff,
     *cost,lambda,df,cv,gcv,coef,s,lev,xwy,hs(1,1),hs(1,2),hs(1,3),hs(1,
     *4),sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
      I23038 = (method)
      if(.not.(I23038.eq.( 2)))goto 23039
      fu=3d0+(targdf-df)**2
      goto 23038
23039 continue
      if(.not.(I23038.eq.( 3)))goto 23040
      fu=gcv
      goto 23038
23040 continue
      if(.not.(I23038.eq.( 4)))goto 23041
      fu=cv
23041 continue
23038 continue
C
C  update  a, b, v, w, and x
C
      if(fu .gt. fx)then
      go to 60
      endif
      if(u .ge. x)then
      a = x
      endif
      if(u .lt. x) then
      b = x
      endif
      v = w
      fv = fw
      w = x
      fw = fx
      x = u
      fx = fu
      go to 20
60    if(u .lt. x)then
      a = u
      endif
      if(u .ge. x)then
      b = u
      endif
      if(fu .le. fw)then
      go to 70
      endif
      if(w .eq. x)then
      go to 70
      endif
      if(fu .le. fv)then
      go to 80
      endif
      if(v .eq. x)then
      go to 80
      endif
      if(v .eq. w)then
      go to 80
      endif
      go to 20
70    v = w
      fv = fw
      w = u
      fw = fu
      go to 20
80    v = u
      fv = fu
      go to 20
C
C  end of main loop
C
90    continue
      if(method.eq.2)then
c$$$Naras fix
c$$$      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,1,tol,wp,ssy,dfoff,cost,
      call sslvr2(xs,ys,ws,n,ldy,nvar,knot,nk,1,wp,ssy,dfoff,cost,
     1        lambda,df,cv,gcv,coef,s,lev,xwy,
     2        hs(1,1),hs(1,2),hs(1,3),hs(1,4),
     3        sg(1,1),sg(1,2),sg(1,3),sg(1,4),abd,p1ip,ier)
      endif
      return
      end

      subroutine stxwx2(x,z,w,k,ldy,pz,xknot,n,y,hs0,hs1,hs2,hs3)
      implicit double precision(a-h,o-z)
      integer k,n,pz,ldy,j,i,pp,ileft,mflag
      double precision z(ldy,pz),w(k),x(k),xknot(n+4),y(n,pz),hs0(n),
     *     hs1(n),hs2(n),hs3(n),eps,vnikx(4,1),work(16)
C Initialise the output vectors
      do 23064 i=1,n
         hs0(i)=0d0
         hs1(i)=0d0
         hs2(i)=0d0
         hs3(i)=0d0
         do 23066 j=1,pz
            y(i,j)=0d0
23066    continue
23064 continue
C Compute X'WX -> hs0,hs1,hs2,hs3  and X'WZ -> y
      ileft = 1
      eps = .1d-9
      do 23068 i=1,k
         ileft = interv(xknot(1),(n+1),x(i), 0,0, ileft,mflag)
         if(mflag.eq. 1)then
            if(x(i).le.(xknot(ileft)+eps))then
               ileft=ileft-1
            else
               return
            endif
         endif
         call bsplvd (xknot,n+8,4,x(i),ileft,work,vnikx,1)
         j= ileft-4+1
         do 23074 pp=1,pz
            y(j,pp) = y(j,pp)+w(i)*z(i,pp)*vnikx(1,1)
23074    continue
         hs0(j)=hs0(j)+w(i)*vnikx(1,1)**2
         hs1(j)=hs1(j)+w(i)*vnikx(1,1)*vnikx(2,1)
         hs2(j)=hs2(j)+w(i)*vnikx(1,1)*vnikx(3,1)
         hs3(j)=hs3(j)+w(i)*vnikx(1,1)*vnikx(4,1)
         j= ileft-4+2
         do 23076 pp=1,pz
            y(j,pp) = y(j,pp)+w(i)*z(i,pp)*vnikx(2,1)
23076    continue
         hs0(j)=hs0(j)+w(i)*vnikx(2,1)**2
         hs1(j)=hs1(j)+w(i)*vnikx(2,1)*vnikx(3,1)
         hs2(j)=hs2(j)+w(i)*vnikx(2,1)*vnikx(4,1)
         j= ileft-4+3
         do 23078 pp=1,pz
            y(j,pp) = y(j,pp)+w(i)*z(i,pp)*vnikx(3,1)
23078    continue
         hs0(j)=hs0(j)+w(i)*vnikx(3,1)**2
         hs1(j)=hs1(j)+w(i)*vnikx(3,1)*vnikx(4,1)
         j= ileft-4+4
         do 23080 pp=1,pz
            y(j,pp) = y(j,pp)+w(i)*z(i,pp)*vnikx(4,1)
23080    continue
         hs0(j)=hs0(j)+w(i)*vnikx(4,1)**2
23068 continue
      return
      end

c$$$ Naras fix
c$$$      subroutine sslvr2(x,y,w,n,ldy,p,knot,nk,method,tol,wp,ssy,dfoff,
      subroutine sslvr2(x,y,w,n,ldy,p,knot,nk,method,wp,ssy,dfoff,
     *     cost,lambda,df,cv,gcv,coef,sz,lev,xwy,hs0,hs1,hs2,hs3,
     *     sg0,sg1,sg2,sg3,abd,p1ip,info)
      implicit double precision(a-h,o-z)
      integer n,p,ldy,nk,method,info
c$$$Naras fix
c$$$      double precision x(n),y(ldy,p),w(n),knot(nk+4),tol,wp(p),ssy(p),
      double precision x(n),y(ldy,p),w(n),knot(nk+4),wp(p),ssy(p),
     *     dfoff,cost,lambda,df,cv,gcv,coef(nk,p),sz(ldy,p),lev(n),
     $     xwy(nk,p), hs0(nk),hs1(nk),hs2(nk),hs3(nk),
     $     sg0(nk),sg1(nk),sg2(nk),sg3(nk), abd(4,nk),p1ip(4,nk)
C     local storage
c$$$  Naras fix, external routine not declared, added line below
      EXTERNAL bvalue

      double precision b0,b1,b2,b3,eps,vnikx(4,1),work(16),xv,bvalue,
     *     rss,tssy
      integer ld4,i,icoef,ileft,j,mflag
      logical fittoo
      fittoo= (method.ne.2)
      ileft = 1
      eps = .1d-10
      ld4=4
C Purpose : Solves the smoothing problem and computes the
C           criterion functions (CV and GCV).
C The coeficients of estimated smooth
      if(fittoo)then
         do 23084 i=1,nk
            do 23086 j=1,p
               coef(i,j) = xwy(i,j)
23086       continue
23084    continue
      endif
      do 23088 i=1,nk
         abd(4,i) = hs0(i)+lambda*sg0(i)
23088 continue
      do 23090 i=1,(nk-1)
         abd(3,i+1) = hs1(i)+lambda*sg1(i)
23090 continue
      do 23092 i=1,(nk-2)
         abd(2,i+2) = hs2(i)+lambda*sg2(i)
23092 continue
      do 23094 i=1,(nk-3)
         abd(1,i+3) = hs3(i)+lambda*sg3(i)
23094 continue
      call dpbfa(abd,ld4,nk,3,info)
      if(info.ne.0)then
         return
      endif
      if(fittoo)then
         do 23100 j=1,p
            call dpbsl(abd,ld4,nk,3,coef(1,j))
23100    continue
C     Value of smooths at the data points
         icoef = 1
         do 23102 i=1,n
            xv = x(i)
            do 23104 j=1,p
c$$$ Naras Fix
c$$$               sz(i,j) = bvalue(knot,nk+8,coef(1,j),nk,4,xv,0)
               sz(i,j) = bvalue(knot,coef(1,j),nk,4,xv,0)
23104       continue
23102    continue
         continue
      endif
C     Compute the criterion functions
C     Get Leverages First
      call sinrp2(abd,ld4,nk,p1ip)
      do 23106 i=1,n
         xv = x(i)
         ileft = interv(knot(1),(nk+1),xv, 0,0, ileft,mflag)
         if(mflag.eq.-1)then
            ileft = 4
            xv = knot(4)+eps
         endif
         if(mflag.eq.1)then
            ileft = nk
            xv = knot(nk+1)-eps
         endif
         j=ileft-3
         call bsplvd(knot,nk+8,4,xv,ileft,work,vnikx,1)
         b0=vnikx(1,1)
         b1=vnikx(2,1)
         b2=vnikx(3,1)
         b3=vnikx(4,1)
         lev(i) = (p1ip(4,j)*b0**2 + 2.*p1ip(3,j)*b0*b1 +
     1          2.*p1ip(2,j)*b0*b2 + 2.*p1ip(1,j)*b0*b3 +
     2           p1ip(4,j+1)*b1**2 + 2.*p1ip(3,j+1)*b1*b2 +
     3        2.*p1ip(2,j+1)*b1*b3 +    p1ip(4,j+2)*b2**2 +
     4        2.*p1ip(3,j+2)*b2*b3 +    p1ip(4,j+3)*b3**2 )*w(i)
23106 continue
C     Evaluate Criteria
      rss = 0d0
      df = 0d0
      sumw=0d0
      gcv=0d0
      cv=0d0
      do 23112 i=1,n
         df = df + lev(i)
23112 continue

      if(fittoo)then
         do 23116 i=1,n
            sumw = sumw + w(i)
            do 23118 j=1,p
               rss = rss + w(i)*wp(j)*(y(i,j)-sz(i,j))**2
               cv = cv +w(i)*wp(j)*((y(i,j)-sz(i,j))/(1-lev(i)))**2
23118       continue
23116    continue
         tssy=0d0
         do 23120 j=1,p
            tssy=tssy+wp(j)*ssy(j)
23120    continue
         gcv=((rss+tssy)/sumw)/((1d0-((dfoff+df-1)*cost+1)/sumw)**2)
C note: the weights should sum to n (the number of original observations)
         cv=(cv+tssy)/sumw
C     lev includes the weights
C     Note that this version of cv omits ALL observations at
C     tied x values, since the data are already collapsed here
      endif
      return
      end

      subroutine sinrp2(abd,ld4,nk,p1ip)
C Purpose :  Computes Inner Products between columns of L(-1)
C             where L = abd is a Banded Matrix with 3 subdiagonals
C                A refinement of Elden's trick is used.
C        Coded by Finbarr O'Sullivan

      implicit double precision(a-h,o-z)
      integer ld4,nk,i,j
      double precision abd(ld4,nk),p1ip(ld4,nk),
     *     wjm3(3),wjm2(2),wjm1(1),c0,c1,c2,c3
      wjm3(1)=0d0
      wjm3(2)=0d0
C     wjm3(1)=0d0   spotted by Brian Ripley 10/8/2015 (earlier actually)
      wjm3(3)=0d0
      wjm2(1)=0d0
      wjm2(2)=0d0
      wjm1(1)=0d0
      do 100 i=1,nk
         j=nk-i+1
         c0 = 1d0/abd(4,j)
         if(j.le.nk-3)then
            c1 = abd(1,j+3)*c0
            c2 = abd(2,j+2)*c0
            c3 = abd(3,j+1)*c0
         else if(j.eq.nk-2)then
            c1 = 0d0
            c2 = abd(2,j+2)*c0
            c3 = abd(3,j+1)*c0
         else if(j.eq.nk-1)then
            c1 = 0d0
            c2 = 0d0
            c3 = abd(3,j+1)*c0
c$$$ Naras fix: the condition (j.eq.nk) is superfluous even if nk = 1
c$$$         else if(j.eq.nk)then
c$$$            c1 = 0d0
c$$$            c2 = 0d0
c$$$            c3 = 0d0
         else 
            c1 = 0d0
            c2 = 0d0
            c3 = 0d0
         endif

         p1ip(1,j) = 0d0- (c1*wjm3(1)+c2*wjm3(2)+c3*wjm3(3))
         p1ip(2,j) = 0d0- (c1*wjm3(2)+c2*wjm2(1)+c3*wjm2(2))
         p1ip(3,j) = 0d0- (c1*wjm3(3)+c2*wjm2(2)+c3*wjm1(1))
         p1ip(4,j) = c0**2 +c1**2*wjm3(1)+2.*c1*c2*wjm3(2)+
     *        2.*c1*c3*wjm3(3)+ c2**2*wjm2(1)+2.*c2*c3*wjm2(2) +
     2        c3**2*wjm1(1)
         wjm3(1)=wjm2(1)
         wjm3(2)=wjm2(2)
         wjm3(3)=p1ip(2,j)
         wjm2(1)=wjm1(1)
         wjm2(2)=p1ip(3,j)
         wjm1(1)=p1ip(4,j)
 100  continue
      return
      end

      subroutine suff2(n,p,ny,match,y,w,ybar,wbar,ssy,work)
      integer match(n),n,ny,p,i,j
      double precision y(n,ny),ybar(p+1,ny),w(n),wbar(p+1),ssy(ny),
     *     work(n)
      double precision tsum

      call tpack(n,p,match,w,wbar)
      do 100 j=1,ny
         do 34 i=1,n
            work(i)=y(i,j)*w(i)
 34      continue
         call tpack(n,p,match,work,ybar(1,j))
         do 36 i=1,p
            if(wbar(i).gt.0d0)then
               ybar(i,j)=ybar(i,j)/wbar(i)
            else
               ybar(i,j)=0d0
            endif
 36      continue
         call untpack(n,p,match,ybar(1,j),work)
         tsum=0d0
         do 40 i=1,n
            tsum=tsum+ w(i)*(y(i,j)-work(i))**2
 40      continue
         ssy(j)=tsum
 100  continue
      return
      end

      subroutine sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,
     *     icenter,dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
     $     xrange,work,iwork, ier)
      integer n,p,nk,method,ier,nef, nefp1, n2,match(*),iwork(n)
      double precision x(n),y(n,p),w(n),knot(nk+4),tol,wp(p),dfoff,
     *     dfmax,cost,lambda,df,cv,gcv,coef(*),s(n,p),lev(nef),
     *     xrange(2), work(*)
C     workspace must be (2*p+2)*nefp1 + (p+16)*nk + n +p
      integer icenter
      logical center
      double precision xmiss,sigtol,xdiff,temp

      if(icenter.eq.0)then
         center=.false.
      else
         center=.true.
      endif
      if(nef.eq.0)then
C     match has not been initialized
         xmiss=1d20
         sigtol=1d-5
         call namat(x,match,n,nef,work,iwork,xmiss,sigtol)
         xrange(1)=work(1)
         xrange(2)=work(nef)
      else
         do 44 i=1,n
            work(match(i))=x(i)
 44      continue
      endif
      xdiff=xrange(2)-xrange(1)
      do 46 i=1,nef
         work(i)=(work(i)-xrange(1))/xdiff
 46   continue
      if(nk.eq.0)then
         call sknotl(work,nef,knot,nk)
         nk=nk-4
      endif
      if(dfmax .gt. dble(nk))then
         dfmax=dble(nk)
      endif
      if(cost.gt.0)then
c$$$ Naras fix
c$$$  if(center) then
c$$$            icenter = 1
c$$$         else
c$$$            icenter = 0
c$$$         endif
         temp=dble(n-icenter)/cost - dfoff
         if(dfmax.gt.temp)then
            dfmax=temp
         endif
      endif
      nefp1=nef+1
      n2=nefp1*(2*p+2)+1
c$$$ Naras fix
c$$$      call sspl1(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,center,
      call sspl1(y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,center,
     *     dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,
     *     work(1),work(nefp1+1),
c          xin,    yin
     $     work(nefp1*(p+1)+1),work(nefp1*(p+2)+1),
c          win,                sout
     *     work(n2),work(n2+p*nk),work(n2+(p+4)*nk),
c          xwy,     hs,           sg
     $     work(n2+(p+8)*nk),work(n2+(p+12)*nk),work(n2+(p+16)*nk),
     $     work(n2+(p+16)*nk+p), ier)
      return
      end

C Memory management subroutine
c$$$ Naras fix
c$$$      subroutine sspl1(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,
c$$$     *center,dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,xin,yin,win,
c$$$     *sout,xwy,hs,sg,abd,p1ip,ssy,work,ier)
c$$$      integer n,p,nefp1,nk,method,ier,match(n),nef
c$$$      double precision x(n),y(n,p),w(n),knot(nk+4),tol,wp(p),dfoff,
      subroutine sspl1(y,w,n,p,knot,nk,method,tol,wp,match,nef,nefp1,
     *center,dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,lev,xin,yin,win,
     *sout,xwy,hs,sg,abd,p1ip,ssy,work,ier)
      integer n,p,nefp1,nk,method,ier,match(n),nef
      double precision y(n,p),w(n),knot(nk+4),tol,wp(p),dfoff,
     *dfmax,cost,lambda,df,cv,gcv,coef(nk,p),s(n,p),lev(nef),xin(nefp1),
     *yin(nefp1,p),win(nefp1),sout(nefp1,p),xwy(nk,p),hs(nk,4),sg(nk,4),
     *abd(4,nk),p1ip(4,nk),ssy(p),work(n)
      logical center
      double precision sbar, wmean,tdfoff

      call suff2(n,nef,p,match,y,w,yin,win,ssy,work)
      if(center)then
         tdfoff=dfoff
         if(cost.gt.0)then
            tdfoff=dfoff-1/cost
         endif
      endif
      call sspl(xin,yin,win,nef,nefp1,p,knot,nk,method,tol,wp,ssy,
     *     tdfoff,dfmax,cost,lambda,df,cv,gcv,coef,sout,lev,xwy,
     $     hs,sg,abd, p1ip,ier)

C now unpack the results
      do 60 j=1,p
         call untpack(n,nef,match,sout(1,j),s(1,j))
         if(center) then
            sbar=wmean(nef,sout(1,j),win)
            do 64 i=1,n
               s(i,j)=s(i,j)-sbar
 64         continue
         endif
 60   continue

      if(center) df=df-1
      return
      end

      subroutine namat(x,match,n,nef,work,iwork,xmiss,tol)
C returns match (order) and work(1:nef) is the sorted unique x
      implicit double precision(a-h,o-z)
      integer match(*),n,nef,iwork(n),index
      double precision x(*),xmiss,work(*),tol,xend,xstart

      do 68 i=1,n
         work(i)=x(i)
         iwork(i)=i
 68   continue
      call sortdi(work,n,iwork,1,n)
      xstart=x(iwork(1))
      index=n
      xend=x(iwork(n))
c     while( . )
 70   if(xend .ge. xmiss .and. index .gt. 1)then
         index=index-1
         xend=x(iwork(index))
         goto 70
      endif

      tol=tol*(xend-xstart)
      index=1
      work(1)=xstart
c     for(i = 1:n) -- not allowed to increment i inside do-loop!
      i=1
 100  if(i .le. n) then
c        while() :
 75      if((x(iwork(i))-xstart).lt.tol) then
            match(iwork(i))=index
            i=i+1
            if(i.gt.n) goto 10
c      this next line was missing; added by TH in Nov/2017!            
            goto 75
         endif
         xstart= x(iwork(i))
         index=index+1
         match(iwork(i))=index
         work(index)=xstart
         i=i+1
         goto 100
      endif

 10   if(xstart .ge. xmiss)then
         nef=index-1
      else
         nef=index
      endif
      return
      end

      subroutine simfit(x,y,w,n,p,dfoff,cost,wp,gcv,coef,s,type,center,
     *     work)
C
C computes constant and linear fits, and selects the best using gcv
C
      implicit double precision (a-h,o-z)
      integer n,p,type
      double precision x(n),y(n,p),w(n),cost,dfoff,wp(p),gcv,coef(2,p),
     *     s(n,p),work(p)
      logical center
      double precision sx,sy,sumw, dcent,sxx,syy,sxy
C center is F for no centering, else T
C Note: if type enters 1 or 2, no selection is made
c$$$Naras fix: xbar is uninitialized; zero seems reasonable
      xbar = 0d0
      if(center) then
         icenter = 1
      else
         icenter = 0
      endif
      dcent=1-dble(icenter)
      sumw=0d0
      gcvc=0d0
      do 81 i=1,n
         sumw=sumw+w(i)
 81   continue

      if(type.ne.1)then
C     either 0 or 2 in which case the linear is needed as well
         sx=0.0
         sxx=0.0
         gcvl=0d0
         do 85 i=1,n
            sx=sx+w(i)*x(i)
 85      continue
         xbar=sx/sumw
         do 87 i=1,n
            sxx=sxx+w(i)*(x(i)-xbar)*x(i)
 87      continue
      endif

      do 100 j=1,p
         sy=0d0
         syy=0d0
         do 91 i=1,n
            sy=sy+w(i)*y(i,j)
 91      continue
         work(j)=sy/sumw
         do 93 i=1,n
            syy=syy+w(i)*(y(i,j)-work(j))*y(i,j)
 93      continue
         gcvc=gcvc+wp(j)*syy
         if(type.ne.1) then
C     once again, do for linear as well
            sxy=0.0
            do 97 i=1,n
               sxy=sxy+w(i)*(x(i)-xbar)*y(i,j)
 97         continue
            coef(2,j)=sxy/sxx
            gcvl=gcvl+wp(j)*(syy -sxy*coef(2,j))
         endif
 100  continue
      if(type.eq.0) then
         gcvc =gcvc/ (sumw* (1 - (dfoff*cost + dcent)/sumw)**2 )
         gcvl=gcvl/(sumw* (1 - (dcent + (dfoff +1)* cost)/sumw)**2)
         if(gcvc.le.gcvl) then
            type=1
            gcv=gcvc
         else
            type=2
            gcv=gcvl
         endif
      else if(type.eq.1) then
         gcv=gcvc/(sumw* (1 - (dfoff*cost + dcent)/sumw)**2 )
      else
         gcv=gcvl/(sumw* (1 - (dcent + (dfoff + 1)*cost)/sumw)**2)
      endif

      if(type.eq.1) then
         do 107 j=1,p
            coef(1,j)=work(j)*dcent
            coef(2,j)=0d0
            do 109 i=1,n
               s(i,j)=coef(1,j)
 109        continue
 107     continue
      else
         do 111 j=1,p
            coef(1,j)=work(j)*dcent-xbar*coef(2,j)
            do 113 i=1,n
               s(i,j)=coef(1,j)+coef(2,j)*x(i)
 113        continue
 111     continue
      endif
      return
      end

      subroutine sspl2(x,y,w,n,p,knot,nk,wp,match,nef,dfoff,dfmax,cost,
     *     lambda,df,gcv,coef,s,type,center,xrange,work,iwork,tol,ier)
C this routine selects from the linear and constant model as well
C see documentation for sspl
C workspace must be (2*p+2)*nefp1 + (p+16)*nk + 2*n
C if type>0 then no selection is performed; the fit is simply computed.
      integer n,p,nk,nef,type,match(n),ier,method,iwork(n)
      double precision x(n),y(n,p),w(n),knot(nk+4),wp(p),dfoff,dfmax,
     *     cost,lambda,df,gcv,coef(*),s(n,p),xrange(2),work(*),tol
      double precision coef1,coef2,cv
c$$$ Naras fix: gcv1 was not declared to be double precision!
      double precision gcv1
      logical center
c$$$ Naras fix: added icenter
      integer icenter
C     center is F for no centering, else T
c$$$ Naras fix: added icenter logic below
      if(center) then
         icenter = 1
      else
         icenter = 0
      endif
      if(type.eq.3)then
         method=1
         call sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,icenter,
     $        dfoff, dfmax,cost,lambda,df,cv,gcv,coef,s,work(1),
     $        xrange,work(n+1),iwork, ier)
         return
      endif

      if(type.gt.0)then
         call simfit(x,y,w,n,p,dfoff,cost,wp,gcv,coef,s,type,center,
     $        work)
c$$$ Naras fix: commented out
c$$$         if(center) then
c$$$            icenter = 1
c$$$         else
c$$$            icenter = 0
c$$$         endif
         df=dble(type) - dble(icenter)
         return
      endif

      method=3
      call sspl0(x,y,w,n,p,knot,nk,method,tol,wp,match,nef,icenter,
     *     dfoff,dfmax,cost,lambda,df,cv,gcv,coef,s,work(1),xrange,
     $     work(n+1),iwork, ier)
      gcv1=gcv
      call simfit(x,y,w,n,p,dfoff,cost,wp,gcv,work,work(2*p+1),type,
     *     center,work((n+2)*p+1))
      if(gcv.le.gcv1) then
C     the coef swapping is so as not to destroy the spline coefs if needed
         if(center) then
            icenter = 1
         else
            icenter = 0
         endif
         df=dble(type) - dble(icenter)
         do 30 j=1,p
            coef1=work(1+(j-1)*2)
            coef2=work(2+(j-1)*2)
            if(type.eq.1)then
               do 25 i=1,n
                  s(i,j)=coef1
 25            continue
            else
               do 27 i=1,n
                  s(i,j) =coef1+coef2*x(i)
 27            continue
            endif
            coef(1+(j-1)*2)=coef1
            coef(2+(j-1)*2)=coef2
 30      continue
      else
         type=3
         gcv=gcv1
      endif
      return
      end

      subroutine psspl2(x,n,p,knot,nk,xrange,coef,coefl,s,order,type)
      implicit double precision(a-h,o-z)
      integer n,p,nk, order, type
      double precision x(n),knot(nk+4),xrange(2),coef(nk,p),coefl(2,1),
     *     s(n,p)
c Var
      double precision ytemp

C switch(type){ case 1, 2, 3 } :
      if(type .eq. 1) then
         do 30 j=1,p
            if(order.ge.1)then
               ytemp=0d0
            else
               ytemp=coefl(1,j)
            endif
            do 35 i=1,n
               s(i,j)=ytemp
 35         continue
 30      continue

      else if(type .eq. 2) then

         if(order.ge.1)then
            do 40 j=1,p
               if(order.eq.1)then
                  ytemp=coefl(2,j)
               else
                  ytemp=0d0
               endif
               do 44 i =1,n
                  s(i,j)=ytemp
 44            continue
 40         continue
         else
            do 46 j=1,p
               do 48 i=1,n
                  s(i,j)=coefl(1,j)+coefl(2,j)*x(i)
 48            continue
 46         continue
         endif

      else if(type .eq. 3) then

         call psspl(x,n,p,knot,nk,xrange,coef,s,order)

      endif
      return
      end

      subroutine psspl(x,n,p,knot,nk,xrange,coef,s,order)
C     make predictions from a fitted smoothing spline, linear or constant
      implicit double precision(a-h,o-z)
      integer n,p,nk,order
      double precision x(n),knot(nk+4),xrange(2),coef(nk,p),s(n,p)
      double precision xcs,xmin,xdif, endv(2),ends(2),xends(2),stemp
c$$$  Naras fix, external routine not declared, added line below
      EXTERNAL bvalue
      double precision bvalue
      integer where

c$$$  Naras fix: In what follows, it appears that stemp is always set
c$$$  becasue order is one of 0,1,2. So initializing to zero should be ok
      stemp = 0d0
      if(order.gt.2.or.order.lt.0)  return

      xdif=xrange(2)-xrange(1)
      xmin=xrange(1)
      xends(1)=0d0
      xends(2)=1d0
      do 23253 j=1,p
         if(order.eq.0)then
c$$$ Naras fix
c$$$            endv(1)=bvalue(knot,nk+8,coef(1,j),nk,4,0d0,0)
c$$$            endv(2)=bvalue(knot,nk+8,coef(1,j),nk,4,1d0,0)
            endv(1)=bvalue(knot,coef(1,j),nk,4,0d0,0)
            endv(2)=bvalue(knot,coef(1,j),nk,4,1d0,0)
         endif
         if(order.le.1)then
c$$$ Naras fix
c$$$            ends(1)=bvalue(knot,nk+8,coef(1,j),nk,4,0d0,1)
c$$$            ends(2)=bvalue(knot,nk+8,coef(1,j),nk,4,1d0,1)
            ends(1)=bvalue(knot,coef(1,j),nk,4,0d0,1)
            ends(2)=bvalue(knot,coef(1,j),nk,4,1d0,1)
         endif
         do 23259 i=1,n
            xcs=(x(i)-xmin)/xdif
            if(xcs.lt.0d0)then
               where=1
            else if(xcs.gt.1d0)then
               where=2
            else
               where=0
            endif

            if(where.gt.0)then
C     beyond extreme knots
c     switch(order){ case 0: / 1 / 2  }  :
               if(order .eq. 0) then
                  stemp=endv(where)+(xcs-xends(where))*ends(where)
               else if(order .eq. 1) then
                  stemp=ends(where)
               else if(order .eq. 2) then
                  stemp=0d0
               endif
            else
c$$$ Naras fix
c$$$               stemp=bvalue(knot,nk+8,coef(1,j),nk,4,xcs,order)
               stemp=bvalue(knot,coef(1,j),nk,4,xcs,order)
            endif

            if(order.gt.0)then
               s(i,j)=stemp/(xdif**dble(order))
            else
               s(i,j)=stemp
            endif

23259    continue
23253 continue
      return
      end

mda <-
function(formula = formula(data), data = sys.frame(sys.parent()),
         subclasses = 3, sub.df = NULL, tot.df = NULL, dimension =
         sum(subclasses) - 1, eps = 100 * .Machine$double.eps, iter = 5,
         weights = mda.start(x, g, subclasses, trace, ...), method =
         polyreg, keep.fitted = (n * dimension < 5000), trace = FALSE,
         ...)
{
    this.call <- match.call()
    m <- match.call(expand.dots = FALSE)
    m[[1]] <- as.name("model.frame")
    m <- m[match(names(m), c("", "formula", "data"), 0)]
    m <- eval(m, sys.frame(sys.parent()))
    Terms <- attr(m, "terms")
    g <- model.extract(m, "response")
    x <- model.matrix(Terms, m)
    if(attr(Terms, "intercept"))x=x[,-1,drop=FALSE]
    dd <- dim(x)
    n <- dd[1]
    m <- dd[2]
    rowmin <- function(mat) {
        ncc <- ncol(mat)
        if (ncc == 1)
            return(drop(mat))
        rowm <- pmin(mat[, 1], mat[, 2])
        if (ncc == 2)
            return(rowm)
        else {
            for (j in seq(from = 3, to = ncc)) rowm <- pmin(rowm,
                mat[, j])
        }
        rowm
    }
    if (length(g) != n)
        stop("g should have length nrow(x)")
    fg <- factor(g)
    if (inherits(weights, "mda")) {
        if (is.null(weights$weights))
            weights <- predict(weights, x, type = "weights",
                g = fg)
        else weights <- weights$weights
    }
    subclasses <- sapply(weights, ncol)
    prior <- table(fg)
    dim(prior) <- NULL
    prior <- prior/sum(prior)
    cnames <- levels(fg)
    g <- as.numeric(fg)
    J <- length(cnames)
    Assign <- split(seq(sum(subclasses)), rep(seq(J), subclasses))
    names(Assign) <- cnames
    if (!is.null(tot.df)) {
        if (tot.df >= sum(subclasses))
            tot.df <- NULL
    }
    if (!is.null(sub.df)) {
        sub.df <- rep(sub.df, length = length(prior))
        sub.df <- pmin(sub.df, subclasses)
        if (all(sub.df == subclasses))
            sub.df <- NULL
    }
    for (counter in seq(iter)) {
        fit <- mda.fit(x, g, weights, assign.theta = Assign,
            Rj = subclasses, sub.df = sub.df, tot.df = tot.df,
            dimension = dimension, eps = eps,
            method = method, trace = trace, ...)
        dmat <- predict.fda(fit, type = "distance")
        sub.prior <- fit$sub.prior
        for (j in seq(J)) {
            TT <- dmat[g == j, Assign[[j]], drop = FALSE]
            TT <- exp(-0.5 * (TT - rowmin(TT)))
            TT <- TT * outer(rep(1, nrow(TT)), sub.prior[[j]])
            weights[[j]][] <- TT/drop(TT %*% rep(1, ncol(TT)))
        }
        pclass <- matrix(1, n, J)
        dmat <- exp(-0.5 * (dmat - rowmin(dmat)))
        for (j in seq(J)) {
            priorj <- sub.prior[[j]]
            ass <- Assign[[j]]
            TT <- dmat[, ass, drop = FALSE] * outer(rep(1, n), priorj)
            TTot <- drop(TT %*% rep(1, length(ass)))
            pclass[, j] <- prior[j] * TTot
        }
        pclass <- pclass/drop(pclass %*% rep(1, J))
        if (trace)
            cat("Iteration", counter, "\tDeviance(multinomial)",
                format(round(ll <- llmult(pclass, g), 5)), "\n")
    }
    if (!trace)
        ll <- llmult(pclass, g)
    if (!keep.fitted)
        fit$fit$fitted.values <- NULL
    dimnames(pclass) <- list(NULL, names(Assign))
    conf <- confusion(softmax(pclass), fg)
    fit <- c(fit, list(weights = weights, prior = prior, assign.theta = Assign,
        deviance = ll, confusion = conf, terms = Terms))
    fit$call <- this.call
    fit$sub.df <- sub.df
    fit$tot.df <- tot.df
    class(fit) <- c("mda", "fda")
    fit
}

set.seed(20101)

data(ESL.mixture)
mixture.train=data.frame(x=I(ESL.mixture$x),y=ESL.mixture$y)

fda.fit=fda(y~x,data=mixture.train)
print(coef(fda.fit))
mda.fit=mda(y~x,data=mixture.train)
print(coef(mda.fit))
mda.predict = predict(mda.fit,mixture.train$x)
print(mda.predict)
fitb=fda(y~x,data=mixture.train,method=bruto,cost=1)
fitm=fda(y~x,data=mixture.train,method=mars,cost=1)
fitb.confusion = confusion(fitb)
print(fitb.confusion)
fitm.confusion = confusion(fitm)
print(fitm.confusion)
mda.ppred=predict(mda.fit,newdata=ESL.mixture$xnew)
##print(mda.ppred)

objects  <- list(
    fda.fit = fda.fit,
    mda.fit = mda.fit,
    fitb = fitb,
    fitm = fitm,
    fitb.confusion = fitb.confusion,
    fitm.confusion = fitm.confusion,
    mda.ppred = mda.ppred)
##saveRDS(objects, "test_results/mda-0.4-results.RDS")

expected  <- readRDS("test_results/mda-0.4-results.RDS")
for (x in names(objects)) {
    cat(sprintf("Testing %s\n", x))
    if (x == "fitb") {
        expect_equal(objects[[x]], expected[[x]], tol = 1e-7)
    } else {
        expect_equal(objects[[x]], expected[[x]])
    }
}

fda <-
function (formula = formula(data), data = sys.frame(sys.parent()), 
    weights, theta, dimension = J - 1, eps = .Machine$double.eps, 
    method = polyreg, keep.fitted = (n * dimension < 5000), ...) 
{
    this.call <- match.call()
    m <- match.call(expand.dots = FALSE)
    m[[1]] <- as.name("model.frame")
    m <- m[match(names(m), c("", "formula", "data", "weights"), 
        0)]
    m <- eval(m, parent.frame())
    Terms <- attr(m, "terms")
    g <- model.extract(m, "response")
#    attr(Terms, "intercept") <- 0 This fails if a factor is in the model formula
    x <- model.matrix(Terms, m)
    if(attr(Terms, "intercept"))x=x[,-1,drop=FALSE]
    dd <- dim(x)
    n <- dd[1]
    weights <- model.extract(m, weights)
    if (!length(weights)) 
        weights <- rep(1, n)
    else if (any(weights < 0)) 
        stop("negative weights not allowed")
    if (length(g) != n) 
        stop("g should have length nrow(x)")
    fg <- factor(g)
    prior <- table(fg)
    prior <- prior/sum(prior)
    cnames <- levels(fg)
    g <- as.numeric(fg)
    J <- length(cnames)
    iswt <- FALSE
    if (missing(weights)) 
        dp <- table(g)/n
    else {
        weights <- (n * weights)/sum(weights)
        dp <- tapply(weights, g, sum)/n
        iswt <- TRUE
    }
    if (missing(theta)) 
        theta <- contr.helmert(J)
    theta <- contr.fda(dp, theta)
    Theta <- theta[g, , drop = FALSE]
    fit <- method(x, Theta, weights, ...)
    if (iswt) 
        Theta <- Theta * weights
    ssm <- t(Theta) %*% fitted(fit)/n
    ed <- svd(ssm, nu = 0)
    thetan <- ed$v
    lambda <- ed$d
    lambda[lambda > 1 - eps] <- 1 - eps
    discr.eigen <- lambda/(1 - lambda)
    pe <- (100 * cumsum(discr.eigen))/sum(discr.eigen)
    dimension <- min(dimension, sum(lambda > eps))
    if (dimension == 0) {
        warning("degenerate problem; no discrimination")
        return(structure(list(dimension = 0, fit = fit, call = this.call), 
                         class = "fda"))
    }
    thetan <- thetan[, seq(dimension), drop = FALSE]
    pe <- pe[seq(dimension)]
    alpha <- sqrt(lambda[seq(dimension)])
    sqima <- sqrt(1 - lambda[seq(dimension)])
    vnames <- paste("v", seq(dimension), sep = "")
    means <- scale(theta %*% thetan, FALSE, sqima/alpha)
    dimnames(means) <- list(cnames, vnames)
    names(lambda) <- c(vnames, rep("", length(lambda) - dimension))
    names(pe) <- vnames
    obj <- structure(list(percent.explained = pe, values = lambda, 
        means = means, theta.mod = thetan, dimension = dimension, 
        prior = prior, fit = fit, call = this.call, terms = Terms), 
        class = "fda")
    obj$confusion <- confusion(predict(obj), fg)
    if (!keep.fitted) 
        obj$fit$fitted.values <- NULL
    obj
}

df.gold <-
  function(x,df.target,tol=0.001,trace.it=FALSE){
    dof=function(x,lambda)sum(x^2/(x^2+lambda))
    goldnew =function(x1,x3){
      phi=(1+sqrt(5))/2#golden ratio
      (x3+phi*x1)/(1+phi)
    }
    if(df.target<1)stop("Cannot have less than 1 as L1 df.target")
    lambda1=0
###find a lambda3 that undercuts
    lambda3=0.001
    df3=dof(x,lambda3)
    df1=dof(x,0)
    while(df3>df.target){
      lambda3=10*lambda3
      df3=dof(x,lambda3)
    }
    lambda2=goldnew(0,lambda3)
    df2=dof(x,lambda2)
    crit = abs(df2-df.target)
    while( crit>tol){
      if(trace.it) cat("df2",df2,"lambda",lambda2,"\n")
      lambda0=goldnew(lambda2,lambda3)
      df0=dof(x,lambda0)
      crit0=abs(df0-df.target)
      if(crit0<crit){
        lambda1=lambda2
        df1=df2
        lambda2=lambda0
        df2=df0
        crit=crit0
      }
      else {
        lambda3=lambda1
        df3=df1
        lambda1=lambda0
        df1=df0
      }
    }
    lambda2
  }

c$$$Naras fixes
c$$$      subroutine marss(nx,n,p,nclass,y,x,w,tagx,maxorder,mmax,penalty,
c$$$     &   thresh,forwstep,interms,prune,bx,fullin,lenb, bestgcv, bestin, 
c$$$     &   flag,cut,dir,res,alpha,beta,scrat,iscrat,trace)
      subroutine marss(nx,n,p,nclass,y,x,w,tagx,maxorder,mmax,penalty,
     &   thresh,iforwstep,interms,iprune,bx,fullin,lenb,bestgcv,bestin,
     &   flag,cut,dir,res,beta,scrat,iscrat,itrace)
      implicit double precision (a-h,o-z)
      integer nx, n,p,nclass,tagx(nx,p),maxorder,mmax,bestin(mmax),
     &     flag(mmax,p),fullin(mmax)
c$$$Naras fixes
c$$$      double precision y(n,nclass),x(nx,p),w(n),bx(nx,mmax),bestgcv,
c$$$     &      cut(mmax,p),dir(mmax,p),res(nx,nclass),alpha(nclass),
c$$$     &      beta(mmax,nclass)
      double precision y(n,nclass),x(nx,p),w(n),bx(nx,mmax),bestgcv,
     &      cut(mmax,p),dir(mmax,p),res(nx,nclass),
     &      beta(mmax,nclass)
      double precision scrat(*)
      integer iscrat(*),itrace,iforwstep,iprune
      logical forwstep, prune, trace, tracec
      common tracec
      if(iforwstep.eq.0)then
         forwstep=.false.
      else
         forwstep=.true.
      endif
      if(iprune.eq.0)then
         prune=.false.
      else
         prune=.true.
      endif
      if(itrace.eq.0)then
         trace=.false.
      else
         trace=.true.
      endif
      tracec=trace
      len1=n*mmax
      len2=mmax
      len3=mmax*mmax
      len4=mmax*nclass
      len5=nclass
      len6=mmax
      len7=mmax
      len8=nclass
      len9=n
      len10=n*mmax
      len11=mmax*mmax
      len12=mmax*nclass
      len13=mmax*mmax
      len14=mmax*mmax
      n1=1
      n2=n1+len1
      n3=n2+len2
      n4=n3+len3
      n5=n4+len4
      n6=n5+len5
      n7=n6+len6
      n8=n7+len7
      n9=n8+len8
      n10=n9+len9
      n11=n10+len10
      n12=n11+len11
      n13=n12+len12
      n14=n13+len13
      n15=n14+len14
c$$$Naras fix
c$$$      call marsnew1(nx, n, p, nclass, y, x, w, tagx, maxorder, mmax, 
c$$$     &   bx, bestgcv, bestin, fullin, lenb, flag, cut, dir, res, 
c$$$     &   alpha, beta, penalty, thresh, forwstep, interms, prune, 
c$$$     &   scrat, scrat(n2), scrat(n3), scrat(n4), scrat(n5), scrat(n6), 
c$$$     &   scrat(n7), scrat(n8), scrat(n9), scrat(n10), scrat(n11), 
c$$$     &   scrat(n12), scrat(n13), scrat(n14), scrat(n15), iscrat, 
c$$$     &   iscrat(1+mmax), iscrat(1+2*mmax), iscrat(1+3*mmax))
      call marsnew1(nx, n, p, nclass, y,
     &     x, w, tagx, maxorder, mmax, 
     &     bx, bestgcv, bestin, fullin, lenb,
     &     flag, cut, dir, res, beta,
     &     penalty, thresh, forwstep, interms, prune, 
     &     scrat, scrat(n2), scrat(n3), scrat(n4), scrat(n5),
     &     scrat(n6), scrat(n8), scrat(n9), scrat(n10),
     &     scrat(n13), scrat(n14), scrat(n15),
     &     iscrat, iscrat(1+mmax), iscrat(1+2*mmax))
      return
      end

c$$$Naras fix
c$$$      subroutine marsnew1(nx, n, p, nclass, y, x, w, tagx, maxorder, 
c$$$     &   mmax, bx, bestgcv, bestin, fullin, lenb, flag, cut, dir, 
c$$$     &   res, alpha, beta, penalty,  thresh, forwstep, interms, 
c$$$     &   prune, bxorth, bxorthm, cov, covsy, ybar, scr1, scr5, scr6,  
c$$$     &   temp, bxsc, r, betasc, varsc, var, work, termlen, in,  
c$$$     &   tempin, qpivot)
c$$$      implicit double precision (a-h,o-z)
c$$$      integer n,nterms2,p,mmax,flag(mmax,p),tagx(nx,p),termlen(mmax), 
c$$$     &   nclass,fullin(mmax)
c$$$      double precision cov(mmax,mmax),covsy(mmax,nclass),critmax,
c$$$     &    x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),
c$$$     &    y(n,nclass),ybar(nclass),scr1(mmax),scr5(mmax),scr6(nclass)
c$$$      double precision temp(n),w(n), cut(mmax,p),dir(mmax,p),
c$$$     &    alpha(nclass),beta(mmax,nclass), bxsc(n,mmax), r(mmax,mmax), 
c$$$     &    dofit, res(nx,nclass),betasc(mmax,nclass), varsc(mmax,mmax), 
c$$$     &    var(mmax,mmax), stopfac, work(*)
      subroutine marsnew1(nx, n, p, nclass, y,
     &     x, w, tagx, maxorder, mmax,
     &     bx, bestgcv, bestin, fullin, lenb,
     &     flag, cut, dir, res, beta,
     &     penalty, thresh, forwstep, interms, prune, 
     &     bxorth, bxorthm, cov, covsy, ybar, 
     &     scr1, scr6, temp, bxsc,
     &     varsc, var, work, 
     &     termlen, tempin, qpivot)
      implicit double precision (a-h,o-z)
      integer n,nterms2,p,mmax,flag(mmax,p),tagx(nx,p),termlen(mmax), 
     &   nclass,fullin(mmax)
c$$$Naras fix
c$$$      double precision cov(mmax,mmax),covsy(mmax,nclass),critmax,
c$$$     &    x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),
c$$$     &    y(n,nclass),ybar(nclass),scr1(mmax),scr5(mmax),scr6(nclass)
c$$$      double precision temp(n),w(n), cut(mmax,p),dir(mmax,p),
c$$$     &    beta(mmax,nclass), bxsc(n,mmax), r(mmax,mmax), 
c$$$     &    dofit, res(nx,nclass),betasc(mmax,nclass), varsc(mmax,mmax), 
c$$$     &    var(mmax,mmax), stopfac, work(*)
      double precision cov(mmax,mmax),covsy(mmax,nclass),critmax,
     &    x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),
     &    y(n,nclass),ybar(nclass),scr1(mmax),scr6(nclass)
      double precision temp(n),w(n), cut(mmax,p),dir(mmax,p),
     &    beta(mmax,nclass), bxsc(n,mmax),
     &    dofit, res(nx,nclass), varsc(mmax,mmax), 
     &    var(mmax,mmax), stopfac, work(*)
      integer tempin(mmax), bestin(mmax),qrank, qpivot(mmax)
      logical forwstep,go, prune, newform, cvar, trace
      common trace
      double precision rtemp(4)
      integer itemp(4)
      tolbx=.01
      stopfac=10.0
      prevcrit=10d9
      if(.not.(interms.eq.1))goto 23000
      dofit=0
      goto 23001
23000 continue
      dofit=0
      do 23002 j=2,lenb 
      dofit=dofit+fullin(j)
23002 continue
      nterms=interms
23001 continue
      if(.not.(forwstep))goto 23004
      fullin(1)=1
      do 23006 i=2,mmax
      fullin(i)=0
23006 continue
      do 23008 i=1,n 
      w(i)=1
23008 continue
      do 23010 i=1, mmax
      termlen(i)=0
      do 23012 j=1, p
      flag(i,j)=0
      cut(i,j)=0
23012 continue
23010 continue
      nterms=1
      nterms2=2
      do 23014 i=1,n 
      bx(i,1)=1
      bxorth(i,1)=1.0/dsqrt(dfloat(n))
23014 continue
      bxorthm(1)=1/dsqrt(dfloat(n))
      do 23016 i=1,n 
      do 23018 j=1, mmax
      bx(i,j)=0.0
23018 continue
23016 continue
      do 23020 i=1,n 
      bx(i,1)=1
23020 continue
      do 23022 k=1, nclass
      ybar(k)=0.0
      do 23024 i=1,n 
      ybar(k)=ybar(k)+y(i,k)/n
23024 continue
23022 continue
      if(.not.(interms.eq.1))goto 23026
      rssnull=0.0
      do 23028 k=1, nclass
      do 23030 i=1,n 
      rssnull=rssnull+(y(i,k)-ybar(k))**2
23030 continue
23028 continue
      goto 23027
23026 continue
      rssnull=0.0
      do 23032 k=1, nclass
      do 23034 i=1,n 
      rssnull=rssnull+res(i,k)**2
23034 continue
23032 continue
23027 continue
      rss=rssnull
      cmm= (1+dofit) + penalty*(.5*dofit)
      gcvnull=(rssnull/n)/(1.0-cmm/n)**2
      if(.not.(trace))goto 23036
c$$$      call dblepr("initial rss=",11,rssnull,1)
23036 continue
      if(.not.(trace))goto 23038
c$$$      call dblepr("initial gcv=",11,gcvnull,1)
23038 continue
      lenb=1
      ii=interms-1
      go=.true.
23040 if(.not.( (ii.lt.(mmax-1)).and.((rss/rssnull).gt.thresh).and.go))
     &     goto 23041
      ii=ii+2
      do 23042 i1=1, nterms
      do 23044 i2=1, nterms
      cov(i1,i2)=0
23044 continue
23042 continue
      do 23046 j=1, nterms
      cov(j,j)=0.0
      do 23048 i=1,n 
      cov(j,j) = cov(j,j) + 
     %           (bxorth(i,j)-bxorthm(j)) * (bxorth(i,j)-bxorthm(j))
23048 continue
23046 continue
      do 23050 k=1,nclass
      do 23052 j=1, nterms
      covsy(j,k)=0.0
      do 23054 i=1,n 
      covsy(j,k)=covsy(j,k)+(y(i,k)-ybar(k))*bxorth(i,j)
23054 continue
23052 continue
23050 continue
      do 23056 ik=1,mmax 
      tempin(ik)=fullin(ik)
23056 continue

c$$$Naras fixes
c$$$      call addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,rss,prevcrit, 
c$$$     &     cov,covsy,y,ybar,x,tagx,w,termlen,mmax,tolbx, nterms,flag,
c$$$     &     maxorder,scr1,scr5,scr6,imax,jmax,kmax,critmax, newform,
c$$$     &     bxsc, r, betasc, temp)

      call addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,rss,prevcrit, 
     &     cov,covsy,y,ybar,x,tagx,termlen,mmax,tolbx, nterms,flag,
     &     maxorder,scr1,scr6,imax,jmax,kmax,critmax, newform,
     &     temp)
      doftemp=dofit
      doftemp=doftemp+1
      if(.not.((imax.gt.1).and.(newform)))goto 23058
      doftemp=doftemp+1
23058 continue
      temprss=rss-critmax
      cmm= (1+doftemp) + penalty*(.5*doftemp)
      gcv=(temprss/n)/(1.0-cmm/n)**2
      go=.false.
      if (.not.(((critmax/rss).gt.thresh).and.
     &          ((gcv/gcvnull).lt.stopfac))) goto 23060
      go=.true.
      dofit=doftemp
      rss=rss-critmax
      kk=tagx(imax,jmax)
c$$$256   format(" ","adding term"," jmax=",i3, "  imax=",i3 ,"  kmax=",i3, 
c$$$     &  "  critmax= ",f8.2,"  cutp=", f9.5," rss=",f8.2, " gcv=",f8.2, 
c$$$     &  " dofit=",f9.3)
      itemp(1)=jmax
      itemp(2)=imax
      itemp(3)=kmax
      rtemp(1)=critmax
      rtemp(2)=x(kk,jmax)
      rtemp(3)=rss
      rtemp(4)=gcv
      if(.not.(trace))goto 23062
c$$$Naras fixes
c$$$      call intpr("adding term ",12,ii,1)
23062 continue
      if(.not.(trace))goto 23064
c$$$      call intpr("var, sp index, parent",21,itemp,3)
23064 continue
      if(.not.(trace))goto 23066
c$$$      call dblepr("critmax cut rss gcv",19,rtemp,4)
23066 continue
      prevcrit=critmax
      do 23068 j=1,p 
      flag(ii,j)=flag(kmax,j)
      flag(ii+1,j)=flag(kmax,j)
      cut(ii,j)=cut(kmax,j)
      cut(ii+1,j)=cut(kmax,j)
      dir(ii,j)=dir(kmax,j)
      dir(ii+1,j)=dir(kmax,j)
23068 continue
      termlen(ii)=termlen(kmax)+1
      termlen(ii+1)=termlen(kmax)+1
      do 23070 i=1,n 
      temp(i)=x(tagx(i,jmax),jmax)
23070 continue
      temp1=temp(imax)
      fullin(ii)=1
      if(.not.((imax.gt.1).and.(newform)))goto 23072
      fullin(ii+1)=1 
23072 continue
      flag(ii,jmax)=1
      flag(ii+1,jmax)=1
      cut(ii,jmax)=temp1
      cut(ii+1,jmax)=temp1
      dir(ii,jmax)=1
      dir(ii+1,jmax)=-1
      if(.not.(fullin(ii+1).eq.0))goto 23074
      termlen(ii+1)=maxorder+1
23074 continue
      do 23076 i=1,n 
      if(.not.( (x(i,jmax)-temp1).gt.0))goto 23078
      bx(i,ii)=bx(i,kmax)*(x(i,jmax)-temp1)
23078 continue
      if(.not.((temp1-x(i,jmax)).ge.0))goto 23080
      bx(i,ii+1)=bx(i,kmax)*(temp1-x(i,jmax))
23080 continue
23076 continue
      if(.not.(nterms.eq.1))goto 23082
      temp1=0.0
      do 23084 i=1,n 
      temp1=temp1+bx(i,2)/n
23084 continue
      do 23086 i=1,n 
      bxorth(i,2)=bx(i,2)-temp1
23086 continue
      goto 23083
23082 continue
      call orthreg(n,n,nterms,bxorth,fullin, bx(1,ii),bxorth(1,nterms2))
23083 continue
      if(.not.(fullin(ii+1).eq.1))goto 23088
      call orthreg(n,n,nterms+1,bxorth,fullin, bx(1,ii+1),
     &       bxorth(1,nterms2+1))
      goto 23089
23088 continue
      do 23090 i=1,n 
      bxorth(i,nterms2+1)=0
23090 continue
23089 continue
      bxorthm(nterms2)=0.0 
      bxorthm(nterms2+1)=0.0
      do 23092 i=1,n 
      bxorthm(nterms2)=bxorthm(nterms2)+bxorth(i,nterms2)/n
      bxorthm(nterms2+1)=bxorthm(nterms2+1)+bxorth(i,nterms2+1)/n
23092 continue
      temp1=0.0
      temp2=0.0
      do 23094 i=1,n 
      temp1=temp1+bxorth(i,nterms2)**2
      temp2=temp2+bxorth(i,nterms2+1)**2 
23094 continue
      if(.not.(temp1.gt.0.0))goto 23096
      do 23098 i=1,n 
      bxorth(i,nterms2) =bxorth(i,nterms2)/dsqrt(temp1) 
23098 continue
23096 continue
      if(.not.(temp2.gt.0.0))goto 23100
      do 23102 i=1,n 
      bxorth(i,nterms2+1)=bxorth(i,nterms2+1)/dsqrt(temp2) 
23102 continue
23100 continue
      lenb=lenb+2
      nterms=nterms+2
      nterms2=nterms2+2
23060 continue
      goto 23040
23041 continue
      rtemp(1)=rss/rssnull
      rtemp(2)=critmax/rss
      rtemp(3)=gcv/gcvnull
      if(.not.(trace))goto 23104
c$$$      call dblepr("stopping forw step; rss crit and gcv ratios",43,
c$$$     &     rtemp,3)
23104 continue
      if(.not.(trace))goto 23106
      if(.not.((rss/rssnull).le.thresh))goto 23108
c$$$      call dblepr("rss ratio=",10,rss/rssnull,1)
23108 continue
      if(.not.((critmax/rss).le.thresh))goto 23110
c$$$      call dblepr ("crit ratio=",11,critmax/rss,1)
23110 continue
c$$$      call dblepr("critmax",7,critmax,1)
c$$$      call dblepr("rss",3,rss,1)
      if(.not.((gcv/gcvnull).gt.stopfac))goto 23112
c$$$      call dblepr("gcv ratio=",10,gcv/gcvnull,1)
23112 continue
23106 continue
23004 continue
      dofit= -1
      do 23114 i=1,nterms
      bestin(i)=fullin(i)
      dofit=dofit+fullin(i)
23114 continue
      if(.not.(trace))goto 23116
c$$$      call intpr("aft forw step",13,nterms,1)
23116 continue
      cvar=.false.
      call qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,bestin,y,qpivot,qrank,
     &     beta,res,rss,cvar,var,varsc,scr1, work)
c$$$Naras fix
c$$$      nt=dofit+1
      nt=int(dofit+1)
      if(.not.(qrank.lt. nt))goto 23118
      do 23120 i=qrank+1,nt
      bestin(qpivot(i))=0
      fullin(qpivot(i))=0
      dofit=dofit-1
23120 continue
23118 continue
      cvar=.true.
      rssfull=rss
      cmm= (1+dofit) + penalty*(.5*dofit)
      bestgcv=(rss/n)/(1.0-cmm/n)**2
      rtemp(1)=bestgcv
      rtemp(2)=rssfull
      rtemp(3)=dofit
      if(.not.(trace))goto 23122
c$$$      call dblepr("full model: gcv rss dofit",25,rtemp,3)
23122 continue
      if(.not.(trace))goto 23124
c$$$      call intpr("terms",5,fullin,lenb)
23124 continue
      if(.not.(prune))goto 23126
c Need var calculated to do drop-one calculations from t values.
      cvar=.true.
      call qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,tempin,y,qpivot,qrank,
     &     beta,res,rss,cvar,var,varsc,scr1,work)
      do 23128 i=1,mmax
      tempin(i)=bestin(i)
23128 continue
23130 if(.not.(dofit.gt.0 ))goto 23131
      jo=1
      rsstemp=10d99
      minterm=0
      do 23132 ii=2, lenb 
      if(.not.(tempin(ii).eq.1))goto 23134
      jo=jo+1
      temp7=0.0
      do 23136 kc=1,nclass
      temp7=temp7+beta(jo,kc)**2/var(jo,jo)
23136 continue
      if(.not.(temp7 .lt. rsstemp))goto 23138
      minterm=ii
      rsstemp=temp7
23138 continue
23134 continue
23132 continue
      rss=rss+rsstemp
      dofit=dofit-1
      cmm= (1.0+dofit) + penalty*(.5*dofit)
      gcv=(rss/n)/(1.0-cmm/n)**2
      tempin(minterm)=0
c$$$100   format(" ","pruning, minterm= ",i4, " gcv=",f9.3,2x, " rss=",f9.3,
c$$$     &     2x," dof=",f9.3," model= ",60(i1,1x))
      if(.not.(gcv.lt. bestgcv))goto 23140
      bestgcv=gcv
      do 23142 i=1,mmax 
      bestin(i)=tempin(i)
23142 continue
23140 continue
      if(.not.(dofit .gt. 0))goto 23144
      cvar=.true.
      call qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,tempin,y,qpivot,qrank,
     &     beta,res,rss,cvar,var,varsc,scr1,work)
23144 continue
      goto 23130
23131 continue
      call qrreg(nx,n,mmax,lenb,nclass,bx,bxsc,bestin,y,qpivot,qrank,
     &     beta,res,rss,cvar,var,varsc,scr1, work)
c$$$101   format(" ","best model gcv=",f9.3," rss=",f9.3,2x,"model= ",
c$$$     &             60(i1,1x))
      if(.not.(trace))goto 23146
c$$$      call intpr("best model",10,bestin,lenb)
23146 continue
      if(.not.(trace))goto 23148
c$$$      call dblepr(" gcv=",4,bestgcv,1)
23148 continue
23126 continue
      return
      end

c$$$Naras fixes
c$$$      subroutine addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,rss,
c$$$     &     prevcrit,cov,covsy,y,ybar,x,tagx,w,termlen,mmax,tolbx,
c$$$     &     nterms,flag, maxorder,scr1,scr5,scr6,imax,jmax,kmax,
c$$$     &     critmax, newform,bxsc,r, betasc, scrat)
c$$$      implicit double precision (a-h,o-z)
c$$$      integer n,nterms,nterms2,p,mmax,flag(mmax,p),v,tagx(nx,p),
c$$$     &        termlen(mmax), nclass, tempin(mmax), minspan, iendspan
c$$$      double precision cov(mmax,mmax),covsy(mmax,nclass),critmax,
c$$$     &     x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),
c$$$     &     y(n,nclass),ybar(nclass),scr1(mmax),scr5(mmax),scr6(nclass), 
c$$$     &     bxsc(n,mmax), r(mmax,mmax),betasc(mmax,nclass), scrat(n),
c$$$     &     w(n)
c$$$
      subroutine addtrm(nx,bx,tempin,bxorth,bxorthm,p,n,nclass,rss,
     &     prevcrit,cov,covsy,y,ybar,x,tagx,termlen,mmax,tolbx,
     &     nterms,flag, maxorder,scr1,scr6,imax,jmax,kmax,
     &     critmax, newform, scrat)
      implicit double precision (a-h,o-z)
      integer n,nterms,nterms2,p,mmax,flag(mmax,p),v,tagx(nx,p),
     &        termlen(mmax), nclass, tempin(mmax), minspan, iendspan
      double precision cov(mmax,mmax),covsy(mmax,nclass),critmax,
     &     x(nx,p),bx(nx,mmax),bxorth(n,mmax),bxorthm(mmax),
     &     y(n,nclass),ybar(nclass),scr1(mmax),scr6(nclass), 
     &     scrat(n)

      double precision temp1, temp2, scr2,sumb, sumbx, su, st, tem
      logical newform, tnewform, trace
      common trace
      critmax=0
      jmax=0
      imax=0
      kmax=0
c$$$  Naras fixes
      kk1 = 0
      k0 = 0
      kk = 0
      do 23150 m=1,nterms 
      nm=0
      do 23152 jjj=1,n 
      if(.not.(bx(jjj,m).gt.0))goto 23154
      nm=nm+1
23154 continue
23152 continue
      tem=-(1d0/(p*nm))*dlog(1d0 - 5d-2)
c$$$Naras fix
c$$$      minspan= -1d0*(dlog(tem)/dlog(2d0))/2.5
      minspan= int(-1d0*(dlog(tem)/dlog(2d0))/2.5)
      tem=(5d-2)/p
c$$$Naras fix
c$$$      iendspan=3d0-dlog(tem)/dlog(2d0)
      iendspan=int(3d0-dlog(tem)/dlog(2d0))
      if(.not.(termlen(m).lt. maxorder))goto 23156
      do 23158 v=1,p 
      if(.not.(flag(m,v).eq.0))goto 23160
      tnewform=.true.
      mm=1
23162 if(.not.((mm.le.nterms).and.tnewform))goto 23163
      mm=mm+1
      if(.not.(tempin(mm).eq.1))goto 23164
      tnewform=.false.
      if(.not.(flag(mm,v).ne.1))goto 23166
      tnewform=.true.
      go to 9911
23166 continue
      do 23168 j=1,p 
      if(.not.(j.ne.v))goto 23170
      if(.not.(flag(mm,j).ne.flag(m,j)))goto 23172
      tnewform=.true.
      go to 9911 
23172 continue
23170 continue
23168 continue
23164 continue
9911  continue
      goto 23162
23163 continue
      if(.not.(tnewform))goto 23174
      nterms2=nterms+1
      do 23176 i=1,n 
      scrat(i)=x(i,v)*bx(i,m)
23176 continue
      if(.not.(nterms.gt.1))goto 23178
      call orthreg(n,n,nterms,bxorth,tempin, scrat,bxorth(1,nterms2))
      goto 23179
23178 continue
      tem=0
      do 23180 i=1,n 
      tem=tem+scrat(i)/n
23180 continue
      do 23182 i=1,n
      bxorth(i,2)=scrat(i)-tem
23182 continue
23179 continue
      bxorthm(nterms2)=0.0 
      do 23184 i=1,n 
      bxorthm(nterms2)=bxorthm(nterms2)+bxorth(i,nterms2)/n
23184 continue
      temp1=0.0
      do 23186 i=1,n 
      temp1=temp1+bxorth(i,nterms2)**2
23186 continue
      if(.not.(temp1.gt.tolbx))goto 23188
      do 23190 i=1,n 
      bxorth(i,nterms2)=bxorth(i,nterms2)/dsqrt(temp1) 
23190 continue
      goto 23189
23188 continue
      do 23192 i=1,n 
      bxorth(i,nterms2)=0
23192 continue
      tnewform=.false.
23189 continue
      do 23194 i1=1, nterms2
      cov(i1,nterms2)=0.0
      cov(nterms2, i1)=0.0
23194 continue
      cov(nterms2,nterms2)=1
      do 23196 kc=1,nclass
      covsy(nterms2,kc)=0.0
      do 23198 i=1,n 
      covsy(nterms2,kc) = covsy(nterms2,kc)+(y(i,kc)-ybar(kc)) * 
     &                                               bxorth(i,nterms2)
23198 continue
23196 continue
      critnew=0.0
      do 23200 kc=1,nclass 
      temp1=0
      do 23202 i=1,n 
      temp1=temp1+y(i,kc)*bxorth(i,nterms2)
23202 continue
      critnew=critnew+temp1**2
23200 continue
      if(.not.(critnew.gt.critmax))goto 23204
      jmax=v
      critmax=critnew
      imax=1
      kmax=m
23204 continue
23174 continue
      if(.not.(tnewform))goto 23206
      nterms2=nterms+1
      nterms21=nterms+2
      goto 23207
23206 continue
      nterms2=nterms
      nterms21=nterms+1
      critnew=0.0
23207 continue
      do 23208 kc=1, nclass
      covsy(nterms21,kc)=0
23208 continue
      do 23210 ii=1,nterms21 
      cov(ii,nterms21)=0
      cov(nterms21,ii)=0
23210 continue
      do 23212 kc=1,nclass 
      scr6(kc)=0
23212 continue
      do 23214 ii=1,nterms21
      scr1(ii)=0
23214 continue
      scr2=0
      su=0
      st=0
      sumbx2=0
      sumb=0.0
      sumbx=0.0
      k=n-1
23216 if(.not.(k.gt.0))goto 23218
      do 23219 i=1,nterms2 
      kk=tagx(k,v)
      kk1=tagx(k+1,v)
      scr1(i)=scr1(i)+(bxorth(kk1,i)-bxorthm(i))*bx(kk1,m)
      cov(i,nterms21)=cov(i,nterms21)+ (x(kk1,v)-x(kk,v))*scr1(i)
      cov(nterms21,i)=cov(i,nterms21)
23219 continue
      scr2=scr2+(bx(kk1,m)**2)*x(kk1,v)
      sumbx2=sumbx2+bx(kk1,m)**2
      sumb=sumb+bx(kk1,m)
      sumbx=sumbx+bx(kk1,m)*x(kk1,v)
      su=st
      st=sumbx-sumb*x(kk,v)
      cov(nterms21,nterms21)= cov(nterms21,nterms21)+ (x(kk1,v)-x(kk,v))
     &     *(2*scr2-sumbx2*(x(kk,v)+x(kk1,v)))+ ( (su*su)-(st*st) )/n
      crittemp=critnew
      do 23221 kc=1, nclass
      scr6(kc)=scr6(kc)+(y(kk1,kc)-ybar(kc))*bx(kk1,m)
      covsy(nterms21,kc)=covsy(nterms21,kc )+(x(kk1,v)-x(kk,v))*scr6(kc)
      temp1=covsy(nterms21,kc)
      temp2=cov(nterms21,nterms21)
      do 23223 jk=1,nterms2 
      temp1=temp1-covsy(jk,kc)*cov(jk,nterms21)
      temp2=temp2-cov(jk,nterms21)*cov(jk,nterms21)
23223 continue
      if(.not.(cov(nterms21,nterms21).gt.0))goto 23225
      if(.not.((temp2/cov(nterms21,nterms21)) .gt. tolbx))goto 23227
      critadd=(temp1*temp1)/temp2
      goto 23228
23227 continue
      critadd=0.0
23228 continue
      goto 23226
23225 continue
      critadd=0
23226 continue
      crittemp=crittemp+critadd
      if(.not.(crittemp.gt.(1.01*rss)))goto 23229
      crittemp=0.0
23229 continue
      if(.not.(crittemp.gt.(2*prevcrit)))goto 23231
      crittemp=0.0
23231 continue
23221 continue
      if(.not.(k.gt.1))goto 23233
      k0=tagx(k-1,v)
23233 continue
      if(.not.((crittemp.gt.critmax) .and. 
     &         (mod(k,minspan).eq.0) .and.
     &               (k.ge.iendspan) .and.
     &           (k.le.(n-iendspan)) .and.
     &              (bx(kk1,m).gt.0) .and. 
     & (.not.( (k.gt.1) .and. (x(kk,v).eq.x(k0,v)))   ))) goto 23235
      jmax=v
      critmax=crittemp
      imax=k
      kmax=m
      newform=tnewform
23235 continue
      k=k-1
      goto 23216
23218 continue
23160 continue
      continue
23158 continue
23156 continue
23150 continue
      return
      end

shrink.mda <-
function (object, sub.df = NULL, tot.df = NULL, ...) 
{
    if (is.null(sub.df) & is.null(tot.df)) {
        warning("No shrinkage parameters (sub.df or tot.df) given")
        return(object)
    }
    dimension <- object$dimension
    lambda <- object$values[seq(dimension)]
    theta.mod <- object$theta.mod
    theta <- object$means
    alpha <- sqrt(lambda)
    sqima <- sqrt(1 - lambda)
    theta <- scale(theta, FALSE, alpha/sqima)
    sub.prior <- unlist(object$sub.prior)
    prior <- object$prior
    Rj <- sapply(object$assign.theta, length)
    dp <- sub.prior * rep(prior, Rj)
    cl <- rep(seq(Rj), Rj)
    P <- diag(dp) + transformPenalty(prior = dp, cl = cl, df = sub.df, 
        tot.df = tot.df)
    K <- t(theta) %*% P %*% theta
    TT <- chol((K + t(K))/2)
    Tinv <- backsolve(TT, diag(length(lambda)))
    M <- t(Tinv) %*% (lambda * Tinv)
    ed <- svd(M)
    thetan <- ed$v
    lambda <- ed$d
    discr.eigen <- lambda/(1 - lambda)
    pe <- (100 * cumsum(discr.eigen))/sum(discr.eigen)
    dimension <- min(dimension, sum(lambda > .Machine$double.eps))
    if (dimension == 0) {
        warning("degenerate problem; no discrimination")
        return(structure(list(dimension = 0, fit = object$fit,
                              call = object$call), 
                         class = "fda"))
    }
    thetan <- thetan[, seq(dimension), drop = FALSE]
    pe <- pe[seq(dimension)]
    alpha <- sqrt(lambda[seq(dimension)])
    sqima <- sqrt(1 - lambda[seq(dimension)])
    dm <- dimnames(theta)
    vnames <- dm[[2]][seq(dimension)]
    means <- scale(theta %*% Tinv %*% thetan, FALSE, sqima/alpha)
    dimnames(means) <- list(dm[[1]], vnames)
    names(lambda) <- c(vnames, rep("", length(lambda) - dimension))
    names(pe) <- vnames
    theta.mod <- theta.mod %*% Tinv %*% thetan
    object$confusion <- object$deviance <- NULL
    incl.names <- c("percent.explained", "values", "means", "theta.mod", 
        "dimension")
    rl <- list(pe, lambda, means, theta.mod, dimension)
    names(rl) <- incl.names
    object$sub.df <- sub.df
    object$tot.df <- tot.df
    object[incl.names] <- rl
    object$weights <- NULL
    update(object, sub.df = sub.df, tot.df = tot.df, weights = object, 
        ...)
}

plot.fda <-
  function (x, data, coords = c(1, 2), group = c("true",
                                            "predicted"), colors, pch, mcolors = colors, mpch,
            pcex = 0.5, mcex = 2.5, ...)
{
  object <- x                         # generic/method
  group <- match.arg(group)
  if (missing(data)) {
    vars <- predict(object, type = "var")
    g <- predict(object)
    group <- "predict"
  }
  else {
    if(group=="predicted"){
       vars <- predict(object, data, type = "var")
       g <- predict(object, data)
     }
    else{
      ff <- terms(object)
      attr(ff, "intercept") <- 0
      m <- model.frame(ff, data)
      x <- model.matrix(ff, m)
      vars <- predict(object, x, type = "var")
      g <- model.extract(m, "response")
    }
  }
  means <- object$means
  if(ncol(means)==1)stop("Only one canonical variate; plot requires at least two")
  g <- as.factor(g)
  cc <- as.numeric(g)
  np <- seq(levels(g))
 mit.colors=c(Orange = "#FF9233", Cyan = "#29D0D0", Lt.Green = "#81C57A",
Dk.Gray = "#575757", Red = "#AD2323", Blue = "#2A4BD7", Green = "#1D6914",
Brown = "#814A19", Purple = "#8126C0", Lt.Gray = "#A0A0A0", Yellow = "#FFEE33",
Pink = "#FFCDF3")
 if (missing(colors)) colors <- mit.colors
 colors <- rep(colors, length = length(np))
  if (missing(pch))
    pch <- paste(np)
  else pch <- rep(paste(pch), length = length(np))
  mcolors <- rep(mcolors, length = length(np))
  if (missing(mpch))
    mpch <- pch
  else mpch <- rep(paste(mpch), length = length(np))
  assign <- object$assign
  if (is.null(assign))
    assign <- split(seq(np), seq(np))
  if (!is.matrix(coords)) {
    coords <- matrix(coords, length(coords), length(coords))
    tt <- lower.tri(coords)
    coords <- cbind(t(coords)[tt], coords[tt])
  }
  for (ii in seq(nrow(coords))) {
    coord.pair <- coords[ii, ]
    plot(rbind(vars[, coord.pair], means[, coord.pair]),
         ..., type = "n", xlab = paste("Canonical Var",
                            coord.pair[1]), ylab = paste("Canonical Var",
                                              coord.pair[2]), main = paste("Discriminant Plot for",
                                                                group, "classes"))
    for (i in np) {
      which <- cc == i
      if (any(which))
        points(vars[which, coord.pair, drop = FALSE], col = colors[i],
               pch = pch[i], cex = pcex)
      points(means[assign[[i]], coord.pair, drop = FALSE],
             col = mcolors[i], pch = 1, cex = mcex)
      points(means[assign[[i]], coord.pair, drop = FALSE],
             col = mcolors[i], pch = mpch[i], cex = mcex/2)
    }
  }
  invisible()
}

confusion.list <-
function (object, truth, ...)
{
    dd <- names(object)
    y <- seq(dd)
    x <- attr(object, "dimension")
    if (!length(x)) 
        x <- seq(dd)
    for (i in y) {
        confi <- confusion(object[, i], truth)
        y[i] <- attr(confi, "error")
    }
    return(x, y)
}

predict.fda <-
function (object, newdata, type = c("class", "variates", "posterior", 
    "hierarchical", "distances"), prior, dimension = J - 1, ...)
{
    dist <- function(x, mean, m = ncol(mean)) (scale(x, mean, 
        FALSE)^2) %*% rep(1, m)
    type <- match.arg(type)
    means <- object$means
    Jk <- dim(means)
    J <- Jk[1]
    k <- Jk[2]
    if (type == "hierarchical") {
        if (missing(dimension)) 
            dimension.set <- seq(k)
        else {
            dimension.set <- dimension[dimension <= k]
            if (!length(dimension.set)) 
                dimension.set <- k
            dimension <- max(dimension.set)
        }
    }
    else dimension <- min(max(dimension), k)
    if (missing(newdata)) 
        y <- predict(object$fit)
    else {
        if (inherits(newdata, "data.frame") || is.list(newdata)) {
            Terms <- delete.response(terms(object))
#            attr(Terms, "intercept") <- 0
            newdata <- model.matrix(Terms, newdata)
            if(attr(Terms,"intercept"))newdata=newdata[,-1,drop=FALSE]
          }
        y <- predict(object$fit, newdata)
    }
    y <- y %*% object$theta[, seq(dimension), drop = FALSE]
    lambda <- object$values
    alpha <- sqrt(lambda[seq(dimension)])
    sqima <- sqrt(1 - lambda[seq(dimension)])
    newdata <- scale(y, FALSE, sqima * alpha)
    if (missing(prior)) 
        prior <- object$prior
    else {
        if (any(prior < 0) | round(sum(prior), 5) != 1) 
            stop("innappropriate prior")
    }
    means <- means[, seq(dimension), drop = FALSE]
    switch(type, variates = return(newdata), class = {
        n <- nrow(newdata)
        prior <- 2 * log(prior)
        mindist <- dist(newdata, means[1, ], dimension) - prior[1]
        pclass <- rep(1, n)
        for (i in seq(2, J)) {
            ndist <- dist(newdata, means[i, ], dimension) - prior[i]
            l <- ndist < mindist
            pclass[l] <- i
            mindist[l] <- ndist[l]
        }
        ## 2001-10-27: Need to provide levels or else if we get an error
        ## if the predicted classes do no contain all possible classes.
        ## Reported by Greg Jefferis <jefferis@stanford.edu>, fix by
        ## Bj/orn-Helge Mevik <bjorn-helge.mevik@matforsk.no>.
        return(factor(pclass, levels = seq(J),
                      labels = dimnames(means)[[1]]))
    }, posterior = {
        pclass <- matrix(0, nrow(newdata), J)
        for (i in seq(J)) pclass[, i] <- exp(-0.5 * dist(newdata, means[i, 
            ], dimension)) * prior[i]
        dimnames(pclass) <- list(dimnames(newdata)[[1]], dimnames(means)[[1]])
        return(pclass/drop(pclass %*% rep(1, J)))
    }, hierarchical = {
        prior <- 2 * log(prior)
        Pclass <- vector("list", length(dimension.set))
        names(Pclass) <- paste("D", dimension.set, sep = "")
        for (ad in seq(along = dimension.set)) {
            d <- dimension.set[ad]
            dd <- seq(d)
            mindist <- dist(newdata[, dd, drop = FALSE], means[1, dd, drop = FALSE], 
                d) - prior[1]
            pclass <- rep(1, nrow(newdata))
            for (i in seq(2, J)) {
                ndist <- dist(newdata[, dd, drop = FALSE], means[i, dd, 
                  drop = FALSE], d) - prior[i]
                l <- ndist < mindist
                pclass[l] <- i
                mindist[l] <- ndist[l]
            }
            levels(pclass) <- dimnames(means)[[1]]
            Pclass[[ad]] <- pclass
        }
        rownames <- dimnames(newdata)[[1]]
        if (is.null(rownames)) 
            rownames <- paste(seq(nrow(newdata)))
        return(structure(Pclass, class = "data.frame", row.names = rownames, 
            dimensions = dimension.set))
    }, distances = {
        dclass <- matrix(0, nrow(newdata), J)
        for (i in seq(J)) dclass[, i] <- dist(newdata, means[i, ], 
            dimension)
        dimnames(dclass) <- list(dimnames(newdata)[[1]], dimnames(means)[[1]])
        return(dclass)
    })
}

mspline <-
function(x, y, w = rep(1, n), df=5, lambda,
           thresh = 1e-04, ...)
{
    this.call <- match.call()
    y <- as.matrix(y)
    np <- ncol(y)
    n <- length(x)
    wp=rep(1/np,np)
    storage.mode(x) <- "double"
    storage.mode(y) <- "double"
    storage.mode(w) <- "double"
    storage.mode(wp) <- "double"

     nknotl <- function(n) {
            a1 <- log(50)/log(2)
            a2 <- log(100)/log(2)
            a3 <- log(140)/log(2)
            a4 <- log(200)/log(2)
            cx <- as.numeric(cut(n, c(0, 50, 200, 800, 3200)))
            if (is.na(cx)) 
                cx <- 5
            floor(switch(cx, n, 2^(a1 + ((a2 - a1) * (n - 50))/150), 
                2^(a2 + ((a3 - a2) * (n - 200))/600), 2^(a3 + 
                  ((a4 - a3) * (n - 800))/2400), 200 + (n - 3200)^0.2) + 
                6)
        }
    nkmax <- nknotl(n) - 4
    coef <- double(nkmax * np)
    smat <- matrix(0,n,np)
    storage.mode(smat) <- "double"
    nk <- as.integer(0)
    knot <- double(nkmax + 4)
    Match <- integer(n)
    nef <- as.integer(0)
    if(df==0){
      method=1
      if(missing(lambda))stop("need a lambda")
         }
      else {
        method =2
        lambda=0
      }
    storage.mode(method) <- "integer"
    storage.mode(df) <- "double"
    storage.mode(lambda) <- "double"
    xrange <- double(2)
     fit <-
        .Fortran("sspl0",
                 x,
                 y,
                 w,
                 as.integer(n),
                 as.integer(np),
                 knot = knot,
                 nk=nk,
                 method,
                 tol=as.double(thresh),
                 wp,
                 Match = Match,
                 nef = nef,
                 center=as.integer(0),
                 dfoff=as.double(0),
                 dfmax = as.double(df+2), 
                 cost = as.double(0),
                 lambda = lambda,
                 df = df,
                 cv=double(1),
                 gcv=double(1),
                 coef = coef,
                 s=smat,
                 lev=double(n),
                 xrange = xrange,
                 work=double((2 * np + 2) * ((n + 1) + 1) + (2 * np + 16) *(n + 1) + 2 * (n + 1) + np),
                 iwork=integer(n),
                 integer(1),
                 PACKAGE = "mda")[c("knot",  "nk", "Match",
                 "nef", "lambda", "df", "coef","s",
                 "xrange","lev")]
    fit$call <- this.call
    structure(fit, class = "mspline")
}

pplot4 <-
function (x, ...) 
{
    par(mfrow = c(3, 2))
    for (i in 1:3) {
        for (j in (i + 1):4) pplot(x[, c(i, j)], xlab = paste("var", 
            i), ylab = paste("var", j), ...)
    }
}

softmax <-
function (x, gap = FALSE) 
{
    d <- dim(x)
    maxdist <- x[, 1]
    pclass <- rep(1, d[1])
    for (i in seq(2, d[2])) {
        l <- x[, i] > maxdist
        pclass[l] <- i
        maxdist[l] <- x[l, i]
    }
    dd <- dimnames(x)[[2]]
    if (gap) {
        x <- abs(maxdist - x)
        x[cbind(seq(d[1]), pclass)] <- drop(x %*% rep(1, d[2]))
        gaps <- do.call("pmin", data.frame(x))
    }
    pclass <- if (is.null(dd) || !length(dd)) 
        pclass
    else factor(pclass, levels = seq(d[2]), labels = dd)
    if (gap) 
        list(class = pclass, gaps = gaps)
    else pclass
}

contr.fda <-
function (p = rep(1, d[1]), contrast.default = contr.helmert(length(p))) 
{
    d <- dim(contrast.default)
    sqp <- sqrt(p/sum(p))
    x <- cbind(1, contrast.default) * outer(sqp, rep(1, d[2] + 
        1))
    qx <- qr(x)
    J <- qx$rank
    qr.qy(qx, diag(d[1])[, seq(2, J)])/outer(sqp, rep(1, J - 
        1))
}

meanPenalty <-
function (prior, cl) 
{
    n <- length(prior)
    Q <- matrix(0, n, n)
    ll <- unique(cl)
    for (lll in ll) {
        which <- cl == lll
        pp <- prior[which]
        npp <- length(pp)
        pp <- diag(npp) - outer(rep(1, npp), pp/sum(pp))
        Q[which, which] <- t(pp) %*% pp
    }
    attr(Q, "cl") <- cl
    attr(Q, "prior") <- prior
    Q
}

confusion.default <-
  function (object, true, ...) 
{
  if (inherits(object, "data.frame")) 
    confusion.list(object, true)
  else {
    jt <- table(predicted=object, true)
    jd <- dimnames(jt)
    jn <- unlist(jd)
    ju <- jn[duplicated(jn)]
    j1 <- jd[[1]][!match(jd[[1]], ju, 0)]
    j2 <- jd[[2]][!match(jd[[2]], ju, 0)]
    jt <- jt[c(ju, j1), c(ju, j2), drop = FALSE]
    realjt <- unclass(jt[ju, ju, drop = FALSE])
    ntot <- sum(jt)
    mismatch <- (ntot - sum(realjt))/ntot
    structure(jt, error = (1 - sum(diag(realjt))/sum(realjt)), 
              mismatch = if (mismatch > 0) 
              mismatch
              else NULL)
  }
}

coef.fda <-
function (object, ...) 
{
    fit <- object$fit
    Coefs <- fit$coef
    if (is.null(Coefs)) 
        stop("No explicit coefficients in this formulation")
    dimension <- object$dimension
    Coefs <- Coefs %*% object$theta[, seq(dimension), drop = FALSE]
    lambda <- object$values
    alpha <- sqrt(lambda[seq(dimension)])
    sqima <- sqrt(1 - lambda[seq(dimension)])
    Coefs <- scale(Coefs, FALSE, sqima * alpha)
    attr(Coefs,"scaled:scale")=NULL
    Coefs
}

predict.mda <-
function (object, newdata, type = c("class", "variates", "posterior", 
    "hierarchical", "weights"), prior = NULL, dimension = R - 
    1, g, ...) 
{
    type <- match.arg(type)
    Rk <- dim(object$means)
    R <- Rk[1]
    k <- Rk[2]
    if (type == "hierarchical") {
        if (missing(dimension)) 
            dimension.set <- seq(k)
        else {
            dimension.set <- dimension[dimension <= k]
            if (!length(dimension.set)) 
                dimension.set <- k
            dimension <- max(dimension.set)
        }
        Pclass <- vector("list", length(dimension.set))
        names(Pclass) <- paste("D", dimension.set, sep = "")
        for (ad in seq(along = dimension.set)) {
            d <- dimension.set[ad]
            Pclass[[ad]] <- if (missing(newdata)) 
                Recall(object, prior = prior, dimension = d, 
                  ...)
            else Recall(object, newdata, prior = prior, dimension = d, 
                ...)
        }
        rownames <- names(Pclass[[1]])
        if (is.null(rownames)) 
            rownames <- paste(seq(along = Pclass[[1]]))
        return(structure(Pclass, class = "data.frame", row.names = rownames, 
            dimensions = dimension.set))
    }
    else dimension <- min(max(dimension), k)
    if (is.null(prior)) 
        prior <- object$prior
    else {
        if (any(prior < 0) | round(sum(prior), 5) != 1) 
            stop("innappropriate prior")
    }
    if (type == "variates") 
        return(NextMethod("predict"))
    rowmin <- function(mat) {
        ncc <- ncol(mat)
        if (ncc == 1) 
            return(drop(mat))
        rowm <- pmin(mat[, 1], mat[, 2])
        if (ncc == 2) 
            return(rowm)
        else {
            for (j in seq(from = 3, to = ncc)) rowm <- pmin(rowm, 
                mat[, j])
        }
        rowm
    }
    dmat <- if (missing(newdata)) 
        predict.fda(object, type = "distances", dimension = dimension, 
            ...)
    else predict.fda(object, newdata, type = "distances", dimension = dimension, 
        ...)
    Assign <- object$assign
    sub.prior <- object$sub.prior
    J <- length(Assign)
    if (type == "weights") {
        if (missing(newdata)) 
            return(object$weights)
        g <- as.numeric(g)
        weights <- Assign
        for (j in seq(J)) {
            TT <- dmat[g == j, Assign[[j]], drop = FALSE]
            TT <- exp(-0.5 * (TT - rowmin(TT)))
            TT <- TT * outer(rep(1, nrow(TT)), sub.prior[[j]])
            weights[[j]] <- TT/drop(TT %*% rep(1, ncol(TT)))
        }
        return(weights)
    }
    pclass <- matrix(1, nrow(dmat), J)
    dmat <- exp(-0.5 * (dmat - rowmin(dmat)))
    for (j in seq(J)) {
        TT <- dmat[, Assign[[j]], drop = FALSE]
        TT <- TT * outer(rep(1, nrow(TT)), sub.prior[[j]])
        pclass[, j] <- prior[j] * drop(TT %*% rep(1, ncol(TT)))
    }
    dimnames(pclass) <- list(NULL, names(Assign))
    switch(type, class = softmax(pclass), posterior = pclass/drop(pclass %*% 
        rep(1, J)))
}

kmeans.start <-
function (x, g, subclasses) 
{
    cnames <- levels(g <- factor(g))
    J <- length(cnames)
    g <- as.numeric(g)
    weights <- as.list(cnames)
    names(weights) <- cnames
    subclasses <- rep(subclasses, length = length(cnames))
    R <- sum(subclasses)
    cl <- rep(seq(J), subclasses)
    cx <- x[seq(R), , drop = FALSE]
    for (j in seq(J)) {
        nc <- subclasses[j]
        which <- cl == j
        xx <- x[g == j, , drop = FALSE]
        if ((nc <= 1) || (nrow(xx) <= nc)) {
            cx[which, ] <- apply(xx, 2, mean)
            wmj <- matrix(1, sum(g == j), 1)
        }
        else {
###            start <- xx[sample(1:nrow(xx), size = nc), ]
          start=nc
          TT <- kmeans(xx, start)
            cx[which, ] <- TT$centers
            wmj <- diag(nc)[TT$cluster, ]
        }
        dimnames(wmj) <-
            list(NULL, paste("s", seq(dim(wmj)[2]), sep = ""))
        weights[[j]] <- wmj
    }
    list(x = cx, cl = factor(cl, labels = cnames), weights = weights)
}

      double precision function wmean(n,y,w)
      integer n
      double precision y(n),w(n),wtot,wsum
      wtot=0e0
      wsum=0e0
      do 23028 i=1,n
      wsum=wsum+y(i)*w(i)
      wtot=wtot+w(i)
23028 continue
      if(.not.(wtot .gt. 0e0))goto 23030
      wmean=wsum/wtot
      goto 23031
23030 continue
      wmean=0e0
23031 continue
      return
      end

pplot <-
function (x, g, colors, pch, add = FALSE, type = "p", ...) 
{
    g <- as.factor(g)
    cc <- as.numeric(g)
    np <- seq(levels(g))
    if (missing(colors)) 
        colors <- np + 1
    else colors <- rep(colors, length = length(np))
    if (missing(pch)) 
        pch <- paste(np)
    else pch <- rep(pch, length = length(np))
    if (!add) 
        plot(x, type = "n", ...)
    for (i in unique(cc)) points(x[cc == i, , drop = FALSE], col = colors[i], 
        pch = pch[i], type = type)
}

bruto <-
function (x, y, w = rep(1, n), wp = rep(1/np, np), dfmax, cost = 2,
          maxit.select = 20, maxit.backfit = 20, thresh = 1e-04,
          trace.bruto = FALSE, start.linear = TRUE, fit.object, ...)
{
    this.call <- match.call()
    y <- as.matrix(y)
    x <- as.matrix(x)
    np <- ncol(y)
    d <- dim(x)
    nq <- d[2]
    n <- d[1]
    xnames <- dimnames(x)[[2]]
    if (!length(xnames))
        xnames <- NULL
    ynames <- dimnames(y)[[2]]
    if (!length(ynames))
        ynames <- NULL
    storage.mode(x) <- "double"
    storage.mode(y) <- "double"
    storage.mode(w) <- "double"
    storage.mode(wp) <- "double"
    storage.mode(cost) <- "double"
    if (missing(fit.object)) {
        nknotl <- function(n) {
            a1 <- log(50)/log(2)
            a2 <- log(100)/log(2)
            a3 <- log(140)/log(2)
            a4 <- log(200)/log(2)
            cx <- as.numeric(cut(n, c(0, 50, 200, 800, 3200)))
            if (is.na(cx))
                cx <- 5
            floor(switch(cx, n, 2^(a1 + ((a2 - a1) * (n - 50))/150),
                2^(a2 + ((a3 - a2) * (n - 200))/600), 2^(a3 +
                  ((a4 - a3) * (n - 800))/2400), 200 + (n - 3200)^0.2) +
                6)
        }
        check.range <- apply(x, 2, var)
        if (any(check.range < .Machine$double.eps))
            stop(paste("A column of x is constant;",
                       "do not include an intercept column"))
        nkmax <- nknotl(n) - 4
        coef <- matrix(double(nkmax * np * nq), ncol = nq)
        ybar <- apply(y * w, 2, sum)/sum(w)
        if (start.linear && (nq * cost > n))
            start.linear <- FALSE
        if (start.linear) {
            start.fit <- polyreg(x, y, w)
            eta <- fitted(start.fit)
            coef[seq(from = 2, by = 2, length = np), ] <-
                t(start.fit$coef)[, -1]
            type <- as.integer(rep(2, nq))
            df <- as.double(rep(1, nq))
        }
        else {
            eta <- outer(rep(1, n), ybar)
            type <- integer(nq)
            df <- double(nq)
        }
        nk <- integer(nq)
        knot <- matrix(double((nkmax + 4) * nq), ncol = nq)
        Match <- matrix(integer(n * nq), ncol = nq)
        nef <- integer(nq)
        lambda <- double(nq)
        xrange <- matrix(double(2 * nq), 2, nq)
        df <- double(nq)
        if (missing(dfmax))
            dfmax <- (2 * nkmax)/3
        if (length(dfmax) != nq)
            dfmax <- rep(dfmax, length = nq)
        if (cost > 0) {
            TD <- (n - sum(df))/cost
            TT <- dfmax > TD
            if (any(TT))
                dfmax[TT] <- TD
        }
        storage.mode(dfmax) <- "double"
    }
    else {
        this.call <- fit.object$call
        ybar <- fit.object$ybar
        nkmax <- fit.object$nkmax
        dfmax <- fit.object$dfmax
        eta <- fit.object$fitted.values
        if (is.null(eta))
            eta <- predict(fit.object, x)
        nk <- fit.object$nk
        knot <- fit.object$knot
        Match <- fit.object$Match
        nef <- fit.object$nef
        lambda <- fit.object$lambda
        coef <- fit.object$coef
        type <- unclass(fit.object$type)
        xrange <- fit.object$xrange
        maxit.select <- 0
        maxit.backfit <- fit.object$nit[2]
        df <- fit.object$df
    }
    maxit <- as.integer(c(maxit.select, maxit.backfit))
    names(df) <- xnames
    names(maxit) <- c("selection", "backfitting")
    gcv.select <- if (maxit.select)
        matrix(double(maxit.select * nq), nq, maxit.select)
    else double(1)
    gcv.backfit <- if (maxit.backfit)
        matrix(double(maxit.backfit * nq), nq, maxit.backfit)
    else double(1)
    df.select <- if (maxit.select)
        matrix(double(maxit.select * nq), nq, maxit.select)
    else double(1)
    names(lambda) <- xnames
    fit <-
        .Fortran("bruto",
                 x,
                 as.integer(n),
                 as.integer(nq),
                 y,
                 as.integer(np),
                 w,
                 knot = knot,
                 nkmax = as.integer(nkmax),
                 nk = nk,
                 wp,
                 Match = Match,
                 nef = nef,
                 dfmax = dfmax,
                 cost = cost,
                 lambda = lambda,
                 df = df,
                 coef = coef,
                 type = type,
                 xrange = xrange,
                 gcv.select = gcv.select,
                 gcv.backfit = gcv.backfit,
                 df.select = df.select,
                 maxit = maxit,
                 nit = maxit,
                 fitted.values = eta,
                 residuals = y - eta,
                 as.double(thresh),
                 double((2 * np + 2) * ((n + 1) + 1) + (2 * np + 16) *
                        (n + 1) + 2 * (n + 1) + np),
                 integer(n),
                 as.integer(trace.bruto),
                 PACKAGE = "mda")[c("knot", "nkmax", "nk", "Match",
                 "nef", "dfmax", "cost", "lambda", "df", "coef", "type",
                 "xrange", "gcv.select", "gcv.backfit", "df.select",
                 "maxit", "nit", "fitted.values", "residuals")]
    if (TN <- fit$nit[1]) {
        TT <- fit$gcv.select[, seq(TN), drop = FALSE]
        dimnames(TT) <- list(xnames, NULL)
    }
    else TT <- NULL
    fit$gcv.select <- TT
    if (TN) {
        TT <- fit$df.select[, seq(TN), drop = FALSE]
        dimnames(TT) <- list(xnames, NULL)
    }
    else TT <- NULL
    fit$df.select <- TT
    if (TN <- fit$nit[2]) {
        TT <- fit$gcv.backfit[, seq(TN), drop = FALSE]
        dimnames(TT) <- list(xnames, NULL)
    }
    else TT <- NULL
    fit$gcv.backfit <- TT
    TT <- factor(fit$type, levels = 1:3, labels = c("excluded",
        "linear", "smooth"))
    names(TT) <- xnames
    fit$type <- TT
    fit$ybar <- ybar
    fit$call <- this.call
    structure(fit, class = "bruto")
}

mda.means <-
function (object, x, y) 
{
    weights <- means <- object$weights
    nn <- names(object$weights)
    for (i in nn) {
        xx <- x[y == i, ]
        ww <- weights[[i]]
        nc <- ncol(ww)
        xm <- matrix(0, ncol(x), nc)
        for (j in seq(nc)) {
            www <- ww[, j]
            www <- www/sum(www)
            xm[, j] <- apply(xx * www, 2, sum)
        }
        means[[i]] <- xm
    }
    means
}

mda.start <-
function (x, g, subclasses = 3, trace.mda.start = FALSE, start.method = c("kmeans", 
    "lvq"), tries = 5, criterion = c("misclassification", "deviance"), 
    ...) 
{
    start.method <- match.arg(start.method)
    criterion <- match.arg(criterion)
    name.criterion <- switch(criterion, misclassification = "Misclassification Error", 
        deviance = "Deviance(multinomial)")
    starter <- get(paste(start.method, "start", sep = "."), mode = "function")
    fg <- factor(g)
    cnames <- levels(fg)
    prior <- table(fg)
    J <- length(cnames)
    n <- length(g)
    g <- as.numeric(fg)
    best.ll <- 1/.Machine$double.eps
    for (try in seq(tries)) {
        start <- starter(x, fg, subclasses)
        weights <- start$weights
        if (criterion == "misclassification") {
            pg <- lvqtest(start, x)
            ll <- attr(confusion(pg, fg), "error")
        }
        else {
            subclasses <- sapply(weights, ncol)
            Assign <- split(seq(sum(subclasses)), rep(seq(J), 
                subclasses))
            names(Assign) <- cnames
            fit <- mda.fit(x, g, weights, assign.theta = Assign, 
                Rj = subclasses, eps = .Machine$double.eps, method = polyreg, 
                ...)
            dmat <- exp(-0.5 * predict.fda(fit, type = "distance"))
            sub.prior <- fit$sub.prior
            pclass <- matrix(1, n, J)
            for (j in seq(J)) {
                priorj <- sub.prior[[j]]
                ass <- Assign[[j]]
                TT <- dmat[, ass, drop = FALSE]
                TT <- TT * outer(rep(1, n), priorj)
                TTot <- drop(TT %*% rep(1, length(ass)))
                wmj <- TT[g == j, , drop = FALSE]/TTot[g == j]
                pclass[, j] <- prior[j] * TTot
                dimnames(wmj) <- list(NULL, paste("s", seq(along = ass), 
                  sep = ""))
                weights[[j]] <- wmj
            }
            pclass <- pclass/drop(pclass %*% rep(1, J))
            ll <- llmult(pclass, g)
        }
        if (trace.mda.start) 
            cat(start.method, "start   \t", name.criterion, format(round(ll, 
                5)), "\n")
        if (ll < best.ll) {
            keep.weights <- weights
            best.ll <- ll
        }
    }
    structure(keep.weights, criterion = best.ll, name.criterion = name.criterion)
}

transformPenalty <-
function (Q, prior, cl, df = NULL, tot.df = NULL) 
{
    if (missing(Q)) 
        Q <- meanPenalty(prior, cl)
    if (missing(prior)) 
        prior <- attr(Q, "prior")
    if (missing(cl)) 
        cl <- attr(Q, "cl")
    transform.pen <- function(Q, prior, df) {
        df.inv <- function(d, df, lambda = NULL, iterations = 10) {
            if (is.null(lambda)) {
                lambda <- 0.1
                while (sum(1/(1 + d * lambda)) >= df) lambda <- lambda * 
                  2
            }
            df.diriv <- function(d, lambda) -sum((d * lambda)/(1 + 
                d * lambda)^2)
            current.df <- sum(1/(1 + d * lambda))
            if (abs((df - current.df)/df) < 1e-04 | iterations == 
                1) 
                return(list(lambda = lambda, df = current.df))
            else {
                lambda <- exp(log(lambda) - (current.df - df)/df.diriv(d, 
                  lambda))
                Recall(d, df, lambda, iterations - 1)
            }
        }
        pQp <- Q/outer(sqrt(prior), sqrt(prior))
        d <- svd(pQp)$d
        lambda <- df.inv(d, df)$lambda
        lambda * Q
    }
    if (!is.null(tot.df)) {
        if (tot.df >= length(prior)) 
            return(Q * 0)
        else return(transform.pen(Q, prior, tot.df))
    }
    else {
        ncl <- unique(cl)
        df <- rep(df, length = length(ncl))
        for (i in seq(along = ncl)) {
            which <- cl == ncl[i]
            Q[which, which] <- Recall(Q[which, which, drop = FALSE], 
                prior[which], tot.df = df[i])
        }
        return(Q)
    }
}

laplacian <-
function (size = 16, compose = FALSE) 
{
    gmat <- matrix(0, size, size)
    xx <- seq(size)
    for (v in xx) gmat[, v] <- sqrt(2/size) * cos(((xx - 0.5) * 
        pi * (v - 1))/size)
    gmat[, 1] <- gmat[, 1]/sqrt(2)
    lvec <- -(2 * size^2) * (1 - cos(((xx - 1) * pi)/size))
    gmat <- kronecker(gmat, gmat)
    lvec <- rep(lvec, rep(size, size)) + rep(lvec, size)
    if (compose) 
        gmat %*% (lvec^2 * t(gmat))
    else list(vectors = gmat, values = lvec^2)
}

      subroutine bruto(x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,dfmax,
     *     cost,lambda,df,coef,type,xrange,gcvsel,gcvbak,dfit,maxit,nit,
     *     eta,resid,thresh,work,iwork,itrace)
      implicit double precision(a-h,o-z)
      integer n,q,p,nkmax,nk(q),match(n,q),nef(q),type(q),maxit(2),
     *     nit(2),iwork(n),itrace
      double precision x(n,q),y(n,p),w(n),knot(nkmax+4,q),wp(p),
     *     dfmax(q),cost,lambda(q),df(q),coef(nkmax*p,q),xrange(2,q),
     *     gcvsel(q,*),gcvbak(q,*),dfit(q,*),eta(n,p),resid(n,p),
     *     thresh,work(*)
      logical trace, select
      if(itrace.eq.0)then
         trace=.false.
      else
         trace=.true.
      endif

      do 01 j=1,p
         do 02 i=1,n
            resid(i,j)=y(i,j)-eta(i,j)
 02      continue
 01   continue

      select=.true.
c$$$ Naras fix
c$$$      call bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,dfmax,
c$$$     *cost,lambda,df,coef,type,xrange,gcvsel,dfit,maxit(1),nit(1),eta,
c$$$     *resid,thresh*10d0,work,iwork,trace)
c$$$      select=.false.
c$$$      call bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,dfmax,
c$$$     *cost,lambda,df,coef,type,xrange,gcvbak,dfit,maxit(2),nit(2),eta,
c$$$     *resid,thresh,work,iwork,trace)

      call bakssp(select,x,n,q,p,w,knot,nkmax,nk,wp,match,nef,dfmax,
     *cost,lambda,df,coef,type,xrange,gcvsel,dfit,maxit(1),nit(1),eta,
     *resid,thresh*10d0,work,iwork,trace)
      select=.false.
      call bakssp(select,x,n,q,p,w,knot,nkmax,nk,wp,match,nef,dfmax,
     *cost,lambda,df,coef,type,xrange,gcvbak,dfit,maxit(2),nit(2),eta,
     *resid,thresh,work,iwork,trace)
      do 04 j=1,p
         do 06 i=1,n
            eta(i,j)=y(i,j)-resid(i,j)
 06      continue
 04   continue
      return
      end

c$$$ Naras fix      
c$$$      subroutine bakssp(select,x,n,q,y,p,w,knot,nkmax,nk,wp,match,nef,
c$$$     *     dfmax,cost,lambda,df,coef,type,xrange,gcv,dfit,maxit,nit,s,
c$$$     *     resid,thresh,work,iwork,trace)
c$$$      implicit double precision(a-h,o-z)
c$$$      integer n,q,p,nkmax,nk(q),match(n,q),nef(q),type(q),maxit,nit,ier,
c$$$     *     ntype,iwork(n)
c$$$      double precision x(n,q),y(n,p),w(n),knot(nkmax+4,q),wp(p),dfmax(q)
c$$$     *     ,cost,lambda(q),df(q),coef(nkmax*p,q),xrange(2,q),
c$$$     *     gcv(q,maxit),dfit(q,maxit),s(n,p),resid(n,p),thresh,work(*)
c$$$
      subroutine bakssp(select,x,n,q,p,w,knot,nkmax,nk,wp,match,nef,
     *     dfmax,cost,lambda,df,coef,type,xrange,gcv,dfit,maxit,nit,s,
     *     resid,thresh,work,iwork,trace)
      implicit double precision(a-h,o-z)
      integer n,q,p,nkmax,nk(q),match(n,q),nef(q),type(q),maxit,nit,ier,
     *     ntype,iwork(n)
      double precision x(n,q),w(n),knot(nkmax+4,q),wp(p),dfmax(q)
     *     ,cost,lambda(q),df(q),coef(nkmax*p,q),xrange(2,q),
     *     gcv(q,maxit),dfit(q,maxit),s(n,p),resid(n,p),thresh,work(*)
      double precision dfoff, gcv0, ndf,gcv1,gcvrat,ndfoff,wmean,sbar,
     *     rss,tol
      logical center, select, trace
      center=.true.
      tol=1d-3
      rss=0d0
      do 08 j=1,p
         sbar=wmean(n,resid(1,j),w)
         do 10 i=1,n
            resid(i,j)=resid(i,j) - sbar
            work(i)=resid(i,j)**2
 10      continue
         rss=rss+wp(j)*wmean(n,work,w)
 08   continue
      dfoff=0
      do 12 k=1,q
         dfoff=dfoff+df(k)
 12   continue
      gcv1=rss/((1-(1+dfoff*cost)/n)**2)
      gcvrat=1d0
      nit=0
 14   if(.not.(nit.lt.maxit.and.gcvrat .gt.thresh ))goto 15
      gcv0=gcv1
      nit=nit+1
      do 16 k=1,q
         gcv(k,nit)=gcv1
         if(.not.select.and.type(k).eq.1)goto 16

         if(type(k) .gt. 1) then
            call psspl2(x(1,k),n,p,knot(1,k),nk(k),xrange(1,k),
     *           coef(1,k),coef(1,k),s,0,type(k))
            do 22 j=1,p
               sbar=wmean(n,s(1,j),w)
               do 24 i=1,n
                  resid(i,j)=resid(i,j) + s(i,j)-sbar
 24            continue
 22         continue
         endif

         ndfoff=dfoff-df(k)
         if(select) then
            ntype=0
         else
            ntype=type(k)
         endif

         call sspl2(x(1,k),resid,w,n,p,knot(1,k),nk(k),wp,match(1,k),
     *        nef(k),ndfoff,dfmax(k),cost,lambda(k),ndf,gcv1,coef(1,k),
     *        s,ntype,center,xrange(1,k),work,iwork,tol,ier)
         gcv(k,nit)=gcv1
         if(select) then
            dfit(k,nit)=ndf
            df(k)=ndf
            dfoff=ndfoff+ndf
            type(k)=ntype
         endif

         if(type(k) .gt. 1) then
            do 32 j=1,p
               do 34 i=1,n
                  resid(i,j)=resid(i,j) - s(i,j)
 34            continue
 32         continue
         endif

 16   continue

      gcvrat=dabs(gcv1-gcv0)/gcv0
      if(trace) then
         call intpr("outer iteration",15,nit,1)
         call dblepr("gcv  ",5,gcv1,1)
         call dblepr("ratio",5,gcvrat,1)
      endif
      continue
      goto 14
 15   continue
      return
      end

      subroutine pbruto(x,n,q,ybar,p,knot,nkmax,nk,coef,type,xrange,eta,
     *     work)
      implicit double precision(a-h,o-z)
      integer n,q,p,nkmax,nk(q),type(q)
      double precision x(n,q),ybar(p),knot(nkmax+4,q),coef(nkmax*p,q),
     *     xrange(2,q),eta(n,p),work(n,p)

      do 38 j=1,p
         do 40 i =1,n
            eta(i,j)=ybar(j)
 40      continue
 38   continue
      do 42 k=1,q
         if(type(k) .eq .1) goto 42
         call psspl2(x(1,k),n,p,knot(1,k),nk(k),xrange(1,k),coef(1,k),
     *        coef(1,k),work,0,type(k))
         do 46 j=1,p
            do 48 i=1,n
               eta(i,j)=eta(i,j) + work(i,j)
 48         continue
 46      continue
 42   continue
      return
      end

model.matrix.mars <-
function (object, x, which = object$selected.terms, full = FALSE, 
    ...) 
{
    if (missing(x)) 
        return(object$x)
    x <- as.matrix(x)
    dd <- dim(x)
    n <- dd[1]
    p <- dd[2]
    nterms <- length(which)
    dir <- object$factor
    cut <- object$cuts
    if (full) {
        bx <- matrix(0, nrow = n, ncol = object$lenb)
        bx[, 1] <- 1
    }
    else bx <- matrix(1, nrow = n, ncol = nterms)
    which <- which[-1]
    for (i in seq(along = which)) {
        j <- which[i]
        if (all(dir[j, ] == 0)) {
            stop("error in factor or which")
        }
        temp1 <- 1
        for (k in 1:p) {
            if (dir[j, k] != 0) {
                temp2 <- dir[j, k] * (x[, k] - cut[j, k])
                temp1 <- temp1 * temp2 * (temp2 > 0)
            }
        }
        if (full) 
            bx[, j] <- temp1
        else bx[, i + 1] <- temp1
    }
    bx
}

      subroutine qrreg(nx,n,px,p,nclass,x,xsc,in,y,qpivot,qrank,beta,
     &     res,rss,cvar,var,varsc,scr1,work)
      implicit double precision (a-h,o-z)
      integer nx,n,p,px, qpivot(p),qrank,nclass,in(p)
      double precision x(nx,p), xsc(n,p), y(n,nclass),res(nx,nclass),
     &     beta(px,nclass),work(*),scr1(p),var(px,p),varsc(px,p)
      logical cvar
      ii=0
      do 23000 j=1,p 
      if(.not.(in(j).eq.1))goto 23002
      ii=ii+1
      do 23004 i=1,n 
      xsc(i,ii)=x(i,j)
23004 continue
23002 continue
23000 continue
      nt=ii
      ijob=101
      info=1
      temp3=1d-2
      do 23006 i=1,p 
      qpivot(i)=i
23006 continue
      call dqrdc2(xsc,n,n,nt,temp3,qrank,scr1,qpivot,work)
      rss=0.0
      do 23008 k=1,nclass
      call dqrsl(xsc,n,n,qrank,scr1,y(1,k),work(1),work(1),beta(1,k),
     &     work(1),res(1,k),ijob,info)
      do 23010 i=1,n 
      res(i,k)=y(i,k)-res(i,k)
      rss=rss+res(i,k)*res(i,k)
23010 continue
23008 continue
      if(.not.(cvar))goto 23012
c$$$ Naras fix
c$$$      call calcvar(nx,n,px,xsc,qrank,qpivot,var,varsc,work)
      call calcvar(nx,px,xsc,qrank,var,varsc)
23012 continue
      return
      end

confusion.fda <-
function (object, data, ...) 
{
    if (missing(data)) 
        return(object$confusion)
    Terms <- terms(object)
    attr(Terms, "intercept") <- 0
    m <- model.frame(Terms, data)
    x <- model.matrix(Terms, m)
    g <- model.extract(m, "response")
    confusion.default(predict(object, x, ...), g)
}

predict.bruto <-
function (object, newdata, type = c("fitted", "terms"), ...)
{
    if (missing(newdata)) {
        z <- fitted(object)
        if (is.null(z)) 
            stop("need to supply newdata")
        else return(z)
    }
    d <- as.integer(dim(newdata))
    type <- match.arg(type)
    nq <- d[2]
    n <- d[1]
    if (nq != length(object$df)) 
        stop(paste("newdata should have the same number of columns",
                   "as the df component of object"))
    ybar <- object$ybar
    np <- as.integer(length(ybar))
    eta <- matrix(double(n * np), n, np)
    Type <- as.numeric(object$type)
    storage.mode(Type) <- "integer"
    storage.mode(newdata) <- "double"
    if (type == "fitted") {
        .Fortran("pbruto", newdata, n, nq, ybar, np, object$knot, object$nkmax, 
            object$nk, object$coef, Type, object$xrange, eta = eta, 
            eta, PACKAGE = "mda")$eta
    }
    else {
        ob <- as.list(seq(nq))
        names(ob) <- dimnames(newdata)[[2]]
        knot <- object$knot
        nk <- object$nk
        xrange <- object$xrange
        coef <- object$coef
        fitm <- matrix(double(n * np), n, np)
        dimnames(fitm) <- list(dimnames(newdata)[[1]], names(ybar))
        for (i in seq(nq)) {
            if (Type[i] > 1) 
                fit <- .Fortran("psspl2", newdata[, i], n, np, knot[, 
                  i], nk[i], xrange[, i], coef[, i], coef[, i], 
                  fit = fitm, as.integer(0), Type[i], PACKAGE = "mda")$fit
            else fit <- fitm
            ob[[i]] <- list(x = newdata[, i], y = fit)
        }
        ob
    }
}

      subroutine sortdi (v,lenv,a,ii,jj)
c
c     puts into a the permutation vector which sorts v into
c     increasing order.  only elements from ii to jj are considered.
c     arrays iu(k) and il(k) permit sorting up to 2**(k+1)-1 elements
c     v is returned sorted
c     this is a modification of cacm algorithm #347 by r. c. singleton,
c     which is a modified hoare quicksort.
c
      integer lenv
      integer t,tt,ii,jj,iu(20),il(20)
      integer a(jj)
      double precision v(lenv), vt, vtt
      m=1
      i=ii
      j=jj
 10   if (i.ge.j) go to 80
 20   k=i
      ij=(j+i)/2
      t=a(ij)
      vt=v(ij)
      if (v(i).le.vt) go to 30
      a(ij)=a(i)
      a(i)=t
      t=a(ij)
      v(ij)=v(i)
      v(i)=vt
      vt=v(ij)
 30   l=j
      if (v(j).ge.vt) go to 50
      a(ij)=a(j)
      a(j)=t
      t=a(ij)
      v(ij)=v(j)
      v(j)=vt
      vt=v(ij)
      if (v(i).le.vt) go to 50
      a(ij)=a(i)
      a(i)=t
      t=a(ij)
      v(ij)=v(i)
      v(i)=vt
      vt=v(ij)
      go to 50
 40   a(l)=a(k)
      a(k)=tt
      v(l)=v(k)
      v(k)=vtt
 50   l=l-1
      if (v(l).gt.vt) go to 50
      tt=a(l)
      vtt=v(l)
 60   k=k+1
      if (v(k).lt.vt) go to 60
      if (k.le.l) go to 40
      if (l-i.le.j-k) go to 70
      il(m)=i
      iu(m)=l
      i=k
      m=m+1
      go to 90
 70   il(m)=k
      iu(m)=j
      j=l
      m=m+1
      go to 90
 80   m=m-1
      if (m.eq.0) return
      i=il(m)
      j=iu(m)
 90   if (j-i.gt.10) go to 20
      if (i.eq.ii) go to 10
      i=i-1
 100  i=i+1
      if (i.eq.j) go to 80
      t=a(i+1)
      vt=v(i+1)
      if (v(i).le.vt) go to 100
      k=i
 110  a(k+1)=a(k)
      v(k+1)=v(k)
      k=k-1
      if (vt.lt.v(k)) go to 110
      a(k+1)=t
      v(k+1)=vt
      go to 100
      end

      subroutine orthreg(nx,n,p,x,in, y,res)
      implicit double precision (a-h,o-z)
      integer n,nx,p, in(p)
      double precision x(nx,p),y(n),res(n)
      do 23000 i=1,n 
      res(i)=y(i)
23000 continue
      do 23002 j=1,p 
      if(.not.(in(j).eq.1))goto 23004
      temp1=0
      temp2=0
      do 23006 i=1,n 
      temp1=temp1+res(i)*x(i,j)
      temp2=temp2+x(i,j)*x(i,j)
23006 continue
      beta=temp1/temp2
      do 23008 i=1,n 
      res(i)=res(i)-beta*x(i,j)
23008 continue
23004 continue
23002 continue
      return
      end

print.fda <-
function (x, ...) 
{
    if (!is.null(cl <- x$call)) {
        cat("Call:\n")
        dput(cl)
    }
    cat("\nDimension:", format(x$dimension), "\n")
    cat("\nPercent Between-Group Variance Explained:\n")
    print(round(x$percent, 2))
    error <- x$confusion
    df <- x$fit
    if (!is.null(df)) 
        df <- df$df
    if (!is.null(df)) {
        cat("\nDegrees of Freedom (per dimension):", format(sum(df)), 
            "\n")
    }
    if (!is.null(error)) {
        n <- as.integer(sum(error))
        error <- format(round(attr(error, "error"), 5))
        cat("\nTraining Misclassification Error:", error, "( N =", 
            n, ")\n")
    }
    invisible(x)
}

\name{glass}
\alias{glass}
\title{Glass Identification Database}
\description{
  The \code{glass} data frame has 214 observations and 10 variables,
  representing glass fragments.
}
\usage{data(glass)}
\format{
  This data frame contains the following columns:
  \describe{
    \item{RI}{refractive index}
    \item{Na}{weight percent in corresponding oxide}
    \item{Mg}{weight percent in corresponding oxide}
    \item{Al}{weight percent in corresponding oxide}
    \item{Si}{weight percent in corresponding oxide}
    \item{K}{weight percent in corresponding oxide}
    \item{Ca}{weight percent in corresponding oxide}
    \item{Ba}{weight percent in corresponding oxide}
    \item{Fe}{weight percent in corresponding oxide}
    \item{Type}{Type of glass:
      \describe{
	\item{1}{building\_windows\_float\_processed,}
	\item{2}{building\_windows\_non\_float\_processed,}
	\item{3}{vehicle\_windows\_float\_processed,}
	\item{4}{vehicle\_windows\_non\_float\_processed (none in this
	  database),}
	\item{5}{containers,}
	\item{6}{tableware,}
	\item{7}{headlamps}
      }
    }
  }
}
\source{
  P. M. Murphy and D. W. Aha (1999),
  UCI Repository of Machine Learning Databases,
  \url{http://archive.ics.uci.edu/ml/datasets/glass+identification}
}
\keyword{datasets}

\name{mars}
\alias{mars}
\title{Multivariate Adaptive Regression Splines}
\description{
  Multivariate adaptive regression splines.
}
\usage{
mars(x, y, w, wp, degree, nk, penalty, thresh, prune, trace.mars,
     forward.step, prevfit, \dots)
}
\arguments{
  \item{x}{a matrix containing the independent variables.}
  \item{y}{a vector containing the response variable, or in the case of
    multiple responses, a matrix whose columns are the response values
    for each variable.}
  \item{w}{an optional vector of observation weights (currently ignored).}
  \item{wp}{an optional vector of response weights.}
  \item{degree}{an optional integer specifying maximum interaction
    degree (default is 1).}
  \item{nk}{an optional integer specifying the maximum number of model
    terms.}
  \item{penalty}{an optional value specifying the cost per degree of
    freedom charge (default is 2).}
  \item{thresh}{an optional value specifying forward stepwise stopping
    threshold (default is 0.001).}
  \item{prune}{an optional logical value specifying whether the model
    should be pruned in a backward stepwise fashion (default is
    \code{TRUE}).}
  \item{trace.mars}{an optional logical value specifying whether info
    should be printed along the way (default is \code{FALSE}).}
  \item{forward.step}{an optional logical value specifying whether
    forward stepwise process should be carried out (default is
    \code{TRUE}).}
  \item{prevfit}{optional data structure from previous fit.  To see the
    effect of changing the penalty parameter, one can use prevfit with
    \code{forward.step = FALSE}.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  An object of class \code{"mars"}, which is a list with the following
  components:

  \item{call}{call used to \code{mars}.}
  \item{all.terms}{term numbers in full model.  \code{1} is the constant
    term.  Remaining terms are in pairs (\code{2 3}, \code{4 5}, and so
    on).  \code{all.terms} indicates nonsingular set of terms.}
  \item{selected.terms}{term numbers in selected model.}
  \item{penalty}{the input penalty value.}
  \item{degree}{the input degree value.}
  \item{thresh}{the input threshold value.}
  \item{gcv}{gcv of chosen model.}
  \item{factor}{matrix with \eqn{ij}-th element equal to 1 if term
    \eqn{i} has a factor of the form \eqn{x_j > c}, equal to \eqn{-1} if
    term \eqn{i} has a factor of the form \eqn{x_j \le c}, and to 0 if
    \eqn{xj} is not in term \eqn{i}.}
  \item{cuts}{matrix with \eqn{ij}-th element equal to the cut point
    \eqn{c} for variable \eqn{j} in term \eqn{i}.}
  \item{residuals}{residuals from fit.}
  \item{fitted}{fitted values from fit.}
  \item{lenb}{length of full model.}
  \item{coefficients}{least squares coefficients for final model.}
  \item{x}{a matrix of basis functions obtained from the input x
    matrix.}
}
\note{
  This function was coded from scratch, and did not use any of
  Friedman's mars code.  It gives quite similar results to  Friedman's
  program in our tests, but not exactly the same results.  We have not 
  implemented Friedman's anova decomposition nor are categorical
  predictors handled properly yet.  Our version does handle multiple
  response variables, however.  }
\author{
  Trevor Hastie and Robert Tibshirani
}
\references{
  J. Friedman, ``Multivariate Adaptive Regression Splines'' (with
  discussion) (1991). 
  \emph{Annals of Statistics}, \bold{19}/1, 1--141.
}
\seealso{
  \code{\link{predict.mars}},
  \code{\link{model.matrix.mars}}.

  Package \pkg{earth} also provides multivariate adaptive regression
  spline models based on the Hastie/Tibshirani mars code in package
  \pkg{mda}, adding some extra features. It can be used in the
  \code{method}
  argument of \code{fda} or \code{mda}.
}
\examples{
data(trees)
fit1 <- mars(trees[,-3], trees[3])
showcuts <- function(obj)
{
  tmp <- obj$cuts[obj$sel, ]
  dimnames(tmp) <- list(NULL, names(trees)[-3])
  tmp
}
showcuts(fit1)

## examine the fitted functions
par(mfrow=c(1,2), pty="s")
Xp <- matrix(sapply(trees[1:2], mean), nrow(trees), 2, byrow=TRUE)
for(i in 1:2) {
  xr <- sapply(trees, range)
  Xp1 <- Xp; Xp1[,i] <- seq(xr[1,i], xr[2,i], len=nrow(trees))
  Xf <- predict(fit1, Xp1)
  plot(Xp1[ ,i], Xf, xlab=names(trees)[i], ylab="", type="l")
}
}
% Converted by Sd2Rd version 0.3-3.
\keyword{smooth}

\name{mda}
\alias{mda}
\alias{print.mda}
\title{Mixture Discriminant Analysis}
\description{
  Mixture discriminant analysis.
}
\usage{
mda(formula, data, subclasses, sub.df, tot.df, dimension, eps,
    iter, weights, method, keep.fitted, trace, \dots)
}
\arguments{
  \item{formula}{of the form \code{y~x} it describes the response and
    the predictors.  The formula can be more complicated, such as
    \code{y~log(x)+z} etc (see \code{\link{formula}} for more details).
    The response should be a factor representing the response variable,
    or any vector that can be coerced to such (such as a logical
    variable).}
  \item{data}{data frame containing the variables in the formula
    (optional).}
  \item{subclasses}{Number of subclasses per class, default is 3.  Can be
    a vector with a number for each class.}
  \item{sub.df}{If subclass centroid shrinking is performed, what is the
    effective degrees of freedom of the centroids per class.  Can be a
    scalar, in which case the same number is used for each class, else a
    vector.}
  \item{tot.df}{The total df for all the centroids can be specified
    rather than separately per class.}
  \item{dimension}{The dimension of the reduced model.  If we know our
    final model will be confined to a discriminant subspace (of the
    subclass centroids), we can specify this in advance and have the EM
    algorithm operate in this subspace.}
  \item{eps}{A numerical threshold for automatically truncating the
    dimension.}
  \item{iter}{A limit on the total number of iterations,  default is 5.}
  \item{weights}{\emph{NOT} observation weights!  This is a special
    weight structure, which for each class assigns a weight (prior
    probability) to each of the observations in that class of belonging
    to one of the subclasses.  The default is provided by a call to
    \code{mda.start(x, g, subclasses, trace, \dots)} (by this time
    \code{x} and \code{g} are known).  See the help for
    \code{\link{mda.start}}.  Arguments for \code{mda.start} can be
    provided via the \code{\dots} argument to mda, and the
    \code{weights} argument need never be accessed.  A previously fit
    mda object can be supplied, in which case the final subclass
    \code{responsibility} weights are used for \code{weights}.  This 
    allows the iterations from a previous fit to be continued.}
  \item{method}{regression method used in optimal scaling.  Default is
    linear regression via the function \code{polyreg}, resulting in the
    usual mixture model.  Other possibilities are \code{mars} and 
    \code{bruto}.  For penalized mixture discriminant models
    \code{gen.ridge} is appropriate.}
  \item{keep.fitted}{a logical variable, which determines whether the
    (sometimes large) component \code{"fitted.values"} of the \code{fit}
    component of the returned \code{mda} object should be kept.  The
    default is \code{TRUE} if \code{n * dimension < 5000}.} 
  \item{trace}{if \code{TRUE}, iteration information is printed.  Note
    that the deviance reported is for the posterior class likelihood,
    and not the full likelihood, which is used to drive the EM algorithm
    under \code{mda}.  In general the latter is not available.}
  \item{\dots}{additional arguments to \code{mda.start} and to
    \code{method}.}
}
\value{
  An object of class \code{c("mda", "fda")}.  The most useful extractor
  is \code{predict}, which can make many types of predictions from this
  object.  It can also be plotted, and any functions useful for fda
  objects will work here too, such as \code{confusion} and \code{coef}.

  The object has the following components:
  \item{percent.explained}{the percent between-group variance explained
    by each dimension (relative to the total explained.)}
  \item{values}{optimal scaling regression sum-of-squares for each
    dimension (see reference).}
  \item{means}{subclass means in the discriminant space.  These are also
    scaled versions of the final theta's or class scores, and can be
    used in a subsequent call to \code{mda} (this only makes sense if
    some columns of theta are omitted---see the references)}
  \item{theta.mod}{(internal) a class scoring matrix which allows
    \code{predict} to work properly.}
  \item{dimension}{dimension of discriminant space.}
  \item{sub.prior}{subclass membership priors, computed in the fit.  No
    effort is currently spent in trying to keep these above a threshold.}
  \item{prior}{class proportions for the training data.}
  \item{fit}{fit object returned by \code{method}.}
  \item{call}{the call that created this object (allowing it to be
    \code{update}-able).}
  \item{confusion}{confusion matrix when classifying the training data.}
  \item{weights}{These are the subclass membership probabilities for
    each member of the training set; see the weights argument.}
  \item{assign.theta}{a pointer list which identifies which elements of
    certain lists belong to individual classes.}
  \item{deviance}{The multinomial log-likelihood of the fit.  Even though
    the full log-likelihood drives the iterations, we cannot in general
    compute it because of the flexibility of the \code{method} used.
    The deviance can increase with the iterations, but generally does not.}

  The \code{method} functions are required to take arguments \code{x}
  and \code{y} where both can be matrices, and should produce a matrix
  of \code{fitted.values} the same size as \code{y}.  They can take
  additional arguments \code{weights} and should all have a \code{\dots}
  for safety sake.  Any arguments to method() can be passed on via the
  \code{\dots} argument of \code{mda}.  The default method
  \code{polyreg} has a \code{degree} argument which allows polynomial
  regression of the required total degree.  See the documentation for
  \code{\link{predict.fda}} for further requirements of \code{method}.
  The package \code{earth} is suggested for this package as well;
  \code{earth} is a more detailed implementation of the mars model, and
  works as a \code{method} argument.
  
  The function \code{mda.start} creates the starting weights; it takes
  additional arguments which can be passed in via the \code{\dots}
  argument to \code{mda}.  See the documentation for \code{mda.start}.
}
\author{
  Trevor Hastie and Robert Tibshirani
}
\seealso{
  \code{\link{predict.mda}},
  \code{\link{mars}},
  \code{\link{bruto}},
  \code{\link{polyreg}},
  \code{\link{gen.ridge}},
  \code{\link{softmax}},
  \code{\link{confusion}}
  %%\code{\link{coef.fda}},
  %%\code{\link{plot.fda}}
}
\references{
  ``Flexible Disriminant Analysis by Optimal Scoring'' by Hastie,
  Tibshirani and Buja, 1994, JASA, 1255-1270.

  ``Penalized Discriminant Analysis'' by Hastie, Buja and Tibshirani, 1995,
  Annals of Statistics, 73-102
    
  ``Discriminant Analysis by Gaussian Mixtures'' by Hastie and
  Tibshirani, 1996, JRSS-B, 155-176.
  
  ``Elements of Statisical Learning - Data Mining, Inference and
  Prediction'' (2nd edition, Chapter 12) by Hastie, Tibshirani and
  Friedman, 2009, Springer
}
\examples{
data(iris)
irisfit <- mda(Species ~ ., data = iris)
irisfit
## Call:
## mda(formula = Species ~ ., data = iris)
##
## Dimension: 4
##
## Percent Between-Group Variance Explained:
##     v1     v2     v3     v4
##  96.02  98.55  99.90 100.00
##
## Degrees of Freedom (per dimension): 5
##
## Training Misclassification Error: 0.02 ( N = 150 )
##
## Deviance: 15.102

data(glass)
# random sample of size 100
samp <- c(1, 3, 4, 11, 12, 13, 14, 16, 17, 18, 19, 20, 27, 28, 31,
          38, 42, 46, 47, 48, 49, 52, 53, 54, 55, 57, 62, 63, 64, 65,
          67, 68, 69, 70, 72, 73, 78, 79, 83, 84, 85, 87, 91, 92, 94,
          99, 100, 106, 107, 108, 111, 112, 113, 115, 118, 121, 123,
          124, 125, 126, 129, 131, 133, 136, 139, 142, 143, 145, 147,
          152, 153, 156, 159, 160, 161, 164, 165, 166, 168, 169, 171,
          172, 173, 174, 175, 177, 178, 181, 182, 185, 188, 189, 192,
          195, 197, 203, 205, 211, 212, 214) 
glass.train <- glass[samp,]
glass.test <- glass[-samp,]
glass.mda <- mda(Type ~ ., data = glass.train)
predict(glass.mda, glass.test, type="post") # abbreviations are allowed
confusion(glass.mda,glass.test)
}
\keyword{classif}
% Converted by Sd2Rd version 0.3-3.

\name{laplacian}
\Rdversion{1.1}
\alias{laplacian}
\title{create penalty object for two-dimensional smoothing.
}
\description{
  Creates a penalty matrix  for use by \code{gen.ridge} for
  two-dimensional smoothing.
}
\usage{
laplacian(size, compose)
laplacian(size = 16, compose = FALSE)
}
\arguments{
  \item{size}{dimension of the image is \code{size x size}; default is 16.}
  \item{compose}{default is \code{compose=FALSE}, which means the
    penalty is returned as an eigen-decomposition. If \code{compose=TRUE}, a penalty matrix is returned.}
}
\details{
Formulas are used to construct a laplacian for smoothing a square image.
}
\value{
  If \code{compose=FALSE}, an eigen-decomposition object is
  returned. The \code{vectors} component is a \code{size^2 x size^2}
  orthogonal matrix, and the \code{$values} component is a \code{size^2}
  vector of non-negative eigen-values. If \code{compose=TRUE}, these are
  multiplied together to form a single matrix.
}
\references{
Here we follow very closely the material on page 635 in JASA 1991 of O'Sullivan's article on discretized Laplacian Smoothing
}
\author{
Trevor Hastie <hastie@stanford.edu
}
\seealso{\code{\link{gen.ridge}},\code{\link{fda}}
}
\keyword{regression}

\name{predict.bruto}
\alias{predict.bruto}
\title{Predict method for BRUTO Objects}
\description{
  Predicted values based on `bruto' additive spline models which are fit
  by adaptive backfitting.
}
\usage{
\method{predict}{bruto}(object, newdata, type=c("fitted", "terms"), \dots)
}
\arguments{
  \item{object}{a fitted bruto object}
  \item{newdata}{values at which predictions are to be made.}
  \item{type}{if type is \code{"fitted"}, the fitted values are
    returned.  If type is \code{"terms"}, a list of fitted terms is
    returned, each with an \code{x} and \code{y} component.  These can
    be used to show the fitted functions.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  Either a fit matrix or a list of fitted terms.
}
\seealso{
  \code{\link{bruto}},
  \code{\link{predict}}
}
\examples{
data(trees)
fit1 <- bruto(trees[,-3], trees[3])
fitted.terms <- predict(fit1, as.matrix(trees[,-3]), type = "terms")
par(mfrow=c(1,2), pty="s")
for(tt in fitted.terms) plot(tt, type="l")
}
% Converted by Sd2Rd version 0.3-3.
\keyword{smooth}

\name{predict.mda}
\alias{predict.mda}
\title{Classify by Mixture Discriminant Analysis}
\description{
  Classify observations in conjunction with \code{mda}.
}
\usage{
\method{predict}{mda}(object, newdata, type, prior, dimension, g, \dots)
}
\arguments{
  \item{object}{a fitted mda object.}
  \item{newdata}{new data at which to make predictions.  If missing, the
    training data is used.}
  \item{type}{kind of predictions: \code{type = "class"} (default)
    produces a fitted factor, \code{type = "variates"} produces a matrix
    of discriminant variables (note that the maximal dimension is
    determined by the number of subclasses), \code{type = "posterior"}
    produces a matrix of posterior probabilities (based on a gaussian
    assumption), \code{type = "hierarchical"} produces the predicted
    class in sequence for models of dimensions specified by
    \code{dimension} argument.}
  \item{prior}{the prior probability vector for each class; the
    default is the training sample proportions.}
  \item{dimension}{the dimension of the space to be used, no larger
    than the dimension component of \code{object}, and in general less
    than the number of subclasses.  \code{dimension} can be a vector for
    use with \code{type = "hierarchical"}.}
  \item{g}{???}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  An appropriate object depending on \code{type}.  \code{object} has a
  component \code{fit} which is regression fit produced by the
  \code{method} argument to \code{mda}.  There should be a
  \code{predict} method for this object which is invoked.  This method
  should itself take as input \code{object} and optionally \code{newdata}.
}
\seealso{
  \code{\link{mda}},
  \code{\link{fda}},
  \code{\link{mars}},
  \code{\link{bruto}},
  \code{\link{polyreg}},
  \code{\link{softmax}},
  \code{\link{confusion}}
}
\examples{
data(glass)
samp <- sample(1:nrow(glass), 100)
glass.train <- glass[samp,]
glass.test <- glass[-samp,]
glass.mda <- mda(Type ~ ., data = glass.train)
predict(glass.mda, glass.test, type = "post") # abbreviations are allowed
confusion(glass.mda, glass.test)
}
\keyword{classif}
% Converted by Sd2Rd version 0.3-3.

\name{ESL.mixture}
\alias{ESL.mixture}
\title{Mixture example from "Elements of Statistical Learning"}
\description{
  A list with training data and other details for the mixture example
}
\usage{data(ESL.mixture)}
\format{
  This list contains the following elements:
  \describe{
    \item{x}{a 200x2 matrix of predictors.}
    \item{y}{a 200 vector of y values taking values 0 or 1.}
    \item{xnew}{a 6831x2 matrix of prediction points, on a 69x99 grid.}
    \item{prob}{a vector of 6831 probabilities - the true probabilities
      of a 1 at each point in \code{xnew}.}
    \item{marginal}{the marginal distribution of the predictors t each
      point in \code{xnew}.}
    \item{px1}{grid values for first coordinate in \code{xnew}.}
    \item{px2}{grid values for second coordinate in \code{xnew}.}
    \item{means}{a 20 x 2 matrix of means used in the generation of
      these data.}
  }
}
\source{
  "Elements of Statistical Learning (second edition)", Hastie, T., Tibshirani, R. and
  Friedman, J. (2009), Springer, New York.
  \url{https://hastie.su.domains/ElemStatLearn/}
}
\keyword{datasets}

\name{model.matrix.mars}
\alias{model.matrix.mars}
\title{Produce a Design Matrix from a `mars' Object}
\description{Produce a design matrix from a `mars' object.}
\usage{
\method{model.matrix}{mars}(object, x, which, full = FALSE, \dots)
}
\arguments{
  \item{object}{a mars object.}
  \item{x}{optional argument; if supplied, the mars basis functions are
    evaluated at these new observations.}
  \item{which}{which columns should be used. The default is to use the
    columns described by the component \code{selected.terms} on
    \code{object}.}
  \item{full}{if \code{TRUE} the entire set of columns are selected,
    even redundant ones.  This is used for updating a mars fit.}
  \item{\dots}{further arguments to be passed from or to methods.}
}
\value{
  A model matrix corresponding to the selected columns.
}
\seealso{
  \code{\link{mars}},
  \code{\link{predict.mars}}
}
\keyword{models}

\name{fda}
\alias{fda}
\alias{print.fda}
\title{Flexible Discriminant Analysis}
\description{
  Flexible discriminant analysis.}
\usage{
fda(formula, data, weights, theta, dimension, eps, method,
    keep.fitted, \dots)
}
\arguments{
  \item{formula}{of the form \code{y~x} it describes the response and
    the predictors.  The formula can be more complicated, such as
    \code{y~log(x)+z} etc (see \code{\link{formula}} for more details).
    The response should be a factor representing the response variable,
    or any vector that can be coerced to such (such as a logical
    variable).}
  \item{data}{data frame containing the variables in the formula
    (optional).}
  \item{weights}{an optional vector of observation weights.}
  \item{theta}{an optional matrix of class scores, typically with less
    than \code{J-1} columns.}
  \item{dimension}{The dimension of the solution, no greater than
    \code{J-1}, where \code{J} is the number classes.  Default is
    \code{J-1}.}
  \item{eps}{a threshold for small singular values for excluding
    discriminant variables; default is \code{.Machine$double.eps}.}
  \item{method}{regression method used in optimal scaling.  Default is
    linear regression via the function \code{polyreg}, resulting in
    linear discriminant analysis.  Other possibilities are \code{mars}
    and \code{bruto}.  For Penalized Discriminant analysis
    \code{gen.ridge} is appropriate.}
  \item{keep.fitted}{a logical variable, which determines whether the
    (sometimes large) component \code{"fitted.values"} of the \code{fit}
    component of the returned fda object should be kept.  The default is
    \code{TRUE} if \code{n * dimension < 5000}.}
  \item{\dots}{additional arguments to \code{method}.}
}
\value{
  an object of class \code{"fda"}.  Use \code{predict} to extract
  discriminant variables, posterior probabilities or predicted class
  memberships.  Other extractor functions are \code{coef},
  \code{confusion} and \code{plot}.

  The object has the following components:
  \item{percent.explained}{the percent between-group variance explained
    by each dimension (relative to the total explained.)}
  \item{values}{optimal scaling regression sum-of-squares for each
    dimension (see reference).  The usual discriminant analysis
    eigenvalues are given by \code{values / (1-values)}, which are used
    to define \code{percent.explained}.}
  \item{means}{class means in the discriminant space.  These are also
    scaled versions of the final theta's or class scores, and can be
    used in a subsequent call to \code{fda} (this only makes sense if
    some columns of theta are omitted---see the references).} 
  \item{theta.mod}{(internal) a class scoring matrix which allows
    \code{predict} to work properly.}
  \item{dimension}{dimension of discriminant space.}
  \item{prior}{class proportions for the training data.}
  \item{fit}{fit object returned by \code{method}.}
  \item{call}{the call that created this object (allowing it to be
    \code{update}-able)}
  \item{confusion}{confusion matrix when classifying the training data.}

  The \code{method} functions are required to take arguments \code{x}
  and \code{y} where both can be matrices, and should produce a matrix
  of \code{fitted.values} the same size as \code{y}.  They can take
  additional arguments \code{weights} and should all have a \code{\dots}
  for safety sake.  Any arguments to \code{method} can be passed on via
  the \code{\dots} argument of \code{fda}.  The default method
  \code{\link{polyreg}} has a \code{degree} argument which allows
  polynomial regression of the required total degree.  See the
  documentation for \code{\link{predict.fda}} for further requirements
  of \code{method}. The package \code{earth} is suggested for this
  package as well; \code{earth} is a more detailed implementation of
  the mars model, and works as a \code{method} argument.
}
\author{
  Trevor Hastie and Robert Tibshirani
}
\seealso{
  \code{\link{predict.fda}},
  \code{\link{plot.fda}},
  \code{\link{mars}},
  \code{\link{bruto}},
  \code{\link{polyreg}},
  \code{\link{softmax}},
  \code{\link{confusion}},
  %% \code{\link{coefficients.fda}}
}
\references{
  ``Flexible Disriminant Analysis by Optimal Scoring''  by Hastie,
  Tibshirani and Buja, 1994, JASA, 1255-1270.

  ``Penalized Discriminant Analysis'' by Hastie, Buja and Tibshirani, 1995,
  Annals of Statistics, 73-102.

  ``Elements of Statisical Learning - Data Mining, Inference and
  Prediction'' (2nd edition, Chapter 12) by Hastie, Tibshirani and
  Friedman, 2009, Springer
}
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
irisfit
## fda(formula = Species ~ ., data = iris)
##
## Dimension: 2 
##
## Percent Between-Group Variance Explained:
##     v1     v2 
##  99.12 100.00 
##
## Degrees of Freedom (per dimension): 5 
##
## Training Misclassification Error: 0.02 ( N = 150 )

confusion(irisfit, iris)
##            Setosa Versicolor Virginica 
##     Setosa     50          0         0
## Versicolor      0         48         1
##  Virginica      0          2        49
## attr(, "error"):
## [1] 0.02

plot(irisfit)

coef(irisfit)
##           [,1]        [,2]
## [1,] -2.126479 -6.72910343
## [2,] -0.837798  0.02434685
## [3,] -1.550052  2.18649663
## [4,]  2.223560 -0.94138258
## [5,]  2.838994  2.86801283

marsfit <- fda(Species ~ ., data = iris, method = mars)
marsfit2 <- update(marsfit, degree = 2)
marsfit3 <- update(marsfit, theta = marsfit$means[, 1:2]) 
## this refits the model, using the fitted means (scaled theta's)
## from marsfit to start the iterations
}
\keyword{classif}
% Converted by Sd2Rd version 0.3-3.

\name{predict.fda}
\alias{predict.fda}
\title{Classify by Flexible Discriminant Analysis}
\description{
  Classify observations in conjunction with \code{fda}.
}
\usage{
\method{predict}{fda}(object, newdata, type, prior, dimension, \dots)
}
\arguments{
  \item{object}{an object of class \code{"fda"}.}
  \item{newdata}{new data at which to make predictions.  If missing, the
    training data is used.}
  \item{type}{kind of predictions: \code{type = "class"} (default)
    produces a fitted factor, \code{type = "variates"} produces a matrix
    of discriminant (canonical) variables, \code{type = "posterior"} produces a
    matrix of posterior probabilities (based on a gaussian assumption),
    and \code{type = "hierarchical"} produces the predicted class in
    sequence for models of all dimensions.}
  \item{prior}{the prior probability vector for each class; the
    default is the training sample proportions.}
  \item{dimension}{the dimension of the space to be used, no larger
    than the dimension component  of \code{object}.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  An appropriate object depending on \code{type}. \code{object} has a
  component \code{fit} which is regression fit produced by the
  \code{method} argument to \code{fda}.  There should be a
  \code{predict} method for this object which is invoked.  This method
  should itself take as input \code{object} and optionally \code{newdata}.
}
\seealso{
  \code{\link{fda}},
  \code{\link{mars}},
  \code{\link{bruto}},
  \code{\link{polyreg}},
  \code{\link{softmax}},
  \code{\link{confusion}}
}
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
irisfit
## Call:
## fda(x = iris$x, g = iris$g)
## 
## Dimension: 2 
##
## Percent Between-Group Variance Explained:
##     v1  v2 
##  99.12 100
confusion(predict(irisfit, iris), iris$Species)
##            Setosa Versicolor Virginica
##     Setosa     50          0         0
## Versicolor      0         48         1
##  Virginica      0          2        49
## attr(, "error"):
## [1] 0.02
}
% Converted by Sd2Rd version 0.3-3.
\keyword{classif}

\name{plot.fda}
\alias{plot.fda}
\title{Plot for Flexible Discriminant Analysis}
\description{Plot in discriminant (canonical) coordinates a \code{fda} or (by inheritance) a \code{mda} object.
}
\usage{
\method{plot}{fda}(x, data, coords, group, colors, pch, mcolors, mpch, pcex, mcex, \dots)
}
\arguments{
  \item{x}{an object of class \code{"fda"}.}
  \item{data}{the data to plot in the discriminant coordinates. If
    \code{group="true"}, then data should be a data frame with the same variables that were used in the fit. If \code{group="predicted"},
    \code{data} need not contain the response variable, and can in fact
    be the correctly-sized \code{"x"} matrix.}
  \item{coords}{vector of coordinates to plot, with default
    \code{coords="c(1,2)"}. All pairs of plots are produced.}
  \item{group}{if \code{group="true"} (the default), each point is color
    and symbol coded according to the response in \code{data}. If
    \code{group="predicted"}, the class of each point is predicted from
    the model, and used instead.}
  \item{colors}{a vector of colors to be used in the plotting.}
  \item{pch}{a vector of plotting characters.}
  \item{mcolors}{a vector of colors for the class centroids; default is \code{colors}.}
  \item{mpch}{a vector of plotting characters for the centroids.}
  \item{pcex}{character expansion factor for the points; defualt is \code{pcex="0.5"}.}
  \item{mcex}{character expansion factor for the centroids; defualt is \code{pcex="2.5"}.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\seealso{
   \code{\link{fda}},
   \code{\link{mda}},
   \code{\link{predict.fda}}
 }
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
plot(irisfit)
data(ESL.mixture)
## Not a data frame
mixture.train=ESL.mixture[c("x","y")] 
mixfit=mda(y~x, data=mixture.train)
plot(mixfit, mixture.train)
plot(mixfit, data=ESL.mixture$xnew, group="pred")
}
\keyword{classif}

\name{coef.fda}
\alias{coef.fda}
\alias{coef.mda}
\title{
Produce coefficients for an fda or mda object
}
\description{
a method for coef for extracting the canonical coefficients from an fda or mda object
}
\usage{
\method{coef}{fda}(object, \dots)
}
\arguments{
  \item{object}{an \code{fda} or \code{mda} object.
}
  \item{\dots}{
not relevant
}
}
\details{
See the references for details.
}
\value{
A coefficient matrix}
\author{
  Trevor Hastie and Robert Tibshirani
}
\seealso{
  \code{\link{predict.fda}},
  \code{\link{plot.fda}},
  \code{\link{mars}},
  \code{\link{bruto}},
  \code{\link{polyreg}},
  \code{\link{softmax}},
  \code{\link{confusion}},
  %% \code{\link{coefficients.fda}}
}
\references{
  ``Flexible Disriminant Analysis by Optimal Scoring''  by Hastie,
  Tibshirani and Buja, 1994, JASA, 1255-1270.

  ``Penalized Discriminant Analysis'' by Hastie, Buja and Tibshirani, 1995,
  Annals of Statistics, 73-102.

  ``Elements of Statisical Learning - Data Mining, Inference and
  Prediction'' (2nd edition, Chapter 12) by Hastie, Tibshirani and
  Friedman, 2009, Springer
}
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
coef(irisfit)
mfit=mda(Species~.,data=iris,subclass=2)
coef(mfit)
}\keyword{classif}

\name{mspline}
\alias{mspline}
\title{
Vector Smoothing Spline
}
\description{
Fit a smoothing spline to a matrix of responses, single x.
}
\usage{
mspline(x, y, w, df = 5, lambda, thresh = 1e-04, \dots)
}
\arguments{
  \item{x}{
x variable (numeric vector).
}
  \item{y}{
response matrix.
}
  \item{w}{
optional weight vector, defaults to a vector of ones.
}
  \item{df}{
requested degrees of freedom, as in \code{smooth.spline}.
}
  \item{lambda}{
can provide penalty instead of df.
}
  \item{thresh}{
convergence threshold for df inversion (to lambda).
}
  \item{\dots}{
holdall for other arguments.
}
}
\details{
  This function is based on the ingredients of \code{smooth.spline},
  and allows for simultaneous smoothing of multiple responses
}
\value{
  A list is returned, with  a number of components, only some of which
  are of interest. These are
  \item{lambda}{The value of lambda used (in case df was supplied)}
  \item{df}{The df used (in case lambda was supplied)}
  \item{s}{A matrix like \code{y} of smoothed responses}
  \item{lev}{Self influences (diagonal of smoother matrix)}
}
\author{
Trevor Hastie
}
\seealso{
\code{\link{smooth.spline}}
}
\examples{
x=rnorm(100)
y=matrix(rnorm(100*10),100,10)
fit=mspline(x,y,df=5)
}

\name{softmax}
\alias{softmax}
\title{Find the Maximum in Each Row of a Matrix}
\description{Find the maximum in each row of a matrix.}
\usage{
softmax(x, gap = FALSE)
}
\arguments{
  \item{x}{a numeric matrix.}
  \item{gap}{if \code{TRUE}, the difference between the largest and
    next largest column is returned.}
}
\value{
  A factor with levels the column labels of \code{x} and values the
  columns corresponding to the maximum column. If \code{gap = TRUE} a
  list is returned, the second component of which is the difference
  between the largest and next largest column of \code{x}.
}
\seealso{
  \code{\link{predict.fda}},
  \code{\link{confusion}},
  \code{\link{fda}} 
  \code{\link{mda}} 
}
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
posteriors <- predict(irisfit, type = "post")
confusion(softmax(posteriors), iris[, "Species"])
}
% Converted by Sd2Rd version 0.3-3.
\keyword{utilities}

\name{mda.start}
\alias{mda.start}
\title{Initialization for Mixture Discriminant Analysis}
\description{Provide starting weights for the \code{mda} function which
  performs discriminant analysis by gaussian mixtures.}
\usage{
mda.start(x, g, subclasses = 3, trace.mda.start = FALSE,
          start.method = c("kmeans", "lvq"), tries = 5,
          criterion = c("misclassification", "deviance"), \dots)
}
\arguments{
  \item{x}{The x data, or an mda object.}
  \item{g}{The response vector g.}
  \item{subclasses}{number of subclasses per class, as in \code{mda}.}
  \item{trace.mda.start}{Show results of each iteration.}
  \item{start.method}{Either \code{"kmeans"} or \code{"lvq"}.  The
    latter requires package \bold{class} (from the \bold{VR} package
    bundle.}
  \item{tries}{Number of random starts.}
  \item{criterion}{By default, classification errors on the training
    data.  Posterior deviance is also an option.}
  \item{\dots}{arguments to be passed to the mda fitter when using
    posterior deviance.}
}
\value{
  A list of weight matrices, one for each class.
}
% Converted by Sd2Rd version 0.3-3.
\keyword{classif}

\name{predict.mars}
\alias{predict.mars}
\title{Predict method for MARS Objects}
\description{
  Predicted values based on `mars' multivariate adaptive regression
  spline models.
}
\usage{
\method{predict}{mars}(object, newdata, \dots)
}
\arguments{
  \item{object}{an object of class \code{"mars"}.}
  \item{newdata}{values at which predictions are to be made.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  the fitted values.
}
\seealso{
  \code{\link{mars}},
  \code{\link{predict}},
  \code{\link{model.matrix.mars}}
}
%%\examples{
%%## see help for mars()
%%}
% Converted by Sd2Rd version 0.3-3.
\keyword{smooth}

\name{confusion}
\alias{confusion}
\alias{confusion.default}
\alias{confusion.list}
\alias{confusion.fda}
\title{Confusion Matrices}
\description{
  Compute the confusion matrix between two factors, or for an fda or
  mda object.
}
\usage{
\method{confusion}{default}(object, true, \dots)
\method{confusion}{fda}(object, data, \dots)
}
\arguments{
  \item{object}{the predicted factor, or an fda or mda model object.}
  \item{true}{the true factor.}
  \item{data}{a data frame (list) containing the test data.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\details{
  This is a generic function.
}
\value{
  For the default method essentially \code{table(object, true)}, but
  with some useful attribute(s).
}
\seealso{
  \code{\link{fda}},
  \code{\link{predict.fda}}
}
\examples{
data(iris)
irisfit <- fda(Species ~ ., data = iris)
confusion(predict(irisfit, iris), iris$Species)
##            Setosa Versicolor Virginica 
##     Setosa     50          0         0
## Versicolor      0         48         1
##  Virginica      0          2        49
## attr(, "error"):
## [1] 0.02
}
% Converted by Sd2Rd version 0.3-3.
\keyword{category}

\name{polyreg}
\alias{polyreg}
\alias{predict.polyreg}
\title{Polynomial Regression}
\description{
  Simple minded polynomial regression.
}
\usage{
polyreg(x, y, w, degree = 1, monomial = FALSE, \dots)
}
\arguments{
  \item{x}{predictor matrix.}
  \item{y}{response matrix.}
  \item{w}{optional (positive) weights.}
  \item{degree}{total degree of polynomial basis (default is 1).}
  \item{monomial}{If \code{TRUE} a monomial basis is used (no cross
    terms).  Default is \code{FALSE}.}
  \item{\dots}{currently not used.}
}
\value{
  A polynomial regression fit, containing the essential ingredients for
  its \code{predict} method.
}
%%\seealso{
%%  \code{\link{predict.polyreg}}
%%}
% Converted by Sd2Rd version 0.3-3.
\keyword{regression}

\name{bruto}
\alias{bruto}
\title{
  Fit an Additive Spline Model by Adaptive Backfitting
}
\description{
  Fit an additive spline model by adaptive backfitting.
}
\usage{
bruto(x, y, w, wp, dfmax, cost, maxit.select, maxit.backfit, 
      thresh = 0.0001, trace.bruto = FALSE, start.linear = TRUE,
      fit.object, \dots)
}
\arguments{
  \item{x}{a matrix of numeric predictors (does not include the column
    of 1s).}
  \item{y}{a vector or matrix of responses.}
  \item{w}{optional observation weight vector.}
  \item{wp}{optional weight vector for each column of \code{y}; the RSS
    and GCV criteria use a weighted sum of squared residuals.}
  \item{dfmax}{a vector of maximum df (degrees of freedom) for each
    term.}
  \item{cost}{cost per degree of freedom; default is 2.}
  \item{maxit.select}{maximum number of iterations during the selection
    stage.}
  \item{maxit.backfit}{maximum number of iterations for the final
    backfit stage (with fixed lambda).} 
  \item{thresh}{convergence threshold (default is 0.0001); iterations
    cease when the relative change in GCV is below this threshold.} 
  \item{trace.bruto}{logical flag.  If \code{TRUE} (default) a progress
    report is printed during the fitting.}
  \item{start.linear}{logical flag.  If \code{TRUE} (default), the model
    starts with the linear fit.}
  \item{fit.object}{This the object returned by \code{bruto()}; if
    supplied, the same model is fit to the presumably new \code{y}.}
  \item{\dots}{further arguments to be passed to or from methods.}
}
\value{
  A multiresponse additive model fit object of class \code{"bruto"} is
  returned.  The model is fit by adaptive backfitting using smoothing
  splines.  If there are \code{np} columns in \code{y}, then \code{np}
  additive models are fit, but the same amount of smoothing (df) is 
  used for  each term.  The procedure chooses between \code{df = 0}
  (term omitted), \code{df = 1} (term linear) or \code{df > 0} (term
  fitted by smoothing spline).  The model selection is based on an
  approximation to the  GCV criterion, which is used at each step of the
  backfitting procedure.  Once the selection process stops, the model is
  backfit using the chosen amount of smoothing.

  A bruto object has the following components of interest:
  \item{lambda}{a vector of chosen smoothing parameters, one for each
    column of \code{x}.}
  \item{df}{the df chosen for each column of \code{x}.}
  \item{type}{a factor with levels \code{"excluded"}, \code{"linear"} or
    \code{"smooth"}, indicating the status of each column of \code{x}.}
  \item{gcv.select gcv.backfit df.select}{The sequence of gcv values and
    df selected during the execution of the function.}
  \item{nit}{the number of iterations used.}
  \item{fitted.values}{a matrix of fitted values.}
  \item{residuals}{a matrix of residuals.}
  \item{call}{the call that produced this object.}
}
\references{
  Trevor Hastie and Rob Tibshirani,
  \emph{Generalized Additive Models},
  Chapman and Hall, 1990 (page 262).

  Trevor Hastie, Rob Tibshirani and Andreas Buja
  ``Flexible Discriminant Analysis by Optimal Scoring''
  JASA 1994, 89, 1255-1270.
}
\seealso{
  \code{predict.bruto}
}
\examples{
data(trees)
fit1 <- bruto(trees[,-3], trees[3])
fit1$type
fit1$df
## examine the fitted functions
par(mfrow=c(1,2), pty="s")
Xp <- matrix(sapply(trees[1:2], mean), nrow(trees), 2, byrow=TRUE)
for(i in 1:2) {
  xr <- sapply(trees, range)
  Xp1 <- Xp; Xp1[,i] <- seq(xr[1,i], xr[2,i], len=nrow(trees))
  Xf <- predict(fit1, Xp1)
  plot(Xp1[ ,i], Xf, xlab=names(trees)[i], ylab="", type="l")
}
}
% Converted by Sd2Rd version 0.3-3.
\keyword{smooth}

\name{gen.ridge}
\alias{gen.ridge}
\alias{predict.gen.ridge}
\title{Penalized Regression}
\description{
  Perform a penalized regression, as used in penalized discriminant
  analysis.
}
\usage{
gen.ridge(x, y, weights, lambda=1, omega, df, \dots)
}
\arguments{
  \item{x, y, weights}{the x and y matrix and possibly a weight vector.}
  \item{lambda}{the shrinkage penalty coefficient.}
  \item{omega}{a penalty object; omega is the eigendecomposition of
    the penalty matrix, and need not have full rank.  By default,
    standard ridge is used.}
  \item{df}{an alternative way to prescribe lambda, using the notion
    of equivalent degrees of freedom.}
  \item{\dots}{currently not used.}
}
\value{
  A generalized ridge regression, where the coefficients are penalized
  according to omega.  See the function definition for further details.
  No functions are provided for producing one dimensional penalty
  objects (omega).
  \code{\link{laplacian}()} creates a two-dimensional penalty
  object, suitable for (small) images.
}
\seealso{\code{\link{laplacian}}}
\keyword{regression}
